/*==========================================================================*/
/* This program processes a script file to test message flows,              */
/* MsgTest.c contains the programs mainline and usage display functions     */
/*==========================================================================*/
/* (C) Copyright Tim Armstrong 2003, 2004, 2005, 2006, 2007, 2008           */
/*==========================================================================*/
/* MsgTest uses a conventional memory structure where individual control    */
/* blocks are chained off each other. The root control block is of type     */
/* MainCtrlBlk and is pointed to by the global variable kpMCB. Most of the  */
/* programs control blocks are contained in double-linked-lists with Head   */
/* and Tail pointers.                                                       */
/*                                                                          */
/* MainCtrlBlk                                                              */
/* ===================                                                      */
/* TestInfoBlk*       lpTIBReadyHead;                                       */
/* TestInfoBlk*       lpTIBReadyTail;                                       */
/*   This linked list contains the TestInfoBlk control blocks of <Test>'s   */
/*   that are waiting for exectuion. <Test>'s are loaded onto this linked   */
/*   list during the script parsing phase that occurs during program        */
/*   initialisation.                                                        */
/* TestInfoBlk*       lpTIBActiveHead;                                      */
/* TestInfoBlk*       lpTIBActiveTail;                                      */
/*   This linked list contains the TestInfoBlk control blocks of <Test>'s   */
/*   that are being executed. It is loaded after initialisation is complete */
/*   by moving the maximum number of active <Test>'s, as configured in the  */
/*   <Control> parameter <Concurrency>, from the Ready linked list to the   */
/*   Active(this) linked list. When a <Test> completes its removed from     */
/*   this linked list and another one copied from the Ready linked list     */
/*   until both lists have been exhausted or a fatal error occurs.          */
/* VarInfoBlk*        lpVIBHead;                                            */
/* VarInfoBlk*        lpVIBTail;                                            */
/*   As part of program invocation a user may specify extra parameters      */
/*   after the script name. Those parameters are held as <Str> variables    */
/*   on this linked list with name %1%, %2%, ..., %n%                       */
/* ForBackRefBlk*     lpFBRHead;                                            */
/* ForBackRefBlk*     lpFBRTail;                                            */
/*   This linked list is used as a stack during parsing to contain          */
/*   references to the ForStartBlk control blocks so that the ForStartBlk   */
/*   control block pointer in the ForEndBlk can be set approrpriately       */
/* IfBackRefBlk*      lpIFBHead;                                            */
/* IfBackRefBlk*      lpIFBTail;                                            */
/*   This linked list is used as a stack during parsing to contain          */
/*   references to the IfStartBlk control blocks so that the IfStartBlk     */
/*   control block pointer is available for setting IfCondFalseTarget's     */
/* IntervalInfoBlk*   lpIIBHead;                                            */
/* IntervalInfoBlk*   lpIIBTail;                                            */
/*   This linked list is to keep track of timing intervals requested by     */
/*   the user. Intervals are at script scope rather than test scope.        */
/* ControlInfoBlk*    lpCIB;                                                */
/*   Pointer to the control block loaded with the information from the      */
/*   <Control> section of a script during parsing                           */
/* DefaultsInfoBlk*   lpDIB;                                                */
/*   Pointer to the control block loaded with the information from the      */
/*   <Defaults> section of a script during parsing. This includes default   */
/*   values for variables and commands                                      */
/* ConnectionDefBlk*  lpCDBHead;                                            */
/* ConnectionDefBlk*  lpCDBTail;                                            */
/*   This linked list is used to hold the connection definition control     */
/*   blocks which contain information about the queue managers that the     */
/*   program has connected to                                               */
/* SecondaryParseBlk* lpSPB;                                                */
/*   This control block is used for "Secondary" parsing of <MQMDFile>'s     */
/*   and <RFH2File>'s during the execution of a script <Test>'s             */
/*                                                                          */
/* MsgSetParseBlk*    lpMSPB;                                               */
/*   This control block is used for parsing message sets                    */
/*                                                                          */
/* TestInfoBlk                                                              */
/* ===================                                                      */
/* TestInfoBlk*       lpTIBNext;                                            */ 
/* TestInfoBlk*       lpTIBPrev;                                            */
/*   Pointers to the next and previous <Test>'s on the linked list          */ 
/* CommandCtrlBlk*    lpCCBHead;                                            */
/* CommandCtrlBlk*    lpCCBTail;                                            */
/*   <Test>'s consist of a series of commands that change the value of a    */
/*   variable or do something. This linked list contains the command        */
/*   elements attached to the <Test> during the script parsing phase        */
/* CommandCtrlBlk*    lpCCBCurr;                                            */
/*   Pointer to the currently executing command. Commands go through        */
/*   several states. S_INIT where the command control blocks values are     */
/*   initialised. S_RUN where the program attempts to execute the requested */
/*   function until it succeeds, runs out of time or fails. S_COMPLETE      */
/*   is the state set when S_RUN succeeds. S_FAILURE is the state set when  */
/*   S_RUN fails. S_NEXT is the final state telling the command processor   */
/*   to move onto the next command or if there are no more commands then    */
/*   to complete the <Test>                                                 */
/* VarInfoBlk*        lpVIBHead;                                            */
/* VarInfoBlk*        lpVIBTail;                                            */
/*   This linked list is used to contains a <Test>'s variables. A variables */
/*   value is not set until the variable command <Str>, <Num>, <MQMD>,      */
/*   <RFH2> etc. is executed. The first time a variable command executes it */
/*   will copy the value of a variable with the same name from the          */
/*   <Defaults> section.                                                    */
/* BuffInfoBlk*       lpBIBHead;                                            */
/* BuffInfoBlk*       lpBIBTail;                                            */
/*   This linked list is used to contains the <Test>'s buffers.             */
/* QueueInfoBlk*      lpQIBHead;                                            */
/* QueueInfoBlk*      lpQIBTail;                                            */
/*   When an MQ queue needs to be opened during the execution of a Test the */
/*   queue opening function checks the QueueInfoBlk's on this linked list   */
/*   to see if it has already been opened. If so a pointer to the QIB       */
/*   containing the MQHOBJ handle of the opened queue is passed back. If a  */
/*   matching QIB cannot be found then the queue is opened, a new QIB is    */
/*   added to this linked list and the pointer to the new QIB is returned.  */
/*==========================================================================*/
/* MsgTest uses the EXPAT streaming XML parser whose copyright information  */
/* appears immediately below                                                */
/*==========================================================================*/
/* MsgTest uses the PCRE Perl Compatible Regular Expressions library whose  */
/* copyright information appears immediately below                          */
/*==========================================================================*/

/*==========================================================================*/
/* EXPAT Copyright information                                              */
/*==========================================================================*/
/*
Copyright (c) 1998, 1999, 2000 Thai Open Source Software Center Ltd
                               and Clark Cooper
Copyright (c) 2001, 2002 Expat maintainers.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/*==========================================================================*/
/* End of EXPAT Copyright information                                       */
/*==========================================================================*/

/*==========================================================================*/
/* PCRE Copyright information                                               */
/*==========================================================================*/
/*
           Copyright (c) 1997-2005 University of Cambridge

-----------------------------------------------------------------------------
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    * Neither the name of the University of Cambridge nor the names of its
      contributors may be used to endorse or promote products derived from
      this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
-----------------------------------------------------------------------------
*/
/*==========================================================================*/
/* End of PCRE Copyright information                                        */
/*==========================================================================*/

/*==========================================================================*/
/* Includes                                                                 */
/*==========================================================================*/
#ifndef __TJAINC_MSGTEST_H
#include "MsgTest.h"
#endif /* __TJAINC_MSGTEST_H */


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     ShowUsage                                                            */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Display usage information on the console                             */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     N/A                                                                  */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
ShowUsage()
{
#ifdef __MSGTEST_S
    printf("MsgTestS V1.3.0 Compiled on %s at %s\n", __DATE__, __TIME__);
    printf("(C) Copyright Tim Armstrong 2003, 2004, 2005, 2006, 2007\n");
    
    printf("Usage: MsgTestS <ScriptFile> %%1%% %%2%% ... %%n%%\n");
#else
    printf("MsgTest V1.3.0 Compiled on %s at %s\n", __DATE__, __TIME__);
    printf("(C) Copyright Tim Armstrong 2003, 2004, 2005, 2006, 2007\n");
    
    printf("Usage: MsgTest <ScriptFile> %%1%% %%2%% ... %%n%%\n");
#endif
}

MainCtrlBlk* kpMCB;
/*==========================================================================*/
/* FUNCTION:                                                                */
/*     main                                                                 */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Programs mainline. Calls the various script processing functions     */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     argc  Count of program parameters                                    */
/*     argv  Pointer to program parameter pointers                          */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     Zero    If successful                                                */
/*     NonZero If FAILURE                                                   */
/*==========================================================================*/
int
main(int argc, char** argv)
{
    MainCtrlBlk oMCB;

    /* Set main control block pointer */
    kpMCB = &oMCB;

    /* No parameters entered so tell the user how to run the program */
    if(2 > argc)
    {
        ShowUsage();
        return(99999);
    }

    /* Initialise the programs control blocks */
    if(FAILURE == InitCBs(argc, argv))
    {
        ConsoleMsg(153, "InitCBs\0");
        FreeResources();
        return(88888);
    }

    /* Load the script file removing excess whitespace */
    if(FAILURE == LoadXMLFile(oMCB.lpszScriptFileName, &oMCB.lpszScriptBuff,
                              &oMCB.lScriptXmlLen, TRUE))
    {
        ConsoleMsg(153, "LoadXMLFile\0");
        FreeResources();
        return(77777);
    }

    /* Parse the script file */
    if(FAILURE == ParseScriptFile())
    {
        ConsoleMsg(153, "ParseScriptFile\0");
        FreeResources();
        return(66666);
    }

    /* Resolve any program argument variables in the CIB and DIB */
    if(FAILURE == ResolveCIBVars())
    {
        ConsoleMsg(153, "ResolveCIBVars\0");
        FreeResources();
        return(55555);
    }

    /* Open the log file if any */
    if(FAILURE == OpenLogFile())
    {
        ConsoleMsg(153, "OpenLogFile\0");
        FreeResources();
        return(44444);
    }

    /* Open the report file if any */
    if(FAILURE == OpenReportFile())
    {
        ConsoleMsg(153, "OpenReportFile\0");
        FreeResources();
        return(33333);
    }

    /* Open the user log file if any */
    if(FAILURE == OpenUserLogFile())
    {
        ConsoleMsg(153, "OpenUserLogFile\0");
        FreeResources();
        return(22222);
    }

    /* Connect to MQ etc. */
    if(FAILURE == OpenResources())
    {
        ConsoleMsg(153, "OpenResources\0");
        FreeResources();
        return(11111);
    }

    /* Load Message Set Files */
    if(FAILURE == LoadAndParseMsgSetFiles())
    {
        ConsoleMsg(153, "LoadAndParseMsgSetFiles\0");
        FreeResources();
        return(9999);
    }

    /* Load Message Set Files */
    if(FAILURE == PrepareMsgSets())
    {
        ConsoleMsg(153, "PrepareMsgSets\0");
        FreeResources();
        return(8888);
    }

    /* Complete the initialisation of the control blocks related to each test */
    if(FAILURE == PrepareTests())
    {
        ConsoleMsg(153, "PrepareTests\0");
        FreeResources();
        return(7777);
    }

    /* Run the actual tests */
    RunTests();

    /* Report run time statistics for intervals and message rates */
    ReportStats();

    /* Close files, close queues, free memory */
    FreeResources();

    /* Return for main() is the number of failed and incomplete tests plus the number of failed commands */
    return oMCB.lFailedTests + oMCB.lIncompleteTests + oMCB.lScriptFailedGetFile + oMCB.lScriptFailedPutFile +
           oMCB.lScriptFailedGetMsg + oMCB.lScriptFailedPutMsg + oMCB.lScriptFailedCompare;
}
