/*==========================================================================*/
/* (C) Copyright Tim Armstrong 2003, 2004, 2005, 2006, 2007, 2008           */
/*==========================================================================*/
#ifndef __TJAINC_MSGTEST_H
#define __TJAINC_MSGTEST_H

#ifdef WIN32
#include <process.h>
#include <wtypes.h>
#include <winbase.h>
#else
#include <unistd.h>
#include <sys/time.h>
#include <errno.h>
#include <locale.h>
#include <langinfo.h>
#endif /* WIN32 */

#ifndef __TJAINC_CMQC_H
#define __TJAINC_CMQC_H
#include <cmqc.h>
#endif /* __TJAINC_CMQC_H */

#ifndef __TJAINC_CMQXC_H
#define __TJAINC_CMQXC_H
#include <cmqxc.h>
#endif  /* __TJAINC_CMQXC_H */

#ifndef __TJAINC_CMQCFC_H
#define __TJAINC_CMQCFC_H
#include <cmqcfc.h>
#endif  /* __TJAINC_CMQCFC_H */

#ifndef __TJAINC_CTYPE_H
#define __TJAINC_CTYPE_H
#include <ctype.h>
#endif /* __TJAINC_CTYPE_H */

#ifndef __TJAINC_EXPAT_H
#define __TJAINC_EXPAT_H
#include "expat.h"
#endif /* __TJAINC_EXPAT_H */

#ifndef __TJAINC_STDARG_H
#define __TJAINC_STDARG_H
#include <stdarg.h>
#endif /* __TJAINC_STDARG_H */

#ifndef __TJAINC_STDIO_H
#define __TJAINC_STDIO_H
#include <stdio.h>
#endif /* __TJAINC_STDIO_H */

#ifndef __TJAINC_STDLIB_H
#define __TJAINC_STDLIB_H
#include <stdlib.h>
#endif /* __TJAINC_STDLIB_H */

#ifndef __TJAINC_STRING_H
#define __TJAINC_STRING_H
#include <string.h>
#endif /* __TJAINC_STRING_H */

#ifndef __TJAINC_MATH_H
#define __TJAINC_MATH_H
#include <math.h>
#endif /* __TJAINC_MATH_H */

#ifndef __TJAINC_TIME_H
#define __TJAINC_TIME_H
#include <time.h>
#endif /* __TJAINC_TIME_H */

#ifndef __TJAINC_PCRE_H
#include "pcre.h"
#endif /* __TJAINC_PCRE_H */


/*==========================================================================*/
/* Typedefs                                                                 */
/*==========================================================================*/
#ifdef WIN32
    typedef __int16 Int16;
    typedef __int64 Int64;
#else
    typedef short Int16;
    typedef long long int Int64;
#endif /* WIN32 */


/*==========================================================================*/
/* Enumerations                                                             */
/*==========================================================================*/
enum eArithmeticEnum{A_UNKNOWN, A_ADD, A_SUBTRACT, A_MULTIPLY, A_DIVIDE,
                     A_EVAL};

enum eBrowseEnum{B_NONE, B_BROWSEFIRST, B_BROWSENEXT, B_BROWSEAUTO};

enum eCommandEnum{C_UNKNOWN, C_GETFILE, C_PUTFILE, C_INLINEDATA, C_OVERLAY,
                  C_EXTRACT, C_SUBSTRING, C_GETMSG, C_PUTMSG, C_SELECT,
                  C_WAITONMSGDEPTH, C_WAITONTIME, C_COMPARE, C_COMMIT,
                  C_BACKOUT, C_SYSTEM, C_FORSTART, C_FOREND, C_FINALLYSTART,
                  C_FINALLYEND, C_INTERVALSTART, C_INTERVALEND, C_VAR, C_MATH,
                  C_STDIN, C_STDOUT, C_USERLOG, C_MQSCCOMMAND, C_IFSTART, 
                  C_IFEND, C_ELSEIF, C_ELSE, C_CONTINUE, C_BREAK, C_EXITTEST,
                  C_FAILTEST, C_LENGTH, C_POSITION};

enum eElementEnum{E_UNKNOWN, E_MSGTEST, E_CONTROL, E_DEFAULTS, E_TEST,
                  E_CONNECTION, E_QMGR, E_CHANNEL, E_HOST, E_PORT, E_TESTLOG,
                  E_TESTREPORT, E_TESTUSERLOG, E_TESTINTERVAL, E_TESTMSGSET,
                  E_TESTMAXWAIT, E_NEXTLOGMSGINTERVAL, E_CONCURRENCY,
                  E_CONCURRENCYWAIT, E_ONERROR, E_NUM, E_DOUBLE, E_STR,
                  E_RANDOMNUM, E_MQMD, E_RFH2, E_LIST, E_GETFILE, E_PUTFILE,
                  E_INLINEDATA, E_OVERLAY, E_EXTRACT, E_SUBSTRING, E_GETMSG,
                  E_PUTMSG, E_SELECT, E_WAITONMSGDEPTH, E_WAITONTIME,
                  E_COMPARE, E_COMMIT, E_BACKOUT, E_SYSTEM, E_FOR, E_FOREVER,
                  E_FOREACH, E_FINALLY, E_BUFFER, E_FILE, E_FILEPREFIX,
                  E_REGEX, E_LISTFILE, E_MQMDFILE, E_RFH2FILE, E_DIR,
                  E_CMDMAXWAIT, E_GMOOPTIONS, E_PMOOPTIONS, E_GMOMATCHOPTIONS,
                  E_GMOWAIT, E_DATA, E_Q, E_DEPTH, E_INTERVAL, E_INTERVALMILLIS,
                  E_TIME, E_RFH2COMP, E_MASK, E_COMMAND, E_IGNORERC,
                  E_SYNCPOINT, E_VERSION, E_REPORT, E_MSGTYPE, E_EXPIRY,
                  E_FEEDBACK, E_ENCODING, E_CODEDCHARSETID, E_FORMAT,
                  E_PRIORITY, E_PERSISTENCE, E_MSGID, E_CORRELID,
                  E_BACKOUTCOUNT, E_REPLYTOQ, E_REPLYTOQMGR, E_USERIDENTIFIER,
                  E_ACCOUNTINGTOKEN, E_APPLIDENTITYDATA, E_PUTAPPLTYPE,
                  E_PUTAPPLNAME, E_PUTDATE, E_PUTTIME, E_APPLORIGINDATA,
                  E_GROUPID, E_MSGSEQNUMBER, E_OFFSET, E_MSGFLAGS,
                  E_ORIGINALLENGTH, E_MQRFH2, E_FOLDER, E_LISTELEMENT,
                  E_SETIDENTITYCONTEXT, E_SETALLCONTEXT, E_INTERVALSTART,
                  E_INTERVALEND, E_INTERVALINMSGID, E_INTERVALINCORRELID,
                  E_CONVERT, E_CLEAR, E_CONTINUEONRC, E_BREAKONRC,
                  E_SUPPRESSRC, E_MESSAGERATE, E_PARSERNAME, E_BROWSE,
                  E_GROUP, E_APPEND, E_STDIN, E_STDOUT, E_USERLOG,
                  E_LOGANDCONTINUE, E_ADD, E_SUBTRACT, E_MULTIPLY, E_DIVIDE,
                  E_EVAL, E_CONVTABLEDIR, E_LOCALENCODING, E_LOCALCODEDCHARSETID,
                  E_KEYREPOSITORY, E_SSLCIPHERSPEC, E_SSLPEERNAME,
                  E_MQSCCOMMAND, E_MQSCREPLYDYNAMICQ, E_REMOVEDLH, E_REMOVEXQH,
                  E_IF, E_ELSEIF, E_ELSE, E_CONTINUE, E_BREAK, E_EXITTEST,
                  E_FAILTEST, E_LENGTH, E_POSITION, E_MAXMSGLENGTH};

enum eForTypeEnum{F_UNKNOWN, F_FOR, F_FOREVER, F_FOREACH};

enum eGroupEnum{G_NONE, G_GROUPNEXT, G_GROUPLAST, G_GROUPAUTO};

enum eIntervalEnum{I_UNKNOWN, I_START, I_END};

enum eMsgModeEnum{M_UNKNOWN, M_BURST, M_PACED, M_RANDOM, M_INCREMENTAL};

enum eMsgSetTreeNodeEnum{N_UNKNOWN, N_ROOT, N_MSG, N_SEGMENT, N_ELEMENT};

enum eOutputFormatEnum{O_UNKNOWN, O_FIXED}; /* ???TA O_SDV(Separator Delimited) O_XML */

enum eMSLLBlockEnum{P_UNKNOWN, P_MSGSET, P_MSGTEMPLATES, P_MESSAGES, P_MSG,
                    P_ELEMENT, P_SEGDEF, P_SEGMENT, P_MSGDEF, P_ROOT};

enum eCmqValueTypeEnum{Q_UNKNOWN, Q_MQRO, Q_MQMT, Q_MQFMT, Q_MQMI, Q_MQCI, Q_MQGI,
                       Q_MQACT, Q_MQAT, Q_MQMF, Q_MQMO, Q_MQGMO, Q_MQPMO, Q_MQMD,
                       Q_MQEI, Q_MQPRI, Q_MQPER, Q_MQOL, Q_MQENC, Q_MQCCSI,
                       Q_MQFB, Q_MQIIH, Q_MQIAUT, Q_MQITII, Q_MQITS, Q_MQICM,
                       Q_MQISS, Q_MQCIH, Q_MQCRC, Q_MQCC, Q_MQRC, Q_MQCUOWC,
                       Q_MQCGWI, Q_MQCLT, Q_MQCODL, Q_MQCADSD, Q_MQCCT, Q_MQCTES,
                       Q_MQCFAC, Q_MQCFUNC, Q_MQCSC, Q_MQRFH};

enum eRecursionCallEnum{R_UNKNOWN, R_ROOT, R_CHILD, R_SIBLING};

enum eCmdStateEnum{S_UNKNOWN, S_INIT, S_RUN, S_COMPLETE, S_NEXT, S_FAILURE,
                   S_BREAK, S_CONTINUE};

enum eElementTypeEnum{T_UNKNOWN, T_CHAR, T_HEX, T_INT8, T_INT16, T_INT32, T_INT64, T_PD};

enum eVariableEnum{V_UNKNOWN, V_NUM, V_DOUBLE, V_STR, V_RANDOMNUM, V_LIST, V_MQMD, V_RFH2};

enum eWaitTypeEnum{W_UNKNOWN, W_INTERVAL, W_INTERVALMILLIS, W_TIME};

enum eSetContextTypeEnum{X_UNKNOWN, X_NONE, X_SETIDENTITY, X_SETALL};

enum eEvalTypeEnum{Y_UNKNOWN, Y_ADD, Y_SUBTRACT, Y_MULTIPLY, Y_DIVIDE, Y_REM};

enum ePredicateTypeEnum{Z_UNKNOWN, Z_EQUALTO, Z_NOTEQUALTO, Z_GREATER, Z_LESS,
                        Z_GREATEREQUAL, Z_LESSEQUAL};

/*==========================================================================*/
/* Defines                                                                  */
/*==========================================================================*/
#define MQSC_EXTEND_LEN            65500
#define INITIAL_BUFF_LEN           32732
#define RESOLVE_BUFF_LEN           4096
#define ERROR_WORK_BUFF_LEN        4096
#define FILE_BUFF_LEN              1024
#define LOG_BUFF_LEN               1024
#define REPORT_BUFF_LEN            1024
#define EXTRA_MSG_INFO_LEN         1024
#define VAR_INFO_LEN               1024
#define STDIN_BUFF_LEN             1024
#define NAME_BUFF_LEN              256
#define ATTR_BUFF_LEN              256
#define MAX_TESTS                  256
#define WORK_BUFF_LEN              128
#define CODE_PAGE_BUFF_LEN         32
#define NUM_BUFF_LEN               16
#define TIME_BUFF_LEN              16
#define MAX_SCRIPT_PARSE_DEPTH     16
#define MAX_CTRLBLK_PARSE_DEPTH    2
#define MAX_MSGSET_PARSE_DEPTH     32
#define TRUE                       1
#define FALSE                      0
#define UNKNOWN                    -1
#define FAILURE                    -1
#define SUCCESS                    0
#define MAX_REASON_CODES           16
#define EXTRA_LINE                 999
#define TEST_EXTRA_LINE            998
#define EXTRA_LINE_ERR             997
#define TEST_EXTRA_LINE_ERR        996
#define TEST_EXTRA_LINE_PCF        995
#define MTCC_OK                    0
#define MTCC_WARN                  1
#define MTCC_FAIL                  2
#define MTRC_NONE                  0
#define MTRC_LOGIC_FAIL            9970 /* Generic logic error failure code     */
#define MTRC_INIT_FAIL             9980 /* Generic initialisation failure code  */
#define MTRC_INIT_BUFF_FAIL        9981 /* Buffer location/allocation failure   */
#define MTRC_INIT_RESOLVE_FAIL     9982 /* Variable resolution failure code     */
#define MTRC_INIT_CONVNUM_FAIL     9983 /* Numeric conversion failure           */
#define MTRC_INIT_LOCATE_FAIL      9984 /* Generic location  failure            */
#define MTRC_RUN_FAIL              9990 /* Generic run failure code             */
#define MTRC_RUN_BUFF_FAIL         9991 /* Buffer location/allocation failure   */
#define MTRC_RUN_MATCH_FAIL        9992 /* Compare matching failed              */
#define MTRC_MALLOC_FAIL           9993 /* Memory allocation failed             */

/*==========================================================================*/
/* Flags                                                                    */
/*==========================================================================*/
#define MSTN_NONE                  0x00000000
#define MSTN_SET_SEGLEN            0x00000001
#define MSTN_SET_SEGLEN_M1         0x00000002
#define MSTN_SET_SEGLEN_M2         0x00000004
#define MSTN_SET_SEGLEN_M4         0x00000008
#define MSTN_SET_SEGLEN_MASK       0x0000000f

/*==========================================================================*/
/* Structs                                                                  */
/*==========================================================================*/
/*--------------------------------------------------------------------------*/
/* Connection Defintion Block structure                                     */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct ConnectionDefBlk*       lpCDBNext;
    struct ConnectionDefBlk*       lpCDBPrev;
    char*                          lpszConnectionName;
    char*                          lpszConnQMgrName;
    char*                          lpszConnChannelName;
    char*                          lpszConnHostName;
    char*                          lpszConnPortNum;
    char*                          lpszConnKeyRepository;
    char*                          lpszConnSSLCipherSpec;
    char*                          lpszConnSSLPeerName;
    char*                          lpszConnMaxMsgLength;
    MQLONG                         lCCSID;
    MQHCONN                        hConn;
    long                           bWasAlreadyConnected;
} ConnectionDefBlk;

/*--------------------------------------------------------------------------*/
/* Linked List structure. All program elements that are in lists have Next  */
/* and Prev element pointers as the first two elements in their structures  */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct LinkedListBlk*          lpLLBNext;
    struct LinkedListBlk*          lpLLBPrev;
} LinkedListBlk;

/*--------------------------------------------------------------------------*/
/* Folder Info Block structure                                              */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct FolderInfoBlk*          lpFIBNext;
    struct FolderInfoBlk*          lpFIBPrev;
    char*                          lpszFolderName;
    char*                          lpszFolderData;
    MQLONG                         lFolderDataLen;
    long                           bAlreadyMatched;
} FolderInfoBlk;

/*--------------------------------------------------------------------------*/
/* List Element Info Block structure                                        */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct ListElementInfoBlk*     lpLEBNext;
    struct ListElementInfoBlk*     lpLEBPrev;
    char*                          lpszListElementValue;
} ListElementInfoBlk;

/*--------------------------------------------------------------------------*/
/* Queue Info Block structure                                               */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct QueueInfoBlk*           lpQIBNext;
    struct QueueInfoBlk*           lpQIBPrev;
    char*                          lpszTargQName;
    char*                          lpszTargQMgrName;
    char*                          lpszTargDynQName;
    char*                          lpszConnQMgrName;
    char*                          lpszResolvedQName;
    MQHCONN                        hConn;
    MQHOBJ                         hQ;
} QueueInfoBlk;

/*--------------------------------------------------------------------------*/
/* Variable Info Block structure                                            */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct VarInfoBlk*             lpVIBNext;
    struct VarInfoBlk*             lpVIBPrev;
    enum eVariableEnum             eVarType;
    char*                          lpszVarName;
    char*                          lpszVarFormat;
    /* Depending on eVarType one of the following fields will be used */
    char*                          lpszCurrStrValue;
    void*                          lpCtrlBlk;
} VarInfoBlk;

/*--------------------------------------------------------------------------*/
/* List Info Block structure                                                */
/*--------------------------------------------------------------------------*/
typedef struct
{
    ListElementInfoBlk*            lpLEBHead;
    ListElementInfoBlk*            lpLEBTail;
    char*                          lpLEBPointerArray; /* ???TA Implement V130 */
    char*                          lpszListFileName;
    char*                          lpszDir;
    char*                          lpszResListFileName;
    char*                          lpszResDir;
    long                           lListElementCount;
    long                           bDynamicLIB;
} ListInfoBlk;

/*--------------------------------------------------------------------------*/
/* MQMD Info Block structure                                                */
/*--------------------------------------------------------------------------*/
typedef struct
{
    /* MQMD related fields */
    char*                          lpszVersion;
    char*                          lpszReport;
    char*                          lpszMsgType;
    char*                          lpszExpiry;
    char*                          lpszFeedback;
    char*                          lpszEncoding;
    char*                          lpszCodedCharSetId;
    char*                          lpszFormat;
    char*                          lpszPriority;
    char*                          lpszPersistence;
    char*                          lpszMsgId;
    char*                          lpszCorrelId;
    char*                          lpszBackoutCount;
    char*                          lpszReplyToQ;
    char*                          lpszReplyToQMgr;
    char*                          lpszUserIdentifier;
    char*                          lpszAccountingToken;
    char*                          lpszApplIdentityData;
    char*                          lpszPutApplType;
    char*                          lpszPutApplName;
    char*                          lpszPutDate;
    char*                          lpszPutTime;
    char*                          lpszApplOriginData;
    char*                          lpszGroupId;
    char*                          lpszMsgSeqNumber;
    char*                          lpszOffset;
    char*                          lpszMsgFlags;
    char*                          lpszOriginalLength;
    /* MsgTest related fields */
    char*                          lpszMQMDFileName;
    char*                          lpszDir;
    char*                          lpszResMQMDFileName;
    char*                          lpszResDir;
    long                           bDynamicMIB;
} MQMDInfoBlk;

/*--------------------------------------------------------------------------*/
/* RFH2 Info Block structure                                                */
/*--------------------------------------------------------------------------*/
typedef struct
{
    FolderInfoBlk*                 lpFIBHead;
    FolderInfoBlk*                 lpFIBTail;
    char*                          lpszFormat;
    char*                          lpszRFH2FileName;
    char*                          lpszDir;
    char*                          lpszResRFH2FileName;
    char*                          lpszResDir;
    long                           lRFH2FolderDataLen;
    long                           lRFH2FolderCount;
    long                           bDynamicRIB;
    long                           lEncoding;
    long                           lCodedCharSetId;
    long                           bUseRIBEncodingAndCCSID;
} RFH2InfoBlk;

/*--------------------------------------------------------------------------*/
/* Buffer Info Block structure                                              */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct BuffInfoBlk*            lpBIBNext;
    struct BuffInfoBlk*            lpBIBPrev;
    char*                          lpszBuffName;
    char*                          lpszBuff;
    long                           lBuffLen;
    long                           lDataLen;
} BuffInfoBlk;

/*--------------------------------------------------------------------------*/
/* Interval Info Record structure                                           */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct IntervalInfoRec*        lpIIRNext;
    struct IntervalInfoRec*        lpIIRPrev;
    Int64                          i64IntervalStart;
    Int64                          i64IntervalEnd;
    long                           bHaveStart;
    long                           bHaveEnd;
    long                           bEndBeforeStart;
} IntervalInfoRec;

/*--------------------------------------------------------------------------*/
/* Interval Info Block structure                                            */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct IntervalInfoBlk*        lpIIBNext;
    struct IntervalInfoBlk*        lpIIBPrev;
    char*                          lpszIntervalName;
    IntervalInfoRec*               lpIIRHead;
    IntervalInfoRec*               lpIIRTail;
    long                           lIIRCount;
    FILE*                          lpIntervalFile;
    long                           bHaveIntervalFile;
    long                           bHaveStartCmd;
    long                           bHaveEndCmd;
} IntervalInfoBlk;

/*--------------------------------------------------------------------------*/
/* Message Rate Block structure                                             */
/*--------------------------------------------------------------------------*/
typedef struct
{
    char*                          lpszFrom;
    char*                          lpszTo;
    char*                          lpszIncr;
    char*                          lpszInterval;
    char*                          lpszResFrom;
    char*                          lpszResTo;
    char*                          lpszResIncr;
    char*                          lpszResInterval;
    long                           lFrom;
    long                           lTo;
    long                           lIncr;
    long                           lInterval;
    enum eMsgModeEnum              eMsgRateMode;
    time_t                         oFirstSecond;
    time_t                         oIncrAtSecond;
    time_t                         oProcessingSecond;
    time_t                         oNowSecond;
    long                           lMessagesThisSecond;
    long                           lMessagesLeft;
    long                           bFirstTime;
    long                           bInitForLoop;
    long                           bIncrementalDone;
    Int64                          i64ProcessingStart;
    Int64                          i64ProcessingNow;
    Int64                          i64ProcessingNext;
    Int64                          i64ProcessingIncr;
    Int64                          i64RandomIncr;
} MessageRateBlk;

/*==========================================================================*/
/* Message Set structures                                                   */
/*==========================================================================*/
/*--------------------------------------------------------------------------*/
/* Message Set File Load structure                                          */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct MsgSetFileLoadBlk*      lpMSFLNext;
    struct MsgSetFileLoadBlk*      lpMSFLPrev;
    char*                          lpszFileName;
    char*                          lpszDir;
    char*                          lpszParserName;
    char*                          lpszMsgSetName;
} MsgSetFileLoadBlk;

/*--------------------------------------------------------------------------*/
/* All Message Set linked list elements have the first 5 elements in common */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct MsgSetLinkedListBlk*    lpMSLLNext;
    struct MsgSetLinkedListBlk*    lpMSLLPrev;
    enum eMSLLBlockEnum            eMSLLBlockType;
    char*                          lpszNameOrType;
    long                           lNameOrTypeLen;
} MsgSetLinkedListBlk;

/*--------------------------------------------------------------------------*/
/* Element Definition Block structure                                       */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct ElementBlk*             lpELBNext;
    struct ElementBlk*             lpELBPrev;
    enum eMSLLBlockEnum            eMSLLBlockType;
    char*                          lpszName;
    long                           lNameLen;
    char*                          lpszElementType;
    enum eElementTypeEnum          eElementType;
    char*                          lpszElementSep;
    char*                          lpszLen;
    char*                          lpszPos;
    char*                          lpszOffset;
    char*                          lpszJustify;
    char*                          lpszPadChar;
    char*                          lpszEndAlign;
    char*                          lpszMinOccurs;
    char*                          lpszMaxOccurs;
    char*                          lpszDefaultValue;
    long                           lLen;
    long                           lPos;
    long                           lOffset;
    long                           lEndAlign;
    long                           bCmqValues;
    long                           lMinOccurs;
    long                           lMaxOccurs;
} ElementBlk;

/*--------------------------------------------------------------------------*/
/* Segment Definition Block structure                                       */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct SegmentDefBlk*          lpSDBNext;
    struct SegmentDefBlk*          lpSDBPrev;
    enum eMSLLBlockEnum            eMSLLBlockType;
    char*                          lpszType;
    long                           lTypeLen; 
    char*                          lpszOutputFormat;
    enum eOutputFormatEnum         eOutputFormat;
    char*                          lpszElementSep;
    char*                          lpszSegSep;
    char*                          lpszCCSID;
    char*                          lpszEncoding;
    MsgSetLinkedListBlk*           lpMSLLHead;
    MsgSetLinkedListBlk*           lpMSLLTail;
} SegmentDefBlk;

/*--------------------------------------------------------------------------*/
/* Segment Block structure                                                  */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct SegmentBlk*             lpSGBNext;
    struct SegmentBlk*             lpSGBPrev;
    enum eMSLLBlockEnum            eMSLLBlockType;
    char*                          lpszType;
    long                           lTypeLen;
    char*                          lpszOutputFormat;
    enum eOutputFormatEnum         eOutputFormat;
    char*                          lpszElementSep;
    char*                          lpszSegSep;
    char*                          lpszMinOccurs;
    char*                          lpszMaxOccurs;
    char*                          lpszCCSID;
    char*                          lpszEncoding;
    SegmentDefBlk*                 lpSDB;
    long                           lMinOccurs;
    long                           lMaxOccurs;
} SegmentBlk;

/*--------------------------------------------------------------------------*/
/* Message Defintion Block structure                                        */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct MsgDefBlk*              lpMDBNext;
    struct MsgDefBlk*              lpMDBPrev;
    enum eMSLLBlockEnum            eMSLLBlockType;
    char*                          lpszType;
    long                           lTypeLen;
    char*                          lpszOutputFormat;
    enum eOutputFormatEnum         eOutputFormat;
    char*                          lpszElementSep;
    char*                          lpszSegSep;
    char*                          lpszMsgSep;
    char*                          lpszCCSID;
    char*                          lpszEncoding;
    long                           bTruncate;
    MsgSetLinkedListBlk*           lpMSLLHead;
    MsgSetLinkedListBlk*           lpMSLLTail;
} MsgDefBlk;

/*--------------------------------------------------------------------------*/
/* Message Set Tree Node structure                                          */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct MsgSetTreeNode*         lpParent;
    struct MsgSetTreeNode*         lpNextSibling;
    struct MsgSetTreeNode*         lpPrevSibling;
    struct MsgSetTreeNode*         lpFirstChild;
    struct MsgSetTreeNode*         lpLastChild;
    struct MsgSetTreeNode*         lpNextSiblingSameName;
    struct MsgSetTreeNode*         lpPrevSiblingSameName;
    long                           lIndexSameName;
    enum eMsgSetTreeNodeEnum       eMsgSetTreeNodeType;
    char*                          lpszNodeName;
    long                           lNodeNameLen;
    char*                          lpszNodeValue;
    long                           lNodeValueLen;
    long                           lNodeDepth;
    MsgDefBlk*                     lpMDB;
    SegmentDefBlk*                 lpSDB;
    SegmentBlk*                    lpSGB;
    MsgSetLinkedListBlk*           lpMSLL;
    MsgSetLinkedListBlk*           lpNextRequiredMSLL;
    long                           lFlags;
    long                           lOutputDataLen;
    char*                          lpszBuffLoc;
} MsgSetTreeNode;

/*--------------------------------------------------------------------------*/
/* Message Set Header                                                       */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct MsgSetHdr*              lpMSHNext;
    struct MsgSetHdr*              lpMSHPrev;
    MsgSetTreeNode*                lpMSFRootNode;
    MsgSetTreeNode*                lpOutRootNode;
    MsgDefBlk*                     lpMDBHead;
    MsgDefBlk*                     lpMDBTail;
    SegmentDefBlk*                 lpSDBHead;
    SegmentDefBlk*                 lpSDBTail;
    char*                          lpszMsgSetName;
} MsgSetHdr;

/*==========================================================================*/
/* Command Block structures                                                 */
/*==========================================================================*/
typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
} CommandCtrlBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    char*                          lpszForName;
    VarInfoBlk*                    lpVIB;
    char*                          lpszFrom;
    char*                          lpszTo;
    char*                          lpszIncr;
    char*                          lpszList;
    long                           lFrom;
    long                           lTo;
    long                           lCurrent;
    long                           lIncr;
    ListInfoBlk*                   lpLIB;
    ListElementInfoBlk*            lpCurrLEB;
    enum eForTypeEnum              eForType;
    char*                          lpszResFrom;
    char*                          lpszResTo;
    char*                          lpszResIncr;
    struct ForEndBlk*              lpForEndCmd;
    long                           bCompleteDuringInit;
} ForStartBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    char*                          lpszForName;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    struct ForStartBlk*            lpForStartCmd;
} ForEndBlk;

typedef struct
{
    struct ForBackRefBlk*          lpFBRNext;
    struct ForBackRefBlk*          lpFBRPrev;
    struct ForStartBlk*            lpForStartCmd;
} ForBackRefBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    struct IfStartBlk*             lpIFS;
    struct IfEndBlk*               lpIFE;
    struct IfGenericBlk*           lpIFGNext;
} IfGenericBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    struct IfStartBlk*             lpIFS;
    struct IfEndBlk*               lpIFE;
    struct IfGenericBlk*           lpIFGNext;
    char*                          lpszCond;
    char*                          lpszResCond;
    long                           bGotATrueCondition;
    long                           bACondChecksLastCC;
    long                           bACondChecksLastRC;
} IfStartBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    struct IfStartBlk*             lpIFS;
    struct IfEndBlk*               lpIFE;
    struct IfGenericBlk*           lpIFGNext;
    char*                          lpszCond;
    char*                          lpszResCond;
} ElseIfBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    struct IfStartBlk*             lpIFS;
    struct IfEndBlk*               lpIFE;
    struct IfGenericBlk*           lpIFGNext;
} ElseBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    struct IfStartBlk*             lpIFS;
    struct IfEndBlk*               lpIFE;
    struct IfGenericBlk*           lpIFGNext;
} IfEndBlk;

typedef struct
{
    struct IfBackRefBlk*           lpIFBNext;
    struct IfBackRefBlk*           lpIFBPrev;
    struct IfStartBlk*             lpIfStartCmd;
} IfBackRefBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    ForStartBlk*                   lpBreakContinueFSB;
} ContinueBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    ForStartBlk*                   lpBreakContinueFSB;
} BreakBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
} ExitTestBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
} FailTestBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
} FinallyStartBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    char*                          lpszFinallyName;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
} FinallyEndBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    BuffInfoBlk*                   lpBIB;
    ForStartBlk*                   lpBreakContinueFSB;
    char*                          lpszMQMDVarName;
    char*                          lpszRFH2VarName;
    char*                          lpszBuffName;
    char*                          lpszFileName;
    char*                          lpszMQMDFileName;
    char*                          lpszRFH2FileName;
    char*                          lpszDir;
    char*                          lpszCmdMaxWait;
    char*                          lpszResMQMDVarName;
    char*                          lpszResRFH2VarName;
    char*                          lpszResBuffName;
    char*                          lpszResFileName;
    char*                          lpszResMQMDFileName;
    char*                          lpszResRFH2FileName;
    char*                          lpszResDir;
    char*                          lpszResCmdMaxWait;
    long                           lMaxWait;
    long                           lSuppressRCCount;
    long                           lSuppressRC[MAX_REASON_CODES];
    long                           lContinueOnRCCount;
    long                           lContinueOnRC[MAX_REASON_CODES];
    long                           lBreakOnRCCount;
    long                           lBreakOnRC[MAX_REASON_CODES];
} GetFileBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    BuffInfoBlk*                   lpBIB;
    BuffInfoBlk*                   lpTempDataVarBIB;
    ForStartBlk*                   lpBreakContinueFSB;
    char*                          lpszDataVarName;
    char*                          lpszMQMDVarName;
    char*                          lpszRFH2VarName;
    char*                          lpszBuffName;
    char*                          lpszFileName;
    char*                          lpszMQMDFileName;
    char*                          lpszRFH2FileName;
    char*                          lpszDir;
    char*                          lpszResDataVarName;
    char*                          lpszResMQMDVarName;
    char*                          lpszResRFH2VarName;
    char*                          lpszResBuffName;
    char*                          lpszResFileName;
    char*                          lpszResMQMDFileName;
    char*                          lpszResRFH2FileName;
    char*                          lpszResDir;
    long                           bAppend;
    long                           lSuppressRCCount;
    long                           lSuppressRC[MAX_REASON_CODES];
    long                           lContinueOnRCCount;
    long                           lContinueOnRC[MAX_REASON_CODES];
    long                           lBreakOnRCCount;
    long                           lBreakOnRC[MAX_REASON_CODES];
} PutFileBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    BuffInfoBlk*                   lpBIB;
    char*                          lpszBuffName;
    char*                          lpszData;
    char*                          lpszResBuffName;
} InlineDataBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    BuffInfoBlk*                   lpBIB;
    char*                          lpszPos;
    char*                          lpszLen;
    char*                          lpszBuffName;
    char*                          lpszData;
    char*                          lpszResBuffName;
    char*                          lpszResData;
    char*                          lpszResPos;
    char*                          lpszResLen;
    long                           lPos;
    long                           lLen;
} OverlayBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    BuffInfoBlk*                   lpBIB;
    VarInfoBlk*                    lpVIB;
    char*                          lpszVarName;
    char*                          lpszPos;
    char*                          lpszLen;
    char*                          lpszBuffName;
    char*                          lpszResVarName;
    char*                          lpszResBuffName;
    char*                          lpszResPos;
    char*                          lpszResLen;
    long                           lPos;
    long                           lLen;
} ExtractBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    VarInfoBlk*                    lpVIB;
    VarInfoBlk*                    lpVIBFrom;
    char*                          lpszVarName;
    char*                          lpszFromVarName;
    char*                          lpszPos;
    char*                          lpszLen;
    char*                          lpszResVarName;
    char*                          lpszResFromVarName;
    char*                          lpszResPos;
    char*                          lpszResLen;
    long                           lPos;
    long                           lLen;
} SubstringBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    BuffInfoBlk*                   lpBIB;
    QueueInfoBlk*                  lpQIB;
    ConnectionDefBlk*              lpCDB;
    ForStartBlk*                   lpBreakContinueFSB;
    IntervalInfoBlk*               lpIIB;
    char*                          lpszMQMDVarName;
    char*                          lpszRFH2VarName;
    char*                          lpszConnectionName;
    char*                          lpszBuffName;
    char*                          lpszQName;
    char*                          lpszCmdMaxWait;
    char*                          lpszGmoOptions;
    char*                          lpszGmoMatchOptions;
    char*                          lpszGmoWait;
    char*                          lpszMsgId;
    char*                          lpszCorrelId;
    char*                          lpszGroupId;
    char*                          lpszMsgSeqNumber;
    char*                          lpszResMQMDVarName;
    char*                          lpszResRFH2VarName;
    char*                          lpszResConnectionName;
    char*                          lpszResBuffName;
    char*                          lpszResQName;
    char*                          lpszResCmdMaxWait;
    char*                          lpszResGmoWait;
    char*                          lpszResMsgId;
    char*                          lpszResCorrelId;
    char*                          lpszResGroupId;
    char*                          lpszResMsgSeqNumber;
    long                           lMaxWait;
    long                           lGmoWait;
    long                           lSuppressRCCount;
    long                           lSuppressRC[MAX_REASON_CODES];
    long                           lContinueOnRCCount;
    long                           lContinueOnRC[MAX_REASON_CODES];
    long                           lBreakOnRCCount;
    long                           lBreakOnRC[MAX_REASON_CODES];
    long                           bRunDoneOnce;
    long                           bConvert;
    long                           bIntervalInMsgId;
    long                           bIntervalInCorrelId;
    long                           bSyncpoint;
    long                           bRemoveDLH;
    long                           bRemoveXQH;
    long                           bGmoWait;
    enum eBrowseEnum               eBrowseType;
} GetMsgBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    BuffInfoBlk*                   lpBIB;
    BuffInfoBlk*                   lpTempDataVarBIB;
    QueueInfoBlk*                  lpQIB;
    ConnectionDefBlk*              lpCDB;
    ForStartBlk*                   lpBreakContinueFSB;
    IntervalInfoBlk*               lpIIB;
    MessageRateBlk*                lpMRB;
    char*                          lpszDataVarName;
    char*                          lpszMQMDVarName;
    char*                          lpszRFH2VarName;
    char*                          lpszConnectionName;
    char*                          lpszBuffName;
    char*                          lpszQName;
    char*                          lpszQMgrName;
    char*                          lpszPmoOptions;
    char*                          lpszMsgId;
    char*                          lpszCorrelId;
    char*                          lpszGroupId;
    char*                          lpszMsgSeqNumber;
    char*                          lpszFormat;
    char*                          lpszResDataVarName;
    char*                          lpszResMQMDVarName;
    char*                          lpszResRFH2VarName;
    char*                          lpszResConnectionName;
    char*                          lpszResBuffName;
    char*                          lpszResQName;
    char*                          lpszResQMgrName;
    char*                          lpszResMsgId;
    char*                          lpszResCorrelId;
    char*                          lpszResGroupId;
    char*                          lpszResMsgSeqNumber;
    char*                          lpszResFormat;
    long                           lSuppressRCCount;
    long                           lSuppressRC[MAX_REASON_CODES];
    long                           lContinueOnRCCount;
    long                           lContinueOnRC[MAX_REASON_CODES];
    long                           lBreakOnRCCount;
    long                           lBreakOnRC[MAX_REASON_CODES];
    long                           bSuppressRFH2;
    long                           bIntervalInMsgId;
    long                           bIntervalInCorrelId;
    long                           bSyncpoint;
    long                           bSetIdentityContext;
    long                           bSetAllContext;
    enum eGroupEnum                eGroupType;
} PutMsgBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    BuffInfoBlk*                   lpBIB;
    MsgSetHdr*                     lpMsgSetHdr;
    ForStartBlk*                   lpBreakContinueFSB;
    char*                          lpszMQMDVarName;
    char*                          lpszBuffName;
    char*                          lpszMsgSetName;
    char*                          lpszMsgNum;
    char*                          lpszResMQMDVarName;
    char*                          lpszResBuffName;
    char*                          lpszResMsgSetName;
    char*                          lpszResMsgNum;
    long                           lMsgNum;
    MsgSetTreeNode*                lpLastMsgSetTreeNode;
    long                           lSuppressRCCount;
    long                           lSuppressRC[MAX_REASON_CODES];
    long                           lContinueOnRCCount;
    long                           lContinueOnRC[MAX_REASON_CODES];
    long                           lBreakOnRCCount;
    long                           lBreakOnRC[MAX_REASON_CODES];
} SelectBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    QueueInfoBlk*                  lpQIB;
    ConnectionDefBlk*              lpCDB;
    char*                          lpszConnectionName;
    char*                          lpszQName;
    char*                          lpszDepth;
    char*                          lpszCmdMaxWait;
    char*                          lpszResConnectionName;
    char*                          lpszResQName;
    char*                          lpszResDepth;
    char*                          lpszResCmdMaxWait;
    long                           lDepth;
    long                           lMaxWait;
} WaitDepthBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    enum eWaitTypeEnum             eWaitType;
    char*                          lpszInterval;
    char*                          lpszIntervalMillis;
    char*                          lpszTime;
    char*                          lpszResInterval;
    char*                          lpszResIntervalMillis;
    char*                          lpszResTime;
    long                           lInterval;
    long                           lIntervalMillis;
} WaitTimeBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    BuffInfoBlk*                   lpBIB;
    char*                          lpszBuffName;
    char*                          lpszCompFileName;
    char*                          lpszCompMaskFileName;
    char*                          lpszCompRegexFileName;
    char*                          lpszDir;
    char*                          lpszCompRFH2VarName1;
    char*                          lpszCompRFH2VarName2;
    char*                          lpszResBuffName;
    char*                          lpszResCompFileName;
    char*                          lpszResCompMaskFileName;
    char*                          lpszResCompRegexFileName;
    char*                          lpszResDir;
    char*                          lpszCompFileBuff;
    long                           lCompFileBuffLen;
    long                           lCompFileDataLen;
    char*                          lpszCompMaskBuff;
    long                           lCompMaskBuffLen;
    long                           lCompMaskDataLen;
    char*                          lpszCompRegexBuff;
    long                           lCompRegexBuffLen;
    long                           lCompRegexDataLen;
    long                           bGotCompFile;
    long                           bGotCompMask;
    long                           bGotCompRegex;
    long                           bLogAndContinue;
    long                           bCallOutMade;
    long                           lSubjectCurrPos;
    long                           lPatternCurrPos;
    long                           lPatternCurrItemLen;
    pcre_extra                     oPcreExtra;
    pcre*                          lpCompiledRegex;
} CompareBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    ForStartBlk*                   lpBreakContinueFSB;
    ConnectionDefBlk*              lpCDB;
    char*                          lpszConnectionName;
    char*                          lpszResConnectionName;
    long                           lSuppressRCCount;
    long                           lSuppressRC[MAX_REASON_CODES];
    long                           lContinueOnRCCount;
    long                           lContinueOnRC[MAX_REASON_CODES];
    long                           lBreakOnRCCount;
    long                           lBreakOnRC[MAX_REASON_CODES];
} CommitBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    ForStartBlk*                   lpBreakContinueFSB;
    ConnectionDefBlk*              lpCDB;
    char*                          lpszConnectionName;
    char*                          lpszResConnectionName;
    long                           lSuppressRCCount;
    long                           lSuppressRC[MAX_REASON_CODES];
    long                           lContinueOnRCCount;
    long                           lContinueOnRC[MAX_REASON_CODES];
    long                           lBreakOnRCCount;
    long                           lBreakOnRC[MAX_REASON_CODES];
} BackoutBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    char*                          lpszCommand;
    long                           bIgnoreRC;
    char*                          lpszResCommand;
} SystemBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    BuffInfoBlk*                   lpBIB;
    QueueInfoBlk*                  lpQIB;
    QueueInfoBlk*                  lpQIBMR;
    ForStartBlk*                   lpBreakContinueFSB;
    ConnectionDefBlk*              lpCDB;
    char*                          lpszConnectionName;
    char*                          lpszBuffName;
    char*                          lpszQName;
    char*                          lpszMqscReplyQName;
    char*                          lpszMqscReplyDynamicQName;
    char*                          lpszCommand;
    char*                          lpszCmdMaxWait;
    char*                          lpszResConnectionName;
    char*                          lpszResBuffName;
    char*                          lpszResQName;
    char*                          lpszResMqscReplyDynamicQName;
    char*                          lpszResCommand;
    char*                          lpszResCmdMaxWait;
    long                           lMaxWait;
    long                           lSuppressRCCount;
    long                           lSuppressRC[MAX_REASON_CODES];
    long                           lContinueOnRCCount;
    long                           lContinueOnRC[MAX_REASON_CODES];
    long                           lBreakOnRCCount;
    long                           lBreakOnRC[MAX_REASON_CODES];
} MqscCommandBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    IntervalInfoBlk*               lpIIB;
} IntervalStartBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    IntervalInfoBlk*               lpIIB;
} IntervalEndBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    enum eVariableEnum             eVarType;
    char*                          lpszVarName;
    char*                          lpszVarFormat;
    char*                          lpszNewStrValue;
    char*                          lpszRandomNumFrom;
    char*                          lpszRandomNumTo;
    char*                          lpszResRandomNumFrom;
    char*                          lpszResRandomNumTo;
    long                           lRandomNumFrom;
    long                           lRandomNumTo;
    long                           bRandomNumSeeded;
    long                           bClear;
    void*                          lpNewCtrlBlk;
    char*                          lpszResVarName;
    VarInfoBlk*                    lpVIB;
} VarBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    char*                          lpszVarName;
    char*                          lpszResVarName;
    VarInfoBlk*                    lpVIB;
} StdInBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    char*                          lpszTextOut;
    char*                          lpszResTextOut;
} StdOutBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    char*                          lpszTextOut;
    char*                          lpszResTextOut;
} UserLogBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    enum eArithmeticEnum           eMathFunctionType;
    char*                          lpszVarName;
    char*                          lpszRemName;
    char*                          lpszValue;
    char*                          lpszResVarName;
    char*                          lpszResRemName;
    char*                          lpszResValue;
    VarInfoBlk*                    lpVIB;
    VarInfoBlk*                    lpVIBRem;
} MathFunctionBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    char*                          lpszVarName;
    char*                          lpszOf;
    char*                          lpszResVarName;
    char*                          lpszResOf;
} LengthBlk;

typedef struct
{
    struct CommandCtrlBlk*         lpCCBNext;
    struct CommandCtrlBlk*         lpCCBPrev;
    enum eCommandEnum              eCommandType;
    enum eCmdStateEnum             eCommandState;
    time_t                         oStartTime;
    time_t                         oExpiryTime;
    time_t                         oNextLogMsgTime;
    char*                          lpszVarName;
    char*                          lpszOf;
    char*                          lpszIn;
    char*                          lpszFrom;
    char*                          lpszResVarName;
    char*                          lpszResOf;
    char*                          lpszResIn;
    char*                          lpszResFrom;
    long                           lFrom;
} PositionBlk;

/*--------------------------------------------------------------------------*/
/* Test Info Block structure                                                */
/*--------------------------------------------------------------------------*/
typedef struct
{
    struct TestInfoBlk*            lpTIBNext;
    struct TestInfoBlk*            lpTIBPrev;
    CommandCtrlBlk*                lpCCBHead;
    CommandCtrlBlk*                lpCCBTail;
    CommandCtrlBlk*                lpCCBCurr;
    VarInfoBlk*                    lpVIBHead;
    VarInfoBlk*                    lpVIBTail;
    BuffInfoBlk*                   lpBIBHead;
    BuffInfoBlk*                   lpBIBTail;
    QueueInfoBlk*                  lpQIBHead;
    QueueInfoBlk*                  lpQIBTail;
    char*                          lpszTestName;
    long                           lTestNum;
    long                           bCmdResult;
    long                           bFinallyResetFailedTest;
    long                           lLastCC;
    long                           lLastRC;
    long                           lLastDataLen;
    long                           lTestCurrentMsgRate;
    long                           lTestTotalGetFile;
    long                           lTestSuccessfulGetFile;
    long                           lTestFailedGetFile;
    long                           lTestTotalPutFile;
    long                           lTestSuccessfulPutFile;
    long                           lTestFailedPutFile;
    long                           lTestTotalGetMsg;
    long                           lTestSuccessfulGetMsg;
    long                           lTestFailedGetMsg;
    long                           lTestTotalPutMsg;
    long                           lTestSuccessfulPutMsg;
    long                           lTestFailedPutMsg;
    long                           lTestTotalCompare;
    long                           lTestSuccessfulCompare;
    long                           lTestFailedCompare;
    long                           bTestHasGetFile;
    long                           bTestHasPutFile;
    long                           bTestHasGetMsg;
    long                           bTestHasPutMsg;
    long                           bTestHasCompare;
    char*                          lpszLastFile;
    char*                          lpszLastDir;
    char*                          lpszLastQ;
    char*                          lpszLastQMgr;
    ForStartBlk*                   lpBreakContinueFSB;
    Int64                          i64CommandInit;
} TestInfoBlk;

/*--------------------------------------------------------------------------*/
/* Control Info Block structure                                             */
/*--------------------------------------------------------------------------*/
typedef struct
{
    char*                          lpszConnQMgrName;
    char*                          lpszConnChannelName;
    char*                          lpszConnHostName;
    char*                          lpszConnPortNum;
    char*                          lpszConnKeyRepository;
    char*                          lpszConnSSLCipherSpec;
    char*                          lpszConnSSLPeerName;
    char*                          lpszConnMaxMsgLength;
    char*                          lpszTestLogFileName;
    char*                          lpszTestLogDir;
    long                           bTestLogAppend;
    char*                          lpszTestReportFileName;
    char*                          lpszTestReportDir;
    long                           bTestReportAppend;
    char*                          lpszTestUserLogFileName;
    char*                          lpszTestUserLogDir;
    long                           bTestUserLogAppend;
    char*                          lpszTestIntervalFilePrefix;
    char*                          lpszTestIntervalDelimiter;
    char*                          lpszTestIntervalFileExtension;
    char*                          lpszTestIntervalDir;
    long                           bTestIntervalAppend;
    char*                          lpszFilePrefix;
    char*                          lpszParserName;
    MsgSetFileLoadBlk*             lpMSFLHead;
    MsgSetFileLoadBlk*             lpMSFLTail;
    ConnectionDefBlk*              lpCurrCDB;
    long                           lTestMaxWait;
    long                           lNextLogMsgInterval;
    long                           lConcurrency;
    long                           lConcurrencyWait;
    char*                          lpszOnError;
    char*                          lpszConvTableDir;
    char*                          lpszLocalEncoding;
    char*                          lpszLocalCodedCharSetId;
    long                           lLocalEncoding;
    long                           lLocalCodedCharSetId;
    FILE*                          lpTestLogFile;
    FILE*                          lpTestReportFile;
    FILE*                          lpTestUserLogFile;
} ControlInfoBlk;

/*--------------------------------------------------------------------------*/
/* Defaults Info Block structure                                            */
/*--------------------------------------------------------------------------*/
typedef struct
{
    VarInfoBlk*                    lpVIBHead;
    VarInfoBlk*                    lpVIBTail;
    GetFileBlk*                    lpGFB;
    PutFileBlk*                    lpPFB;
    InlineDataBlk*                 lpIDB;
    OverlayBlk*                    lpOVB;
    ExtractBlk*                    lpEXB;
    SubstringBlk*                  lpSSB;
    GetMsgBlk*                     lpGMB;
    PutMsgBlk*                     lpPMB;
    SelectBlk*                     lpSLB;
    WaitDepthBlk*                  lpWDB;
    WaitTimeBlk*                   lpWTB;
    CompareBlk*                    lpCMB;
    CommitBlk*                     lpCOB;
    BackoutBlk*                    lpBOB;
    SystemBlk*                     lpSYB;
} DefaultsInfoBlk;

/*--------------------------------------------------------------------------*/
/* Command sub-elements                                                     */
/*--------------------------------------------------------------------------*/
typedef struct
{
    char*                          lpszBuffName;
    char*                          lpszFileName;
    char*                          lpszListFileName;
    char*                          lpszRegexFileName;
    char*                          lpszMQMDFileName;
    char*                          lpszRFH2FileName;
    char*                          lpszDir;
    char*                          lpszData;
    char*                          lpszQName;
    char*                          lpszMqscReplyDynamicQName;
    char*                          lpszQMgrName;
    char*                          lpszDepth;
    char*                          lpszInterval;
    char*                          lpszIntervalMillis;
    char*                          lpszTime;
    char*                          lpszCommand;
    char*                          lpszMaskFileName;
    char*                          lpszCmdMaxWait;
    char*                          lpszGmoOptions;
    char*                          lpszPmoOptions;
    char*                          lpszGmoMatchOptions;
    char*                          lpszGmoWait;
    char*                          lpszMsgId;
    char*                          lpszCorrelId;
    char*                          lpszGroupId;
    char*                          lpszMsgSeqNumber;
    char*                          lpszFormat;
    char*                          lpszCompRFH2VarName1;
    char*                          lpszCompRFH2VarName2;
    long                           bIgnoreRC;
    long                           bConvert;
    long                           bClear;
    long                           bLogAndContinue;
    long                           bIntervalInMsgId;
    long                           bIntervalInCorrelId;
    long                           bSyncpoint;
    long                           bSetIdentityContext;
    long                           bSetAllContext;
    long                           bRemoveDLH;
    long                           bRemoveXQH;
    long                           bAppend;
    enum eBrowseEnum               eBrowseType;
    enum eGroupEnum                eGroupType;
    IntervalInfoBlk*               lpIIB;
    MessageRateBlk*                lpMRB;
} CSE;

/*--------------------------------------------------------------------------*/
/* Secondary Parse Block structure                                          */
/*--------------------------------------------------------------------------*/
typedef struct
{
    char*                          lpszCtrlBlkFileName;
    char*                          lpszCtrlBlkBuff;
    long                           lCtrlBlkXmlLen;
    long                           lCtrlBlkParseDepth;
    char*                          lplpszElementName[MAX_CTRLBLK_PARSE_DEPTH + 1];
    enum eElementEnum              eElementType[MAX_CTRLBLK_PARSE_DEPTH + 1];
} SecondaryParseBlk;

/*--------------------------------------------------------------------------*/
/* Message Set Parse Block structure                                        */
/*--------------------------------------------------------------------------*/
typedef struct
{
    char*                          lpszMsgSetBuff;
    long                           lMsgSetDataLen;
    long                           lMsgSetParseDepth;
    ElementBlk*                    lpCurrELB;
    SegmentDefBlk*                 lpCurrSDB;
    SegmentBlk*                    lpCurrSGB;
    MsgDefBlk*                     lpCurrMDB;
    MsgSetTreeNode*                lpCurrMSTN;
    MsgSetHdr*                     lpCurrMSH;
    char*                          lplpszElementName[MAX_MSGSET_PARSE_DEPTH + 1];
    enum eMSLLBlockEnum            eMsgSetType[MAX_MSGSET_PARSE_DEPTH + 1];
} MsgSetParseBlk;

/*--------------------------------------------------------------------------*/
/* Main Control Block structure                                             */
/*--------------------------------------------------------------------------*/
typedef struct
{
    TestInfoBlk*                   lpTIBReadyHead;
    TestInfoBlk*                   lpTIBReadyTail;
    TestInfoBlk*                   lpTIBActiveHead;
    TestInfoBlk*                   lpTIBActiveTail;
    VarInfoBlk*                    lpVIBHead;
    VarInfoBlk*                    lpVIBTail;
    ForBackRefBlk*                 lpFBRHead;
    ForBackRefBlk*                 lpFBRTail;
    IfBackRefBlk*                  lpIFBHead;
    IfBackRefBlk*                  lpIFBTail;
    IntervalInfoBlk*               lpIIBHead;
    IntervalInfoBlk*               lpIIBTail;
    ControlInfoBlk*                lpCIB;
    DefaultsInfoBlk*               lpDIB;
    ConnectionDefBlk*              lpCDBHead;
    ConnectionDefBlk*              lpCDBTail;
    MsgSetHdr*                     lpMSHHead;
    MsgSetHdr*                     lpMSHTail;
    MsgSetLinkedListBlk            oDummyRootMSLL;
    char*                          lpszScriptFileName;
    char*                          lpszScriptBuff;
    char*                          lpszScriptName;
    long                           lScriptXmlLen;
    long                           lProcessing;
    long                           lTotalTests;
    long                           lSuccessfulTests;
    long                           lFailedTests;
    long                           lIncompleteTests;
    long                           lScriptTotalGetFile;
    long                           lScriptSuccessfulGetFile;
    long                           lScriptFailedGetFile;
    long                           lScriptTotalPutFile;
    long                           lScriptSuccessfulPutFile;
    long                           lScriptFailedPutFile;
    long                           lScriptTotalGetMsg;
    long                           lScriptSuccessfulGetMsg;
    long                           lScriptFailedGetMsg;
    long                           lScriptTotalPutMsg;
    long                           lScriptSuccessfulPutMsg;
    long                           lScriptFailedPutMsg;
    long                           lScriptTotalCompare;
    long                           lScriptSuccessfulCompare;
    long                           lScriptFailedCompare;
    long                           bScriptHasGetFile;
    long                           bScriptHasPutFile;
    long                           bScriptHasGetMsg;
    long                           bScriptHasPutMsg;
    long                           bScriptHasCompare;
    long                           bScriptHasMqscCommand;
    long                           bConcurrencyWait;
    long                           bGmoWait;
    long                           bMessageRate;
    long                           bConnectionRequired;
    long                           lWaitTotal;
    long                           lScriptParseDepth;
    long                           lFileErrno;
    long                           lSuppressRCCount;
    long                           lSuppressRC[MAX_REASON_CODES];
    long                           lContinueOnRCCount;
    long                           lContinueOnRC[MAX_REASON_CODES];
    long                           lBreakOnRCCount;
    long                           lBreakOnRC[MAX_REASON_CODES];
    char*                          lplpszElementName[MAX_SCRIPT_PARSE_DEPTH + 1];
    enum eElementEnum              eElementType[MAX_SCRIPT_PARSE_DEPTH + 1];
    long                           lMsgSetSegmentLen[MAX_MSGSET_PARSE_DEPTH + 1];
    char*                          lpszCurrNameAttr;
    char*                          lpszCurrFormatAttr;
    char*                          lpszCurrOutputFormatAttr;
    char*                          lpszCurrMQMDAttr;
    char*                          lpszCurrRFH2Attr;
    char*                          lpszCurrR1Attr;
    char*                          lpszCurrR2Attr;
    char*                          lpszCurrFromAttr;
    char*                          lpszCurrToAttr;
    char*                          lpszCurrIncrAttr;
    char*                          lpszCurrCondAttr;
    char*                          lpszCurrModeAttr;
    char*                          lpszCurrIntervalAttr;
    char*                          lpszCurrPosAttr;
    char*                          lpszCurrLenAttr;
    char*                          lpszCurrTypeAttr;
    char*                          lpszCurrMsgSepAttr;
    char*                          lpszCurrSegSepAttr;
    char*                          lpszCurrElementSepAttr;
    char*                          lpszCurrMinOccursAttr;
    char*                          lpszCurrMaxOccursAttr;
    char*                          lpszCurrCCSIDAttr;
    char*                          lpszCurrEncodingAttr;
    char*                          lpszCurrMsgSetNameAttr;
    char*                          lpszCurrMsgNumAttr;
    char*                          lpszCurrOffsetAttr;
    char*                          lpszCurrJustifyAttr;
    char*                          lpszCurrPadCharAttr;
    char*                          lpszCurrEndAlignAttr;
    char*                          lpszCurrCmqValuesAttr;
    char*                          lpszCurrTruncateAttr;
    char*                          lpszCurrListAttr;
    char*                          lpszCurrDataVarAttr;
    char*                          lpszCurrConnectionAttr;
    char*                          lpszCurrOfAttr;
    char*                          lpszCurrInAttr;
    char*                          lpszCurrDelimiterAttr;
    char*                          lpszCurrFileExtensionAttr;
    char*                          lpszCurrRemAttr;
    char*                          lpszErrorWorkBuff;
    char*                          lpszMqscReplyBuff;
    long                           lMqscReplyBuffLen;
    long                           bInitComplete;
    enum eForTypeEnum              eForType;
    long                           bLoggingSuppressed;
    long                           bReportMessageRate;
    long                           lMessageRateAchieved;
    long                           lMessageRateMissed;
    long                           lParseCharDataLen;
    char*                          lpszParseCharDataBuff;
    SecondaryParseBlk*             lpSPB;
    MsgSetParseBlk*                lpMSPB;
    ListInfoBlk*                   lpLIB;
    ListInfoBlk*                   lpNullLIB;
    MQMDInfoBlk*                   lpMIB;
    MQMDInfoBlk*                   lpNullMIB;
    RFH2InfoBlk*                   lpRIB;
    RFH2InfoBlk*                   lpNullRIB;
    CSE*                           lpCSE;
    MQLONG                         lLocalIntegerEncoding;
    Int64                          i64CountFrequency;
    Int64                          i64OneHalfMilliSecond;
    Int64                          i64RunStart;
    time_t                         oRunStart;
    char                           lpszTimeBuff[TIME_BUFF_LEN + 4];
} MainCtrlBlk;

extern MainCtrlBlk*                kpMCB;

/*==========================================================================*/
/* Functions                                                                */
/*==========================================================================*/
/*--------------------------------------------------------------------------*/
/* Initialisation                                                           */
/*--------------------------------------------------------------------------*/
long
InitCBs(int argc, char** argv);

long
OpenResources();

long
ResolveCIBVars();

long
PrepareTests();

/*--------------------------------------------------------------------------*/
/* Cmq header file value translation                                        */
/*--------------------------------------------------------------------------*/
long
ResolveCharCmqValue(TestInfoBlk* lpTIB, BuffInfoBlk* lpBIB,
                    const char* const kpszCmqValue,
                    long lOutputDataLen,
                    long bCmqValues);

long
ResolveHexCmqValue(TestInfoBlk* lpTIB, BuffInfoBlk* lpBIB,
                   const char* const kpszCmqValue,
                   long lOutputDataLen,
                   long bCmqValues);

long
ResolveIntCmqValue(TestInfoBlk* lpTIB, BuffInfoBlk* lpBIB,
                   const char* const kpszCmqValue,
                   long lOutputDataLen,
                   long bCmqValues);

void
TranslateToValue(TestInfoBlk* lpTIB, void* vpResult,
                 void* vpInitialValue, const char* const kpszCmqValue,
                 enum eCmqValueTypeEnum eCmqValueType);

long
ExtractMQToken(char** lplpszStartToken, char** lplpszEndToken, 
               char* lpszMaxPos, long lMinLen);

void
TurnOnBits(long* lplLongResult, long lOrWith);

void
TurnOffBits(long* lplLongResult, long lOrWith);


/*--------------------------------------------------------------------------*/
/* Script parsing and program control block generation                      */
/*--------------------------------------------------------------------------*/
long
ParseScriptFile();

static void XMLCALL
PSFParseStartElement(void* vpUserData, const char* kpszElementName, const char** kpkpAtts);

static void XMLCALL
PSFParseEndElement(void* vpUserData, const char* kpszElementName);

static void XMLCALL
PSFParseStartCData(void* vpUserData);

static void XMLCALL 
PSFParseEndCData(void* vpUserData);

static void XMLCALL
PSFParseCharData(void* vpUserData, const XML_Char* kpszCD, int lLen);

static void XMLCALL
PSFComments(void* vpUserData, const XML_Char* kpszCD);

void
PSFDisplayParseError(long lMsgNum);

enum eElementEnum
PSFGetElementType(const char* kpszElementName);

GetFileBlk*
CreateGetFileBlk();

PutFileBlk*
CreatePutFileBlk();

InlineDataBlk*
CreateInlineDataBlk();

OverlayBlk*
CreateOverlayBlk();

ExtractBlk*
CreateExtractBlk();

SubstringBlk*
CreateSubstringBlk();

GetMsgBlk*
CreateGetMsgBlk();

PutMsgBlk*
CreatePutMsgBlk();

SelectBlk*
CreateSelectBlk();

WaitDepthBlk*
CreateWaitDepthBlk();

WaitTimeBlk*
CreateWaitTimeBlk();

CompareBlk*
CreateCompareBlk();

CommitBlk*
CreateCommitBlk();

BackoutBlk*
CreateBackoutBlk();

SystemBlk*
CreateSystemBlk();

MqscCommandBlk*
CreateMqscCommandBlk();

ForStartBlk*
CreateForStartBlk();

ForEndBlk*
CreateForEndBlk();

IfStartBlk*
CreateIfStartBlk();

ElseIfBlk*
CreateElseIfBlk();

ElseBlk*
CreateElseBlk();

IfEndBlk*
CreateIfEndBlk();

ContinueBlk*
CreateContinueBlk();

BreakBlk*
CreateBreakBlk();

ExitTestBlk*
CreateExitTestBlk();

FailTestBlk*
CreateFailTestBlk();

FinallyStartBlk*
CreateFinallyStartBlk();

FinallyEndBlk*
CreateFinallyEndBlk();

IntervalStartBlk*
CreateIntervalStartBlk();

IntervalEndBlk*
CreateIntervalEndBlk();

MessageRateBlk*
CreateMessageRateBlk();

VarBlk*
CreateVarBlk(enum eVariableEnum eVarType);

VarInfoBlk*
CreateVarInfoBlk(char* lpszVarName);

IntervalInfoBlk*
CreateIntervalInfoBlk(char* lpszIntervalName);

FolderInfoBlk*
CreateFolderInfoBlk(TestInfoBlk* lpTIB, long bPad);

ListElementInfoBlk*
CreateListElementInfoBlk(const char* const kpszListElement);

ListInfoBlk*
CreateListInfoBlk(long bDynamicLIB);

MQMDInfoBlk*
CreateMQMDInfoBlk(long bDynamicMIB);

RFH2InfoBlk*
CreateRFH2InfoBlk(long bDynamicRIB, MQRFH2* lpoRFH2);

MsgSetFileLoadBlk*
CreateAndAddMsgSetFileLoadBlk();

StdInBlk*
CreateStdInBlk();

StdOutBlk*
CreateStdOutBlk();

UserLogBlk*
CreateUserLogBlk();

MathFunctionBlk*
CreateMathFunctionBlk(enum eArithmeticEnum eMathFunction);

LengthBlk*
CreateLengthBlk();

PositionBlk*
CreatePositionBlk();

ConnectionDefBlk*
CreateConnectionDefBlk();

void
SetCDBToCIBValues(ConnectionDefBlk* lpCDB);

TestInfoBlk*
CreateTestInfoBlk();

/*--------------------------------------------------------------------------*/
/* Control block parsing and support                                        */
/*--------------------------------------------------------------------------*/
long
ParseCtrlBlk(TestInfoBlk* lpTIB, char* lpszCtrlBlkType);

static void XMLCALL
PCBParseStartElement(void* vpUserData, const char* kpszElementName,
                     const char** kpkpAtts);

static void XMLCALL
PCBParseEndElement(void* vpUserData, const char* kpszElementName);

static void XMLCALL
PCBParseStartCData(void* vpUserData);

static void XMLCALL 
PCBParseEndCData(void* vpUserData);

static void XMLCALL
PCBParseCharData(void* vpUserData, const XML_Char* kpszCD, int lLen);

static void XMLCALL
PCBComments(void* vpUserData, const XML_Char* kpszCD);

void
PCBDisplayParseError(long lMsgNum);

enum eElementEnum
PCBGetElementType(const char* kpszElementName);

/*--------------------------------------------------------------------------*/
/* Message Set parsing and support                                          */
/*--------------------------------------------------------------------------*/
long
LoadAndParseMsgSetFiles();

long
ParseMsgSet();

static void XMLCALL
PMSParseStartElement(void* vpUserData, const char* kpszElementName,
                     const char** kpkpAtts);

static void XMLCALL
PMSParseEndElement(void* vpUserData, const char* kpszElementName);

static void XMLCALL
PMSParseStartCData(void* vpUserData);

static void XMLCALL 
PMSParseEndCData(void* vpUserData);

static void XMLCALL
PMSParseCharData(void* vpUserData, const XML_Char* kpszCD, int lLen);

static void XMLCALL
PMSComments(void* vpUserData, const XML_Char* kpszCD);

void
PMSDisplayParseError(long lMsgNum);

enum eMSLLBlockEnum
PMSGetElementType(const char* kpszElementName);

ElementBlk*
CreateElementBlk();

SegmentDefBlk*
CreateSegmentDefBlk();

SegmentBlk*
CreateSegmentBlk();

MsgDefBlk*
CreateMsgDefBlk();

MsgSetTreeNode*
CreateMsgSetTreeNode(enum eMsgSetTreeNodeEnum eMsgSetTreeNodeType,
                     const char* const kpszNodeName, const char* const kpszNodeValue,
                     MsgSetLinkedListBlk* lpMSLL);

MsgSetTreeNode*
AddLastChild(MsgSetTreeNode* lpParent,
             enum eMsgSetTreeNodeEnum eMsgSetTreeNodeType,
             const char* const kpszNodeName,
             const char* const kpszNodeValue,
             MsgSetLinkedListBlk* lpMSLL);

MsgSetTreeNode*
AddFirstChild(MsgSetTreeNode* lpParent,
              enum eMsgSetTreeNodeEnum eMsgSetTreeNodeType,
              const char* const kpszNodeName,
              const char* const kpszNodeValue,
              MsgSetLinkedListBlk* lpMSLL);

MsgSetTreeNode*
AddPrevSibling(MsgSetTreeNode* lpTNCurr,
              enum eMsgSetTreeNodeEnum eMsgSetTreeNodeType,
              const char* const kpszNodeName,
              const char* const kpszNodeValue,
              MsgSetLinkedListBlk* lpMSLL);

long
ParseNewLineMsgSet();

void
AddSingleElementMsg(const char* const kpszMsgType,
                    const char* const kpszElementType,
                    const char* const kpszElementData);

long
PrepareMsgSets();

void
SetMsgTreeMSLLs(MsgSetHdr* lpMSH,
                MsgSetTreeNode* lpTNCall,
                MsgSetLinkedListBlk* lpRecursionMSLLHead,
                MsgDefBlk* lpRecursionMDB,
                SegmentBlk* lpRecursionSGB,
                long lRecursionDepth,
                long lMsgNumber);

void
SetMsgTreeSameNameLinks(MsgSetTreeNode* lpTNCall);

void
CreateOutMsgTreeNode(MsgSetTreeNode* lpTNMsgSetFile,
                     MsgSetTreeNode* lpTNOutParent, 
                     MsgSetTreeNode* lpTNOut,
                     MsgDefBlk* lpRecursionMDB,
                     SegmentBlk* lpRecursionSGB,
                     long lRecursionDepth,
                     long lMsgNumber);

MsgSetTreeNode*
CreateTreeNodeFromMsgDefBlk(MsgSetTreeNode* lpTNCurrOutParent,
                            MsgSetTreeNode* lpTNLastWithSameName,
                            MsgDefBlk* lpMDB,
                            long lRecursionDepth);


MsgSetTreeNode*
CreateTreeNodeFromSegmentBlk(MsgSetTreeNode* lpTNCurrOutParent,
                             MsgSetTreeNode* lpTNCurrOut,
                             MsgSetTreeNode* lpTNLastWithSameName,
                             MsgDefBlk* lpMDB,
                             SegmentBlk* lpSGB,
                             long lRecursionDepth);

MsgSetTreeNode*
CreateTreeNodeFromElementBlk(MsgSetTreeNode* lpTNCurrOutParent,
                             MsgSetTreeNode* lpTNCurrOut,
                             MsgSetTreeNode* lpTNLastWithSameName,
                             MsgDefBlk* lpMDB,
                             SegmentBlk* lpSGB,
                             ElementBlk* lpELB,
                             long lRecursionDepth);

MsgSetTreeNode*
FindMatchingChild(MsgSetTreeNode* lpTNParent, char* lpszChildName,
                  long lChildNameLen, long lChildIndexSameName,
                  MsgSetTreeNode** lplpTNLastWithSameName);

MsgSetTreeNode*
FindMatchingSibling(MsgSetTreeNode* lpTNSibling, char* lpszSiblingName,
                    long lSiblingNameLen, long lSiblingIndexSameName,
                    MsgSetTreeNode** lplpTNLastWithSameName);

MsgSetHdr*
CheckAndAddMSH(const char* const kpszMsgSetName);

MsgSetHdr*
FindMsgSetHdr(const char* const kpszMsgSetName);

void
GenerateMsgInBuff(TestInfoBlk* lpTIB,
                  BuffInfoBlk*lpBIB,
                  MsgSetTreeNode* lpMsgMSTN,
                  VarInfoBlk* lpVIBMQMD);

void
GenerateNodeData(TestInfoBlk* lpTIB,
                 BuffInfoBlk* lpBIB,
                 MsgSetTreeNode* lpTNData,
                 VarInfoBlk* lpVIBMQMD,
                 long lSegmentIndex);

void
NameListInErrorWorkBuff(MsgSetTreeNode* lpTN, long* lplNameListLen);

long
CheckValueIsNumeric(MsgSetTreeNode* lpTN, TestInfoBlk* lpTIB, const char* const kpszNumStr);

long
CheckValueIsHex(MsgSetTreeNode* lpTN, TestInfoBlk* lpTIB, const char* const kpszHexStr);

void
ResolveNodeValue(MsgSetTreeNode* lpTN, TestInfoBlk* lpTIB,
                 char** lplpszResStr, char** lplpszVarStr,
                 long bDisplayMsg61);


/*--------------------------------------------------------------------------*/
/* PCRE Callout Support                                                     */
/*--------------------------------------------------------------------------*/
static int __cdecl
PCRECallout(pcre_callout_block *cb);


/*--------------------------------------------------------------------------*/
/* Script test and command execution functions                              */
/*--------------------------------------------------------------------------*/
void
RunTests();

long
ExecuteGetFileCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecutePutFileCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteInlineDataCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteOverlayCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteExtractCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteSubstringCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteGetMsgCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecutePutMsgCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteSelectCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteWaitOnMsgDepthCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteWaitOnTimeCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteCompareCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteCommitCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteBackoutCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteSystemCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteMqscCommandCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteForStartCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteForEndCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteIfStartCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteElseIfCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteElseCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteIfEndCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteContinueCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteBreakCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteExitTestCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteFailTestCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteFinallyStartCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteFinallyEndCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteIntervalStartCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteIntervalEndCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteVarCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteMathCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteLengthCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecutePositionCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteStdInCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteStdOutCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

long
ExecuteUserLogCmd(TestInfoBlk* lpTIB, CommandCtrlBlk* lpCCB);

void
IfNextCmdIsUserLogThenExecute(TestInfoBlk* lpTIB);

CommandCtrlBlk*
FindFinallyCmd(TestInfoBlk* lpTIB);

long
BCSPutFile(TestInfoBlk* lpTIB, PutFileBlk* lpPFB, long* bSaveFile);

void
LogMsgCorrelAndGroupIds(TestInfoBlk* lpTIB, MQMD* lpMsgDesc);

/*--------------------------------------------------------------------------*/
/* File management functions                                                */
/*--------------------------------------------------------------------------*/
long
LoadFile(char* lpszFileName,
         char** lplpszBuff, long* lplBuffLen,
         long* lplDataLen, long bLogError);

long
SaveFile(char* lpszFileName, char* lpszBuff,
         long lBuffLen, long bSuppress_fopen_Error, long bAppend);

long
FileExists(char* lpszFileName);

long
PFSaveFileRFH2(TestInfoBlk* lpTIB, PutFileBlk* lpPFB);

long
PFSaveFileMQMD(TestInfoBlk* lpTIB, PutFileBlk* lpPFB);

void
LoadLIBFromFile(TestInfoBlk* lpTIB, ListInfoBlk** lplpLIB,
                char* lpszFileName, char* lpszDir);

void
LoadMIBFromFile(TestInfoBlk* lpTIB, MQMDInfoBlk** lplpMIB,
                char* lpszFileName, char* lpszDir);

void
LoadRIBFromFile(TestInfoBlk* lpTIB, RFH2InfoBlk** lplpRIB,
                char* lpszFileName, char* lpszDir);

long
LoadXMLFile(char* lpszXMLFileName, char** lplpszXMLBuff,
            long* lplXMLDataLen, long bLogMsgs);

long
OpenLogFile();

long
OpenReportFile();

long
OpenUserLogFile();

long
OpenIntervalFile();

long
WriteIntervalRecord(IntervalInfoBlk* lpIIB, IntervalInfoRec* lpIIR);

/*--------------------------------------------------------------------------*/
/* Message logging and display                                              */
/*--------------------------------------------------------------------------*/
void
LogMsg(long lMsgNum, ...);

void
ReportMsg(long lMsgNum, ...);

void
UserLogMsg(long lMsgNum, ...);

void
ConsoleMsg(long lMsgNum, ...);

char*
GetMsgPtr(long lMsgNum);

void
DisplayData(const char* const kpszPrefix, char* lpszPrintBuff,
            long lPrintBuffLen, char* lpszData, long lDataLen);

/*--------------------------------------------------------------------------*/
/* Script interval and timing related functions                             */
/*--------------------------------------------------------------------------*/
void
ReportStats();

IntervalInfoRec*
CreateAndAddIntervalInfoRec(IntervalInfoBlk* lpIIB);

void
SetIntervalStart(IntervalInfoRec* lpIIR);

void
SetIntervalEnd(IntervalInfoRec* lpIIR);

void
SetInterval(void* vp8Bytes);

double
CalcInterval(IntervalInfoRec* lpIIR, double* lpdStart,
             double* lpdEnd, double* lpdInterval, double* lpdRunSeconds);

double
CalcCmdExecTime(TestInfoBlk* lpTIB);

void
ControlMessageRate(TestInfoBlk* lpTIB, MessageRateBlk* lpMRB);

void
SleepMillis(long lSleepMillis);

void
GetClockCount(Int64* lpi64ClockCount);

/*--------------------------------------------------------------------------*/
/* Shutdown and resource deallocation functions                             */
/*--------------------------------------------------------------------------*/
void
FreeResources();

void
FreeBIB(BuffInfoBlk** lplpBIB);

void
FreeCDB(ConnectionDefBlk** lplpCDB);

void
FreeELB(ElementBlk** lplpELB);

void
FreeFIB(FolderInfoBlk** lplpFIB);

void
FreeFBR(ForBackRefBlk** lplpFBR);

void
FreeIFB(IfBackRefBlk** lplpIFB);

void
FreeIIB(IntervalInfoBlk** lplpIIB);

void
FreeLEB(ListElementInfoBlk** lplpLEB);

void
FreeLIB(ListInfoBlk** lplpLIB);

void
FreeMRB(MessageRateBlk** lplpMRB);

void
FreeMSFL(MsgSetFileLoadBlk** lplpMSFL);

void
FreeMIB(MQMDInfoBlk** lplpMIB);

void
FreeMSH(MsgSetHdr** lplpMSH);

void
FreeMDB(MsgDefBlk** lplpMDB);

void
FreeSDB(SegmentDefBlk** lplpSDB);

void
FreeMSLL(MsgSetLinkedListBlk** lplpMSLL);

void
FreeSGB(SegmentBlk** lplpSGB);

void
FreeELB(ElementBlk** lplpELB);

void
FreeQIB(QueueInfoBlk** lplpQIB);

void
FreeRIB(RFH2InfoBlk** lplpRIB);

void
FreeSGB(SegmentBlk** lplpSGB);

void
FreeVIB(VarInfoBlk** lplpVIB);

void
FreeCCB(CommandCtrlBlk** lplpCCB);

void
FreeTIB(TestInfoBlk** lplpTIB);

/*--------------------------------------------------------------------------*/
/* Data conversion functions                                                */
/*--------------------------------------------------------------------------*/
long
GetLocalCCSID();

long
ConvertToCCSID(const char* const kpszCodePage);

void
ReverseLong(long *lplNum);

void
CheckAndFixRFH2Longs(MQRFH2* lpoRFH2, long lLocalEncoding, long lCurrentEncoding, long lRequiredEncoding);

/*--------------------------------------------------------------------------*/
/* Utility functions                                                        */
/*--------------------------------------------------------------------------*/
BuffInfoBlk*
FindBuffInfoBlk(BuffInfoBlk* lpBIB, char* lpszFindBuffName);

BuffInfoBlk*
CheckAndAddBIB(TestInfoBlk* lpTIB, char* lpszFindBuffName);

ConnectionDefBlk*
FindConnectionDefBlk(ConnectionDefBlk* lpCBD, const char* const kpszFindBuffName);

FolderInfoBlk*
FindFolderInfoBlk(FolderInfoBlk* lpFIB, char* lpszFindFolderName);

FolderInfoBlk*
CheckAndAddFIB(TestInfoBlk* lpTIB, RFH2InfoBlk* lpRIB, long bPad);

ListElementInfoBlk*
AddLEB(ListInfoBlk* lpLIB, const char* const kpszListElement);

QueueInfoBlk*
OpenQueue(TestInfoBlk* lpTIB, ConnectionDefBlk* lpCDB, QueueInfoBlk* lpQIB,
          char* lpszTargQName, char* lpszTargQMgrName, char* lpszTargDynQName,
          long bOutputOnly);

VarInfoBlk*
FindVarInfoBlk(VarInfoBlk* lpVIB, char* lpszFindVarName);

VarInfoBlk*
CheckAndAddVIB(TestInfoBlk* lpTIB, enum eVariableEnum eVarType,
               char* lpszVarName, char* lpszResVarName,
               char* lpszVarFormat);

IntervalInfoBlk*
FindIntervalInfoBlk(char* lpszFindIntervalName);

MsgSetLinkedListBlk*
FindMsgSetLinkedListBlk(MsgSetLinkedListBlk* lpMSLL, char* lpszFindName, long FindNameLen);

MsgSetLinkedListBlk*
FindNextRequiredMSLL(MsgSetLinkedListBlk* lpMSLL);

MsgSetTreeNode*
FindMsgTreeNode(MsgSetTreeNode* lpTNParent, MsgSetTreeNode* lpTNCurr,
                char* lpszFindName, long lFindNameLen);
void
AddOrUpdateIIR(IntervalInfoBlk* lpIIB, enum eIntervalEnum eIntervalStartEnd);

void
CloneFIB(FolderInfoBlk** lplpFIBTo, FolderInfoBlk* lpFIBFrom);

void
CloneLEB(ListElementInfoBlk** lplpLEBTo, ListElementInfoBlk* lpLEBFrom);

void
CloneLIB(ListInfoBlk** lplpLIBTo, ListInfoBlk* lpLIBFrom);

void
CloneMIB(MQMDInfoBlk** lplpMIBTo, MQMDInfoBlk* lpMIBFrom);

void
CloneRIB(RFH2InfoBlk** lplpRIBTo, RFH2InfoBlk* lpRIBFrom);

void
OverrideLIB(ListInfoBlk* lplpLIBTo, ListInfoBlk* lpLIBFrom);

void
OverrideMIB(MQMDInfoBlk* lplpMIBTo, MQMDInfoBlk* lpMIBFrom);

void
OverrideRIB(RFH2InfoBlk* lplpRIBTo, RFH2InfoBlk* lpRIBFrom);

long
CompareRIBs(TestInfoBlk* lpTIB,
            RFH2InfoBlk* lpCompRIB, RFH2InfoBlk* lpBuffRIB);

long
GetFolderName(TestInfoBlk* lpTIB,
              char** lplpszFolderData, char** lplpszFolderName);

void
ClearAttrs();

void
ClearRC();

void
ClearCIB();

void
ClearCSE();

void
ClearPCDB();

void
ClearLIB(ListInfoBlk* lpLIB);

void
ClearMIB(MQMDInfoBlk* lpMIB);

void
ClearRIB(RFH2InfoBlk* lpRIB);

VarInfoBlk*
AddAutoNumVar(VarInfoBlk** lplpVIBHead,
              VarInfoBlk** lplpVIBTail, char* lpszAutoVarName);

VarInfoBlk*
AddAutoStrVar(VarInfoBlk** lplpVIBHead,
              VarInfoBlk** lplpVIBTail, char* lpszAutoVarName);

void
SetAutoVar(TestInfoBlk* lpTIB, VarInfoBlk* lpVIBHead,
           char* lpszAutoVarName, void* lpVoid);

void
InitTimePidAutoVars(TestInfoBlk* lpTIB,
                    VarInfoBlk* lpVIBHead);

void
InitScriptAutoVars(char** argv);

void
AddToTailOfList(LinkedListBlk** lplpHead, LinkedListBlk** lplpTail,
                LinkedListBlk* lpBlk);

void
RemoveFromList(LinkedListBlk** lplpHead, LinkedListBlk** lplpTail,
               LinkedListBlk* lpBlk);

char*
CloneString(const char* const kpszStringToClone);

char*
CloneStringNull(const char* const kpszStringToClone);

char*
TrimString(const char* const kpszStringToTrim);

void
ChangeIfStr1Empty(char** lplpszStr1, char** lplpszStr2);

void
ChangeIfStr2Exists(char** lplpszStr1, char** lplpszStr2);

void
ResolvePgmValue(char** lplpszResStr, char** lplpszVarStr);

void
ResolveValue(TestInfoBlk* lpTIB, char** lplpszResStr, char** lplpszVarStr,
             long bDisplayMsg61);

void
ResolveSubscript(TestInfoBlk* lpTIB, char** lplpszResStr, char** lplpszVarStr,
                 long bDisplayMsg87);

void
SetMQMDFromMIB(TestInfoBlk* lpTIB, MQMDInfoBlk* lpMIB,
               char* lpszMsgId, char* lpszCorrelId, char* lpszGroupId,
               char* lpszMsgSeqNumber, char* lpszFormat, MQMD* lpMsgDesc,
               long bSetForGet, long bDisplayMsg61);

long
CreateMIBFromMQMD(MQMDInfoBlk** lplpMIB, MQMD* lpMsgDesc);

void
GenByte24Str(char* lpszWorkBuff, char* lpszByte24Value, long bHexIfSpecialChars);

void
GenByte32Str(char* lpszWorkBuff, char* lpszByte32Value);

long
ConvNum(TestInfoBlk* lpTIB,
        const char* const kpszNum, const char* const kpszName);
long
ConvNumFailReturnsZero(const char* const kpszNum);

double
ConvDouble(TestInfoBlk* lpTIB,
           const char* const kpszDouble, const char* const kpszName);

long
IsNumeric(const char* const kpszNum);

long
IsDouble(const char* const kpszNum);

void
HexToLong(long* lpLong, const char* const lpszHexStr);

void
HexToBuff(char* lpszBuff, const char* const lpszHexStr, long lMaxLen);

long
StrlenNull(const char* const kpszStr);

long
FindRC(long lFindRCCount, long lFindRC[], long lCurrRC);

void
RemoveXmlWhiteSpace(char** lplpszXMLBuff, long* lplXMLDataLen, long lDataToProcessLen);

void
SetTIBCodes(TestInfoBlk* lpTIB, long lCC, long lRC);

void
ConvertEscapeSequences(char* lpszConvertStr);

void
FreeIfNotNull(void** vpvpFreeBlock);

long
CheckAndExtendBuff(BuffInfoBlk* lpBIB, long lOutputDataLen);

long
BuffConvertToInt(BuffInfoBlk* lpBIB, const char* const kpszNum, long lOutputDataLen);

void
PackedToBuff(char* lpszBuff, const char* const lpszPackedStr, long lMaxLen);

long
EvaluateCondition(TestInfoBlk* lpTIB, const char* const kpszCond, long bDisplayMsg255);

void
EvaluateMathExpression(TestInfoBlk* lpTIB, const char* const kpszEval, enum eVariableEnum eVarType, void* vpResult);

/*--------------------------------------------------------------------------*/
/* Debug routines                                                           */
/*--------------------------------------------------------------------------*/
void
DisplayMsgTree(char* lpszLabel, MsgSetTreeNode* lpTNOut, 
               long lRecursionDepth,
               long lMsgNumber);

void
DisplayNodeInfo(char* lpszLabel, long lMsgNumber,
                long lRecursionDepth, MsgSetTreeNode* lpTN);
void
DisplayNames(MsgSetTreeNode* lpTN);

void
DisplayBlockInfo(MsgSetTreeNode* lpTN);

#endif /* __TJAINC_MSGTEST_H */
