/*==========================================================================*/
/* MsgTestCtrlBlk.c contains the control block support routines.            */
/*==========================================================================*/
/* (C) Copyright Tim Armstrong 2003, 2004, 2005, 2006, 2007, 2008           */
/*==========================================================================*/
/* The EXPAT copyright information is repeated below                        */
/*==========================================================================*/

/*==========================================================================*/
/* EXPAT Copyright information                                              */
/*==========================================================================*/
/*
Copyright (c) 1998, 1999, 2000 Thai Open Source Software Center Ltd
                               and Clark Cooper
Copyright (c) 2001, 2002 Expat maintainers.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/*==========================================================================*/
/* End of EXPAT Copyright information                                       */
/*==========================================================================*/

/*==========================================================================*/
/* Includes                                                                 */
/*==========================================================================*/
#ifndef __TJAINC_MSGTEST_H
#include "MsgTest.h"
#endif /* __TJAINC_MSGTEST_H */

#ifndef __TJAINC_MSGTESTSTATICMD_H
#include "MsgTestStaticMD.h"
#endif /* __TJAINC_MSGTESTSTATICMD_H */

/*==========================================================================*/
/* Static variables                                                         */
/*==========================================================================*/
/* Element names */
static char* kpszMQRFH2 =             "MQRFH2\0";
static char* kpszRFH2 =               "RFH2\0";
static char* kpszFOLDER =             "Folder\0";


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     ParseCtrlBlk                                                         */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Parse control block information that has been loaded in XML format   */
/*     Currently the MQMD and MQRFH2 control blocks are supported           */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpSPB   Pointer to the main control block structure                  */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     SUCCESS or FAILURE                                                   */
/*==========================================================================*/
long
ParseCtrlBlk(TestInfoBlk* lpTIB, char* lpszCtrlBlkType)
{
    /* Create the parser */
    XML_Parser oParser = XML_ParserCreate(NULL);

    /* Give the parser the test info block pointer for its callback functions */
    XML_SetUserData(oParser, lpTIB);
    
    /* Provide the addresses of our various callback functions */
    XML_SetElementHandler(oParser, PCBParseStartElement, PCBParseEndElement);
    XML_SetCharacterDataHandler(oParser, PCBParseCharData);
    XML_SetCommentHandler(oParser, PCBComments);
    
    /* Call the parser */
    if(XML_STATUS_ERROR == XML_Parse(oParser, kpMCB->lpSPB->lpszCtrlBlkBuff,
                                     kpMCB->lpSPB->lCtrlBlkXmlLen, 1))
    {
        /* EXPAT identified a problem during parsing */
        LogMsg(65, lpTIB->lTestNum, XML_ErrorString(XML_GetErrorCode(oParser)));
        return FAILURE;
    }

    /* Done with the parser so let it know */
    XML_ParserFree(oParser);

    if(FAILURE == lpTIB->bCmdResult)
    {
        /* MsgTest identified a problem during parsing */
        LogMsg(66, lpTIB->lTestNum);
        return FAILURE;
    }

    /* So far so good. Tell the world */
    LogMsg(67, lpTIB->lTestNum, lpszCtrlBlkType, 
           kpMCB->lpSPB->lpszCtrlBlkFileName);
    return SUCCESS;
}

/*--------------------------------------------------------------------------*/
static void XMLCALL
PCBParseStartElement(void* vpUserData, const char* kpszElementName,
                     const char** kpkpAtts)
{
    SecondaryParseBlk* lpSPB;
    TestInfoBlk*       lpTIB;
    enum eElementEnum  eElementIs;
    char               lpszElementNameBuff[NAME_BUFF_LEN + 4];

    /* Get the test info, main control block pointers and */
    lpTIB = vpUserData;
    lpSPB = kpMCB->lpSPB;

    /* Already FAILED? Don't do anything else */
    if(FAILURE == lpTIB->bCmdResult)
    {
        return;
    }

    /* New element so reset the ParseCharData buffer pointer */
    ClearPCDB();

    /* Check element name length */
    if(NAME_BUFF_LEN < StrlenNull(kpszElementName))
    {
        LogMsg(68, lpTIB->lTestNum, "Element", kpszElementName, NAME_BUFF_LEN);
        lpTIB->bCmdResult = FAILURE;
        return;
    }

    /* Copy the element name to our local buffer */
    memset(lpszElementNameBuff, 0x00, NAME_BUFF_LEN + 4);
    strncpy(lpszElementNameBuff, kpszElementName, NAME_BUFF_LEN);

    /* Copy the element name to an element name buffer */
    memset(kpMCB->lpSPB->lplpszElementName[lpSPB->lCtrlBlkParseDepth], 0x00, NAME_BUFF_LEN + 4);
    strncpy(kpMCB->lpSPB->lplpszElementName[lpSPB->lCtrlBlkParseDepth], kpszElementName, NAME_BUFF_LEN);

    /* Not expecting any attributes so its an error */
    if(kpkpAtts[0])
    {
        LogMsg(84, lpTIB->lTestNum, kpkpAtts[0]);
        lpTIB->bCmdResult = FAILURE;
        return;
    }

    /* Which element type are we processing for */
    eElementIs = PCBGetElementType(lpszElementNameBuff);
    if(E_UNKNOWN == eElementIs)
    {
        PCBDisplayParseError(139);
        lpTIB->bCmdResult = FAILURE;
        return;
    }

    /* Now we know which element type it is */
    lpSPB->eElementType[lpSPB->lCtrlBlkParseDepth] = eElementIs;

    /* Handle different initialisation requirements for elements */
    switch(eElementIs)
    {
    case E_MQRFH2:
        ClearRIB(kpMCB->lpRIB);
        break;
    case E_MQMD:
        /* Initialise the MQMD parse string pointers */
        ClearMIB(kpMCB->lpMIB);
        break;
    /* There is no default action for elements that dont need initialising */
    }

    /* Increment the parse element depth */
    lpSPB->lCtrlBlkParseDepth++;
    
    /* Check we haven't exceeded the maximum logical tag depth  */
    /* I.e. MQMD.Version should not contain another element     */
    /* MQMD.Version.xxxx is illegal                             */
    if(MAX_CTRLBLK_PARSE_DEPTH < lpSPB->lCtrlBlkParseDepth)
    {
        lpSPB->lCtrlBlkParseDepth = MAX_CTRLBLK_PARSE_DEPTH;
        PCBDisplayParseError(16);
        lpTIB->bCmdResult = FAILURE;
        return;
    }
}

/*--------------------------------------------------------------------------*/
static void XMLCALL
PCBParseEndElement(void* vpUserData, const char* kpszElementName)
{
    SecondaryParseBlk* lpSPB;
    TestInfoBlk*       lpTIB;
    enum eElementEnum  eElementIs;
    char               lpszElementNameBuff[NAME_BUFF_LEN + 4];

    /* Get the test info and secondary parse control block pointers */
    lpTIB = vpUserData;
    lpSPB = kpMCB->lpSPB;

    /* Already FAILED? Don't do anything else */
    if(FAILURE == lpTIB->bCmdResult)
    {
        return;
    }


    /* Copy the element name to our local buffer */
    memset(lpszElementNameBuff, 0x00, NAME_BUFF_LEN + 4);
    strncpy(lpszElementNameBuff, kpszElementName, NAME_BUFF_LEN);

    /* Decrement the parse element depth */
    lpSPB->lCtrlBlkParseDepth--;
    if(0 > lpSPB->lCtrlBlkParseDepth)
    {
        lpSPB->lCtrlBlkParseDepth = 0;
        LogMsg(72, lpTIB->lTestNum);
        lpTIB->bCmdResult = FAILURE;
        return;
    }

    /* Which element type are we processing for */
    eElementIs = PCBGetElementType(lpszElementNameBuff);
    if(E_UNKNOWN == eElementIs)
    {
        PCBDisplayParseError(139);
        lpTIB->bCmdResult = FAILURE;
        return;
    }

    /* Element processing varies depending on parse depth   */
    switch(lpSPB->lCtrlBlkParseDepth)
    {
    /* At parse depth 0 <MQRFH2>, <MQMD> are the only valid elements */
    case 0:
        switch(eElementIs)
        {
        case E_MQRFH2:
        case E_MQMD:
            /* Nothing need be done but its not an error */
            break;
        default:
            LogMsg(71, lpTIB->lTestNum, lpszElementNameBuff);
            lpTIB->bCmdResult = FAILURE;
            return;
        }
        break;
    /* At parse depth 1 we are assigning the element the   */
    /* contents of the parse char data string              */
    case 1:
        /* Control info block related processing           */
        switch(lpSPB->eElementType[0])
        {
        /* MQRFH2 */
        case E_MQRFH2:
            switch(eElementIs)
            {
            case E_FORMAT:
                kpMCB->lpRIB->lpszFormat = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_FOLDER:
                /* Hold Folder info in the MCB's linked list until it  */
                /* can be consumed by the <MQRFH2> control block       */
                CheckAndAddFIB(lpTIB, kpMCB->lpRIB, TRUE);
                break;
            }
            break;
        /* MQMD */
        case E_MQMD:
            switch(eElementIs)
            {
            case E_VERSION:
                kpMCB->lpMIB->lpszVersion = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_REPORT:
                kpMCB->lpMIB->lpszReport = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_MSGTYPE:
                kpMCB->lpMIB->lpszMsgType = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_EXPIRY:
                kpMCB->lpMIB->lpszExpiry = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_FEEDBACK:
                kpMCB->lpMIB->lpszFeedback = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_ENCODING:
                kpMCB->lpMIB->lpszEncoding = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_CODEDCHARSETID:
                kpMCB->lpMIB->lpszCodedCharSetId = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_FORMAT:
                kpMCB->lpMIB->lpszFormat = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_PRIORITY:
                kpMCB->lpMIB->lpszPriority = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_PERSISTENCE:
                kpMCB->lpMIB->lpszPersistence = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_MSGID:
                kpMCB->lpMIB->lpszMsgId = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_CORRELID:
                kpMCB->lpMIB->lpszCorrelId = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_BACKOUTCOUNT:
                kpMCB->lpMIB->lpszBackoutCount = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_REPLYTOQ:
                kpMCB->lpMIB->lpszReplyToQ = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_REPLYTOQMGR:
                kpMCB->lpMIB->lpszReplyToQMgr = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_USERIDENTIFIER:
                kpMCB->lpMIB->lpszUserIdentifier = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_ACCOUNTINGTOKEN:
                kpMCB->lpMIB->lpszAccountingToken = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_APPLIDENTITYDATA:
                kpMCB->lpMIB->lpszApplIdentityData = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_PUTAPPLTYPE:
                kpMCB->lpMIB->lpszPutApplType = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_PUTAPPLNAME:
                kpMCB->lpMIB->lpszPutApplName = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_PUTDATE:
                kpMCB->lpMIB->lpszPutDate = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_PUTTIME:
                kpMCB->lpMIB->lpszPutTime = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_APPLORIGINDATA:
                kpMCB->lpMIB->lpszApplOriginData = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_GROUPID:
                kpMCB->lpMIB->lpszGroupId = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_MSGSEQNUMBER:
                kpMCB->lpMIB->lpszMsgSeqNumber = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_OFFSET:
                kpMCB->lpMIB->lpszOffset = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_MSGFLAGS:
                kpMCB->lpMIB->lpszMsgFlags = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            case E_ORIGINALLENGTH:
                kpMCB->lpMIB->lpszOriginalLength = CloneString(kpMCB->lpszParseCharDataBuff);
                break;
            default:
                LogMsg(4, "Invalid parse2 depth 1 element", lpszElementNameBuff, __LINE__, __FILE__);
                lpTIB->bCmdResult = FAILURE;
                return;
            }
            break;
        }
        /* Copied the ParseCharData content so reset the buffer */
        ClearPCDB();
        break;
    default:
        ConsoleMsg(4, "Invalid parse2 depth", lpszElementNameBuff, __LINE__, __FILE__);
        kpMCB->lProcessing = FAILURE;
        return;
    }

    /* Clear the element name buffer and element type entries */
    if(0 <= lpSPB->lCtrlBlkParseDepth)
    {
        memset(lpSPB->lplpszElementName[lpSPB->lCtrlBlkParseDepth], 0x00, NAME_BUFF_LEN + 4);
        lpSPB->eElementType[lpSPB->lCtrlBlkParseDepth] = E_UNKNOWN;
    }
}

/*--------------------------------------------------------------------------*/
static void XMLCALL
PCBParseCharData(void* vpUserData, const XML_Char* kpszCD, int lLen)
{
    SecondaryParseBlk* lpSPB;
    TestInfoBlk*       lpTIB;
    long               lPCDBuffLen;

    /* Get the test info, main control block pointers and */
    lpTIB = vpUserData;
    lpSPB = kpMCB->lpSPB;

    /* Already FAILED? Don't do anything else */
    if(FAILURE == lpTIB->bCmdResult)
    {
        return;
    }

    /* ???TA Need to handle multiple calls as according to doco */
    /* ???TA we may not get everything at once                  */

    /* Got some data so allocate and initialise an area to copy it to */
    lPCDBuffLen = (((lLen + 8) / 8) + 1) * 8;
    kpMCB->lParseCharDataLen = lLen;
    kpMCB->lpszParseCharDataBuff = malloc(lPCDBuffLen);
    if(NULL == kpMCB->lpszParseCharDataBuff)
    {
        kpMCB->lParseCharDataLen = 0;
        kpMCB->lpszParseCharDataBuff = NULL;
        LogMsg(3, lPCDBuffLen, "ParseCharData", __LINE__, __FILE__);
        lpTIB->bCmdResult = FAILURE;
        return;
    }
    memset(kpMCB->lpszParseCharDataBuff, 0x00, lPCDBuffLen);

    /* Copy the data contained in the element to our new buffer */
    memcpy(kpMCB->lpszParseCharDataBuff, kpszCD, lLen);
}

/*--------------------------------------------------------------------------*/
static void XMLCALL
PCBComments(void* vpUserData, const XML_Char* kpszComment)
{
    /* Don't do anything with comments at the moment */

    /* Pretend to do something so dont get level 4 warnings */
    /* about unreferenced formal parameters                 */
    void*  vpUD;
    char*  kpszC;
    
    vpUD = vpUserData;
    kpszC = (char*)kpszComment;
}

/*--------------------------------------------------------------------------*/
void
PCBDisplayParseError(long lMsgNum)
{
    long  lIdx;
    char  lpszExtraInfo[EXTRA_MSG_INFO_LEN + 4];
    char* lpszExtraBuff;
    
    lpszExtraBuff = lpszExtraInfo;
    memset(lpszExtraBuff, 0x00, EXTRA_MSG_INFO_LEN + 4);

    for(lIdx = 0; lIdx <= kpMCB->lpSPB->lCtrlBlkParseDepth; lIdx++)
    {
        sprintf(lpszExtraBuff, "%s.", kpMCB->lpSPB->lplpszElementName[lIdx]);
        lpszExtraBuff += StrlenNull(kpMCB->lpSPB->lplpszElementName[lIdx]) + 1;
    }

    LogMsg(lMsgNum, lpszExtraInfo);
}

/*--------------------------------------------------------------------------*/
enum eElementEnum
PCBGetElementType(const char *kpszElementName)
{
    enum eElementEnum eElementIs;

    /****** Ok which element is it ? ******/
    if(0 == memcmp(kpszMQMD, kpszElementName, strlen(kpszMQMD) + 1))
    {
        eElementIs = E_MQMD;
    }
    else if(0 == memcmp(kpszVERSION, kpszElementName, strlen(kpszVERSION) + 1))
    {
        eElementIs = E_VERSION;
    }
    else if(0 == memcmp(kpszREPORT, kpszElementName, strlen(kpszREPORT) + 1))
    {
        eElementIs = E_REPORT;
    }
    else if(0 == memcmp(kpszMSGTYPE, kpszElementName, strlen(kpszMSGTYPE) + 1))
    {
        eElementIs = E_MSGTYPE;
    }
    else if(0 == memcmp(kpszEXPIRY, kpszElementName, strlen(kpszEXPIRY) + 1))
    {
        eElementIs = E_EXPIRY;
    }
    else if(0 == memcmp(kpszFEEDBACK, kpszElementName, strlen(kpszFEEDBACK) + 1))
    {
        eElementIs = E_FEEDBACK;
    }
    else if(0 == memcmp(kpszENCODING, kpszElementName, strlen(kpszENCODING) + 1))
    {
        eElementIs = E_ENCODING;
    }
    else if(0 == memcmp(kpszCODEDCHARSETID, kpszElementName, strlen(kpszCODEDCHARSETID) + 1))
    {
        eElementIs = E_CODEDCHARSETID;
    }
    else if(0 == memcmp(kpszFORMAT, kpszElementName, strlen(kpszFORMAT) + 1))
    {
        eElementIs = E_FORMAT;
    }
    else if(0 == memcmp(kpszPRIORITY, kpszElementName, strlen(kpszPRIORITY) + 1))
    {
        eElementIs = E_PRIORITY;
    }
    else if(0 == memcmp(kpszPERSISTENCE, kpszElementName, strlen(kpszPERSISTENCE) + 1))
    {
        eElementIs = E_PERSISTENCE;
    }
    else if(0 == memcmp(kpszMSGID, kpszElementName, strlen(kpszMSGID) + 1))
    {
        eElementIs = E_MSGID;
    }
    else if(0 == memcmp(kpszCORRELID, kpszElementName, strlen(kpszCORRELID) + 1))
    {
        eElementIs = E_CORRELID;
    }
    else if(0 == memcmp(kpszBACKOUTCOUNT, kpszElementName, strlen(kpszBACKOUTCOUNT) + 1))
    {
        eElementIs = E_BACKOUTCOUNT;
    }
    else if(0 == memcmp(kpszREPLYTOQ, kpszElementName, strlen(kpszREPLYTOQ) + 1))
    {
        eElementIs = E_REPLYTOQ;
    }
    else if(0 == memcmp(kpszREPLYTOQMGR, kpszElementName, strlen(kpszREPLYTOQMGR) + 1))
    {
        eElementIs = E_REPLYTOQMGR;
    }
    else if(0 == memcmp(kpszUSERIDENTIFIER, kpszElementName, strlen(kpszUSERIDENTIFIER) + 1))
    {
        eElementIs = E_USERIDENTIFIER;
    }
    else if(0 == memcmp(kpszACCOUNTINGTOKEN, kpszElementName, strlen(kpszACCOUNTINGTOKEN) + 1))
    {
        eElementIs = E_ACCOUNTINGTOKEN;
    }
    else if(0 == memcmp(kpszAPPLIDENTITYDATA, kpszElementName, strlen(kpszAPPLIDENTITYDATA) + 1))
    {
        eElementIs = E_APPLIDENTITYDATA;
    }
    else if(0 == memcmp(kpszPUTAPPLTYPE, kpszElementName, strlen(kpszPUTAPPLTYPE) + 1))
    {
        eElementIs = E_PUTAPPLTYPE;
    }
    else if(0 == memcmp(kpszPUTAPPLNAME, kpszElementName, strlen(kpszPUTAPPLNAME) + 1))
    {
        eElementIs = E_PUTAPPLNAME;
    }
    else if(0 == memcmp(kpszPUTDATE, kpszElementName, strlen(kpszPUTDATE) + 1))
    {
        eElementIs = E_PUTDATE;
    }
    else if(0 == memcmp(kpszPUTTIME, kpszElementName, strlen(kpszPUTTIME) + 1))
    {
        eElementIs = E_PUTTIME;
    }
    else if(0 == memcmp(kpszAPPLORIGINDATA, kpszElementName, strlen(kpszAPPLORIGINDATA) + 1))
    {
        eElementIs = E_APPLORIGINDATA;
    }
    else if(0 == memcmp(kpszGROUPID, kpszElementName, strlen(kpszGROUPID) + 1))
    {
        eElementIs = E_GROUPID;
    }
    else if(0 == memcmp(kpszMSGSEQNUMBER, kpszElementName, strlen(kpszMSGSEQNUMBER) + 1))
    {
        eElementIs = E_MSGSEQNUMBER;
    }
    else if(0 == memcmp(kpszOFFSET, kpszElementName, strlen(kpszOFFSET) + 1))
    {
        eElementIs = E_OFFSET;
    }
    else if(0 == memcmp(kpszMSGFLAGS, kpszElementName, strlen(kpszMSGFLAGS) + 1))
    {
        eElementIs = E_MSGFLAGS;
    }
    else if(0 == memcmp(kpszORIGINALLENGTH, kpszElementName, strlen(kpszORIGINALLENGTH) + 1))
    {
        eElementIs = E_ORIGINALLENGTH;
    }
    else if(0 == memcmp(kpszMQRFH2, kpszElementName, strlen(kpszMQRFH2) + 1))
    {
        eElementIs = E_MQRFH2;
    }
    else if(0 == memcmp(kpszRFH2, kpszElementName, strlen(kpszRFH2) + 1))
    {
        eElementIs = E_RFH2;
    }
    else if(0 == memcmp(kpszFOLDER, kpszElementName, strlen(kpszFOLDER) + 1))
    {
        eElementIs = E_FOLDER;
    }
    else
    {
        eElementIs = E_UNKNOWN;
    }

    return eElementIs;
}
