/*==========================================================================*/
/* MsgTestMsgSet.c contains the message set processing support routines.    */
/*==========================================================================*/
/* (C) Copyright Tim Armstrong 2003, 2004, 2005, 2006, 2007, 2008           */
/*==========================================================================*/
/* The EXPAT copyright information is repeated below                        */
/*==========================================================================*/

/*==========================================================================*/
/* EXPAT Copyright information                                              */
/*==========================================================================*/
/*
Copyright (c) 1998, 1999, 2000 Thai Open Source Software Center Ltd
                               and Clark Cooper
Copyright (c) 2001, 2002 Expat maintainers.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/*==========================================================================*/
/* End of EXPAT Copyright information                                       */
/*==========================================================================*/

/*==========================================================================*/
/* Includes                                                                 */
/*==========================================================================*/
#ifndef __TJAINC_MSGTEST_H
#include "MsgTest.h"
#endif /* __TJAINC_MSGTEST_H */

#ifndef __TJAINC_MSGTESTSTATICMD_H
#include "MsgTestStaticMD.h"
#endif /* __TJAINC_MSGTESTSTATICMD_H */

/*==========================================================================*/
/* Static variables                                                         */
/*==========================================================================*/
/* Value keywords */
static char* kpszNONE =               "NONE\0";
#ifdef WIN32
static char* kpszDirFileConcat =      "\\\0";
#else
static char* kpszDirFileConcat =      "/\0";
#endif /* WIN32 */

/* Element names */
static char* kpszMSGSET =             "MsgSet\0";
static char* kpszMSGTEMPLATES =       "MessageTemplates\0";
static char* kpszMESSAGES =           "Messages\0";
static char* kpszSEGDEF =             "SegDef\0";
static char* kpszMSGDEF =             "MsgDef\0";
static char* kpszMSG =                "Msg\0";
static char* kpszELEMENT =            "Element\0";
static char* kpszSEGMENT =            "Segment\0";
static char* kpszMESSAGESET =         "message set\0";
static char* kpszNAME =               "Name\0";
static char* kpszLEN =                "Len\0";
static char* kpszPOS =                "Pos\0";
static char* kpszTYPE =               "Type\0";
static char* kpszOUTPUTFORMAT =       "OutputFormat\0";
static char* kpszMSGSEP =             "MsgSep\0";
static char* kpszSEGSEP =             "SegSep\0";
static char* kpszELEMENTSEP =         "ElementSep\0";
static char* kpszMINOCCURS =          "MinOccurs\0";
static char* kpszMAXOCCURS =          "MaxOccurs\0";
static char* kpszCCSID =              "CCSID\0";
static char* kpszJUSTIFY =            "Justify\0";
static char* kpszPADCHAR =            "PadChar\0";
static char* kpszENDALIGN =           "EndAlign\0";
static char* kpszCMQVALUES =          "CmqValues\0";
static char* kpszTRUNCATE =           "Truncate\0";
static char* kpszMLINEDATA =          "M_LineData\0";
static char* kpszSLINEDATA =          "S_LineData\0";
static char* kpszLINEDATA =           "LineData\0";
static char* kpszCHAR =               "Char\0";
static char* kpszHEX =                "Hex\0";
static char* kpszINT8 =               "Int8\0";
static char* kpszINT16 =              "Int16\0";
static char* kpszINT32 =              "Int32\0";
static char* kpszINT64 =              "Int64\0";
static char* kpszPD =                 "PD\0";
static char* kpszROOT =               "Root\0";

/* Parser names */
static char* kpszPN_MSGSET =          "MSGSET\0";
static char* kpszPN_NEWLINE =         "NEWLINE\0";

/* Values */
static char* kpszZERO =               "0\0";
static char* kpszONE =                "1\0";
static char* kpszMINUSONE =           "-1\0";
static char* kpszUNBOUNDED =          "unbounded\0";
static char* kpszFIXED =              "FIXED\0";
static char* kpszYES =                "Yes\0";
static char* kpszLEFT =               "Left\0";
static char* kpszRIGHT =              "Right\0";
static char* kpszEmptyString =        "\0";
static char* kpszSEGLEN =             "%SegLen%\0";
static char* kpszSEGLENM1 =           "%SegLen-1%\0";
static char* kpszSEGLENM2 =           "%SegLen-2%\0";
static char* kpszSEGLENM4 =           "%SegLen-4%\0";
static long  klUNBOUNDED =            -1;
static long  klUNDEFINED =            -1;

/*==========================================================================*/
/* FUNCTION:                                                                */
/*     LoadAndParseMsgSetFiles                                              */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Load and parse any message set files                                 */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     N/A                                                                  */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     SUCCESS or FAILURE                                                   */
/*==========================================================================*/
long
LoadAndParseMsgSetFiles()
{
    MsgSetFileLoadBlk* lpMSFL;
    char               lpszDirAndFile[FILE_BUFF_LEN + 4];
    long               lMsgSetBuffLen;
    long               bLoadFile;
    long               bParseFile;

    /* Process the message set files */
    lpMSFL = kpMCB->lpCIB->lpMSFLHead;
    while(NULL != lpMSFL)
    {
        /* Generate full path file name for loading */
        memset(lpszDirAndFile, 0x00, FILE_BUFF_LEN + 4);
        strncpy(lpszDirAndFile, lpMSFL->lpszDir, FILE_BUFF_LEN);
        strcat(lpszDirAndFile, kpszDirFileConcat);
        strncat(lpszDirAndFile, lpMSFL->lpszFileName, FILE_BUFF_LEN);

        /* Check if a message set header block of the correct name exists */
        /* and if it does not then create one                             */
        kpMCB->lpMSPB->lpCurrMSH = CheckAndAddMSH(lpMSFL->lpszMsgSetName);
        if(NULL == kpMCB->lpMSPB->lpCurrMSH)
        {
            return FAILURE;
        }
        else
        {
            kpMCB->lpMSPB->lpCurrMSTN = kpMCB->lpMSPB->lpCurrMSH->lpMSFRootNode;
        }

        /* Invoke the appropriate parser */
        if(0 == memcmp(kpszPN_MSGSET, lpMSFL->lpszParserName, strlen(kpszPN_MSGSET) + 1))
        {
            /* See if we can load the file */
            bLoadFile = LoadXMLFile(lpszDirAndFile, &kpMCB->lpMSPB->lpszMsgSetBuff,
                                    &kpMCB->lpMSPB->lMsgSetDataLen, TRUE);
            if(FAILURE == bLoadFile)
            {
                LogMsg(137, lpszDirAndFile);
                return FAILURE;
            }

            /* Now parse it */
            bParseFile = ParseMsgSet();
        }
        else if(0 == memcmp(kpszPN_NEWLINE, lpMSFL->lpszParserName, strlen(kpszPN_NEWLINE) + 1))
        {
            /* See if we can load the file */
            bLoadFile = LoadFile(lpszDirAndFile, &kpMCB->lpMSPB->lpszMsgSetBuff,
                                 &lMsgSetBuffLen, &kpMCB->lpMSPB->lMsgSetDataLen, TRUE);
            if(FAILURE == bLoadFile)
            {
                LogMsg(137, lpszDirAndFile);
                return FAILURE;
            }

            /* Tell the world how large a file we loaded */
            LogMsg(156, lpszDirAndFile, kpMCB->lpMSPB->lMsgSetDataLen);

            /* Now parse it */
            bParseFile = ParseNewLineMsgSet();
        }
        else
        {
            LogMsg(152, lpMSFL->lpszParserName, lpszDirAndFile);
            return FAILURE;
        }

        /* Check the result */
        if(FAILURE == bParseFile)
        {
            LogMsg(138, lpszDirAndFile);
            return FAILURE;
        }
        else
        {
            /* So far so good. Tell the world */
            LogMsg(135, lpszDirAndFile);
        }

        /* Finished parsing the buffer so free it */
        if(kpMCB->lpMSPB->lpszMsgSetBuff)
        {
            FreeIfNotNull((void**)&kpMCB->lpMSPB->lpszMsgSetBuff);
            kpMCB->lpMSPB->lpszMsgSetBuff = NULL;
        }

        /* On to the next message set file load block */
        lpMSFL = (MsgSetFileLoadBlk*)lpMSFL->lpMSFLNext;
    }

    return SUCCESS;
}

/*==========================================================================*/
/* FUNCTION:                                                                */
/*     ParseMsgSet                                                          */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Parse message set information that has been loaded in XML format     */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     N/A                                                                  */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     SUCCESS or FAILURE                                                   */
/*==========================================================================*/
long
ParseMsgSet()
{
    /* Parser success of failure value */
    long  bMsgSetParse = SUCCESS;

    /* Create the parser */
    XML_Parser oParser = XML_ParserCreate(NULL);

    /* Give the parser the message set parse success flag pointer */
    /* for its callback functions                                 */
    XML_SetUserData(oParser, &bMsgSetParse);
    
    /* Provide the addresses of our various callback functions */
    XML_SetElementHandler(oParser, PMSParseStartElement, PMSParseEndElement);
    XML_SetCharacterDataHandler(oParser, PMSParseCharData);
    XML_SetCommentHandler(oParser, PMSComments);
    
    /* Call the parser */
    if(XML_STATUS_ERROR == XML_Parse(oParser, kpMCB->lpMSPB->lpszMsgSetBuff,
                                     kpMCB->lpMSPB->lMsgSetDataLen, 1))
    {
        /* EXPAT identified a problem during parsing */
        LogMsg(14, XML_ErrorString(XML_GetErrorCode(oParser)));
        return FAILURE;
    }

    /* Done with the parser so let it know */
    XML_ParserFree(oParser);

    if(FAILURE == bMsgSetParse)
    {
        /* MsgTest identified a problem during parsing */
        LogMsg(15);
        return FAILURE;
    }

    /* Parsed OK */
    return SUCCESS;
}

/*--------------------------------------------------------------------------*/
static void XMLCALL
PMSParseStartElement(void* vpUserData, const char* kpszElementName,
                     const char** kpkpAtts)
{
    MsgSetParseBlk*       lpMSPB;
    enum eMSLLBlockEnum   eMSLLBlockIs;
    char                  lpszElementNameBuff[NAME_BUFF_LEN + 4];
    long*                 lpbMsgSetParse;
    long                  lIdx;

    /* Get pointer to MsgSet parse success or failure value */
    lpbMsgSetParse = vpUserData;

    /* Get the message set parse block pointer */
    lpMSPB = kpMCB->lpMSPB;

    /* Already FAILED? Don't do anything else */
    if(FAILURE == *lpbMsgSetParse)
    {
        return;
    }

    /* New element so reset the ParseCharData buffer pointer */
    ClearPCDB();

    /* Check element name length */
    if(NAME_BUFF_LEN < StrlenNull(kpszElementName))
    {
        LogMsg(18, "Element", kpszElementName, NAME_BUFF_LEN);
        *lpbMsgSetParse = FAILURE;
        return;
    }

    /* Copy the element name to our local buffer */
    memset(lpszElementNameBuff, 0x00, NAME_BUFF_LEN + 4);
    strncpy(lpszElementNameBuff, kpszElementName, NAME_BUFF_LEN);

    /* Copy the element name to an element name buffer */
    memset(lpMSPB->lplpszElementName[lpMSPB->lMsgSetParseDepth], 0x00, NAME_BUFF_LEN + 4);
    strncpy(lpMSPB->lplpszElementName[lpMSPB->lMsgSetParseDepth], kpszElementName, NAME_BUFF_LEN);

    /* If we are inside a <Msg> tag inside the <Messages> section then */
    /* we need to set the eMsgType to P_UNKNOWN, increment the parse   */
    /* depth and check it, and add a new child node to the tree       */
    if((P_MESSAGES == lpMSPB->eMsgSetType[1]) &&
       (P_MSG == lpMSPB->eMsgSetType[2]))
    {
        /* Element type is unknown */
        lpMSPB->eMsgSetType[lpMSPB->lMsgSetParseDepth] = P_UNKNOWN;

        /* Increment the parse element depth */
        lpMSPB->lMsgSetParseDepth++;
    
        /* Check we haven't exceeded the maximum logical tag depth  */
        if(MAX_MSGSET_PARSE_DEPTH < lpMSPB->lMsgSetParseDepth)
        {
            lpMSPB->lMsgSetParseDepth = MAX_MSGSET_PARSE_DEPTH;
            PMSDisplayParseError(140);
            *lpbMsgSetParse = FAILURE;
            return;
        }
        
        /* We know we need to add a node. The NodeValue will come from the ParseCharDataBuff */
        /* during PMSParseEndElement processing and the NodeElementType will be determined   */
        /* during PrepareMsgSets processing                                                  */
        lpMSPB->lpCurrMSTN = AddLastChild(lpMSPB->lpCurrMSTN, N_UNKNOWN, kpszElementName, NULL, NULL);
        return;
    }

    /* Which element type are we processing for */
    eMSLLBlockIs = PMSGetElementType(lpszElementNameBuff);
    if(P_UNKNOWN == eMSLLBlockIs)
    {
        PMSDisplayParseError(139);
        *lpbMsgSetParse = FAILURE;
        return;
    }

    /* Now we know which element type it is */
    lpMSPB->eMsgSetType[lpMSPB->lMsgSetParseDepth] = eMSLLBlockIs;

    /* Clear all attributes before processing for this element */
    ClearAttrs();

    /* The <MsgSet> element itself can have attributes like a schema so ignore them */
    /* for anything else check that its one of the attributes we are expecting       */
    switch(eMSLLBlockIs)
    {
    case P_MSGSET:
        break;
    default:
        /* Process attribute pairs */
        lIdx = 0;
        while(kpkpAtts[lIdx])
        {
            /* Check element attr length */
            if(ATTR_BUFF_LEN < StrlenNull(kpkpAtts[lIdx + 1]))
            {
                LogMsg(18, kpkpAtts[lIdx], kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
                *lpbMsgSetParse = FAILURE;
                return;
            }

            if(0 == memcmp(kpszNAME, kpkpAtts[lIdx], strlen(kpszNAME) + 1))
            {
                /* Copy the Name attr to the current Name attr buffer */
                strncpy(kpMCB->lpszCurrNameAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszPOS, kpkpAtts[lIdx], strlen(kpszPOS) + 1))
            {
                /* Copy the Pos attr to the current Pos attr buffer */
                strncpy(kpMCB->lpszCurrPosAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszLEN, kpkpAtts[lIdx], strlen(kpszLEN) + 1))
            {
                /* Copy the Len attr to the current Len attr buffer */
                strncpy(kpMCB->lpszCurrLenAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszTYPE, kpkpAtts[lIdx], strlen(kpszTYPE) + 1))
            {
                /* Copy the Type attr to the current Type attr buffer */
                strncpy(kpMCB->lpszCurrTypeAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszOUTPUTFORMAT, kpkpAtts[lIdx], strlen(kpszOUTPUTFORMAT) + 1))
            {
                /* Copy the OutputFormat attr to the current OutputFormat attr buffer */
                strncpy(kpMCB->lpszCurrOutputFormatAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszMSGSEP, kpkpAtts[lIdx], strlen(kpszMSGSEP) + 1))
            {
                /* Copy the MsgSep attr to the current MsgSep attr buffer */
                strncpy(kpMCB->lpszCurrMsgSepAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszSEGSEP, kpkpAtts[lIdx], strlen(kpszSEGSEP) + 1))
            {
                /* Copy the SegSep attr to the current SegSep attr buffer */
                strncpy(kpMCB->lpszCurrSegSepAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszELEMENTSEP, kpkpAtts[lIdx], strlen(kpszELEMENTSEP) + 1))
            {
                /* Copy the ElementSep attr to the current ElementSep attr buffer */
                strncpy(kpMCB->lpszCurrElementSepAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszMINOCCURS, kpkpAtts[lIdx], strlen(kpszMINOCCURS) + 1))
            {
                /* Copy the MinOccurs attr to the current MinOccurs attr buffer */
                strncpy(kpMCB->lpszCurrMinOccursAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszMAXOCCURS, kpkpAtts[lIdx], strlen(kpszMAXOCCURS) + 1))
            {
                /* Copy the MaxOccurs attr to the current MaxOccurs attr buffer */
                strncpy(kpMCB->lpszCurrMaxOccursAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszCCSID, kpkpAtts[lIdx], strlen(kpszCCSID) + 1))
            {
                /* Copy the CCSID attr to the current CCSID attr buffer */
                strncpy(kpMCB->lpszCurrCCSIDAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszENCODING, kpkpAtts[lIdx], strlen(kpszENCODING) + 1))
            {
                /* Copy the Encoding attr to the current Encoding attr buffer */
                strncpy(kpMCB->lpszCurrEncodingAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszOFFSET, kpkpAtts[lIdx], strlen(kpszOFFSET) + 1))
            {
                /* Copy the Offset attr to the current Offset attr buffer */
                strncpy(kpMCB->lpszCurrOffsetAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszJUSTIFY, kpkpAtts[lIdx], strlen(kpszJUSTIFY) + 1))
            {
                /* Copy the Justify attr to the current Justify attr buffer */
                strncpy(kpMCB->lpszCurrJustifyAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszPADCHAR, kpkpAtts[lIdx], strlen(kpszPADCHAR) + 1))
            {
                /* Copy the PadChar attr to the current PadChar attr buffer */
                strncpy(kpMCB->lpszCurrPadCharAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszENDALIGN, kpkpAtts[lIdx], strlen(kpszENDALIGN) + 1))
            {
                /* Copy the EndAlign attr to the current EndAlign attr buffer */
                strncpy(kpMCB->lpszCurrEndAlignAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszCMQVALUES, kpkpAtts[lIdx], strlen(kpszCMQVALUES) + 1))
            {
                /* Copy the CmqValues attr to the current CmqValues attr buffer */
                strncpy(kpMCB->lpszCurrCmqValuesAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else if(0 == memcmp(kpszTRUNCATE, kpkpAtts[lIdx], strlen(kpszTRUNCATE) + 1))
            {
                /* Copy the Truncate attr to the current Truncate attr buffer */
                strncpy(kpMCB->lpszCurrTruncateAttr, kpkpAtts[lIdx + 1], ATTR_BUFF_LEN);
            }
            else
            {
                LogMsg(83, kpkpAtts[lIdx], kpkpAtts[lIdx + 1]);
                *lpbMsgSetParse = FAILURE;
                return;
            }

            /* Point to the next pair of attributes */
            lIdx += 2;
        }
        break;
    }

    /* Handle different initialisation requirements for elements */
    switch(eMSLLBlockIs)
    {
    case P_ELEMENT:
        lpMSPB->lpCurrELB = CreateElementBlk();
        break;
    case P_SEGDEF:
        if(NULL != lpMSPB->lpCurrMDB)
        {
            LogMsg(206);
            kpMCB->lProcessing = FAILURE;
        }
        else
        {
            if(NULL != lpMSPB->lpCurrSDB)
            {
                LogMsg(208);
                kpMCB->lProcessing = FAILURE;
            }
            else
            {
                lpMSPB->lpCurrSDB = CreateSegmentDefBlk();
            }
        }
        break;
    case P_SEGMENT:
        lpMSPB->lpCurrSGB = CreateSegmentBlk();
        break;
    case P_MSGDEF:
        if(NULL != lpMSPB->lpCurrSDB)
        {
            LogMsg(207);
            kpMCB->lProcessing = FAILURE;
        }
        else
        {
            if(NULL != lpMSPB->lpCurrMDB)
            {
                LogMsg(209);
                kpMCB->lProcessing = FAILURE;
            }
            else
            {
                lpMSPB->lpCurrMDB = CreateMsgDefBlk();
            }
        }
        break;
    case P_MSG:
        lpMSPB->lpCurrMSTN = AddLastChild(lpMSPB->lpCurrMSH->lpMSFRootNode, N_MSG, kpszMSG, kpMCB->lpszCurrTypeAttr, NULL);
        break;
    /* There is no default action for elements that dont need initialising */
    }

    /* Check to see if anything has gone fatally wrong */
    if(FAILURE == kpMCB->lProcessing)
    {
        *lpbMsgSetParse = FAILURE;
        return;
    }
    
    /* Increment the parse element depth */
    lpMSPB->lMsgSetParseDepth++;
    
    /* Check we haven't exceeded the maximum logical tag depth  */
    if(MAX_MSGSET_PARSE_DEPTH < lpMSPB->lMsgSetParseDepth)
    {
        lpMSPB->lMsgSetParseDepth = MAX_MSGSET_PARSE_DEPTH;
        PMSDisplayParseError(140);
        *lpbMsgSetParse = FAILURE;
        return;
    }
}

/*--------------------------------------------------------------------------*/
static void XMLCALL
PMSParseEndElement(void* vpUserData, const char* kpszElementName)
{
    MsgSetParseBlk*       lpMSPB;
    enum eMSLLBlockEnum   eMSLLBlockIs;
    long*                 lpbMsgSetParse;
    ElementBlk*           lpCheckELB;
    SegmentDefBlk*        lpCheckSDB;
    SegmentBlk*           lpCheckSGB;
    MsgDefBlk*            lpCheckMDB;
    char                  lpszElementNameBuff[NAME_BUFF_LEN + 4];

    /* Get pointer to MsgSet parse success or failure value */
    lpbMsgSetParse = vpUserData;

    /* Get the test info and secondary parse control block pointers */
    lpMSPB = kpMCB->lpMSPB;

    /* Already FAILED? Don't do anything else */
    if(FAILURE == *lpbMsgSetParse)
    {
        return;
    }

    /* Copy the element name to our local buffer */
    memset(lpszElementNameBuff, 0x00, NAME_BUFF_LEN + 4);
    strncpy(lpszElementNameBuff, kpszElementName, NAME_BUFF_LEN);

    /* Decrement the parse element depth */
    lpMSPB->lMsgSetParseDepth--;
    if(0 > lpMSPB->lMsgSetParseDepth)
    {
        lpMSPB->lMsgSetParseDepth = 0;
        LogMsg(17);
        *lpbMsgSetParse = FAILURE;
        return;
    }

    /* If we are inside a <Msg> tag inside the <Messages> section then  */
    /* we need to clear the element name and type value and copy the    */
    /* ParseCharDataBuff value to the current tree node and then reset  */
    /* the current tree node pointer to its parent                      */
    if((P_MESSAGES == lpMSPB->eMsgSetType[1]) &&
       (P_MSG == lpMSPB->eMsgSetType[2]))
    {
        memset(lpMSPB->lplpszElementName[lpMSPB->lMsgSetParseDepth], 0x00, NAME_BUFF_LEN + 4);
        lpMSPB->eMsgSetType[lpMSPB->lMsgSetParseDepth] = P_UNKNOWN;
        /* Node value for <Msg> nodes is set to the Type="" attribute   */
        /* value during PMSParseStartElement processing                 */
        if(0 != memcmp(kpszMSG, kpszElementName, strlen(kpszMSG) + 1))
        {
            lpMSPB->lpCurrMSTN->lpszNodeValue = CloneStringNull(kpMCB->lpszParseCharDataBuff);
        }
        ClearPCDB();
        /* Change pointer to the parent */
        lpMSPB->lpCurrMSTN = (MsgSetTreeNode*)lpMSPB->lpCurrMSTN->lpParent;
        return;
    }

    /* Which element type are we processing for */
    eMSLLBlockIs = PMSGetElementType(lpszElementNameBuff);
    if(P_UNKNOWN == eMSLLBlockIs)
    {
        PMSDisplayParseError(139);
        *lpbMsgSetParse = FAILURE;
        return;
    }

    /* Element processing varies depending on parse depth   */
    switch(lpMSPB->lMsgSetParseDepth)
    {
    /* At parse depth 0 <MSGSET> is the only valid element  */
    case 0:
        switch(eMSLLBlockIs)
        {
        case P_MSGSET:
            break;
        default:
            LogMsg(20, lpszElementNameBuff, kpszMESSAGESET);
            *lpbMsgSetParse = FAILURE;
            return;
        }
        break;
    /* At parse depth 1 <Definitons>, and <Messages> are  */
    /* the only valid elements and only need to exist     */
    case 1:
        switch(eMSLLBlockIs)
        {
        case P_MSGTEMPLATES:
        case P_MESSAGES:
            break;
        default:
            LogMsg(21, lpszElementNameBuff, kpszMESSAGESET);
            *lpbMsgSetParse = FAILURE;
            return;
        }
        break;
    /* At parse depth 2 and beyond processing is dependent on */
    /* which of the parse depth 1 elements we are operating   */
    /* under                                                  */
    case 2:
    case 3:
        switch(lpMSPB->eMsgSetType[1])
        {
        case P_MSGTEMPLATES:
            /* Definitons have a maximum logically reachable parse depth of 3 */
            if(3 < lpMSPB->lMsgSetParseDepth)
            {
                PMSDisplayParseError(139);
                LogMsg(4, "Invalid parse depth", lpszElementNameBuff, __LINE__, __FILE__);
                *lpbMsgSetParse = FAILURE;
            }
            switch(eMSLLBlockIs)
            {
            case P_ELEMENT:
                /* Set the default value for the current element if available */
                lpMSPB->lpCurrELB->lpszDefaultValue = CloneStringNull(kpMCB->lpszParseCharDataBuff);
                /* Are we adding this element to a message definition or a segment definiton */
                if(lpMSPB->lpCurrMDB)
                {
                    /* Elements with duplicate types are not permitted */
                    lpCheckELB = (ElementBlk*)FindMsgSetLinkedListBlk((MsgSetLinkedListBlk*)lpMSPB->lpCurrMDB->lpMSLLHead,
                                                                      lpMSPB->lpCurrELB->lpszName,
                                                                      lpMSPB->lpCurrELB->lNameLen);
                    if(lpCheckELB)
                    {
                        LogMsg(142, lpMSPB->lpCurrELB->lpszName, kpszMSGDEF, lpMSPB->lpCurrMDB->lpszType);
                        *lpbMsgSetParse = FAILURE;
                        return;
                    }
                    /* Seems ok so add the Element to the MsgDef */
                    AddToTailOfList((LinkedListBlk**)&lpMSPB->lpCurrMDB->lpMSLLHead,
                                    (LinkedListBlk**)&lpMSPB->lpCurrMDB->lpMSLLTail,
                                    (LinkedListBlk*)lpMSPB->lpCurrELB);
                }
                else if(lpMSPB->lpCurrSDB)
                {
                    /* Elements with duplicate types are not permitted */
                    lpCheckELB = (ElementBlk*)FindMsgSetLinkedListBlk(lpMSPB->lpCurrSDB->lpMSLLHead,
                                                                      lpMSPB->lpCurrELB->lpszName,
                                                                      lpMSPB->lpCurrELB->lNameLen);
                    if(lpCheckELB)
                    {
                        LogMsg(142, lpMSPB->lpCurrELB->lpszName, kpszSEGDEF, lpMSPB->lpCurrSDB->lpszType);
                        *lpbMsgSetParse = FAILURE;
                        return;
                    }
                    /* Seems ok so add the Element to the SegDef */
                    AddToTailOfList((LinkedListBlk**)&lpMSPB->lpCurrSDB->lpMSLLHead,
                                    (LinkedListBlk**)&lpMSPB->lpCurrSDB->lpMSLLTail,
                                    (LinkedListBlk*)lpMSPB->lpCurrELB);
                }
                else
                {
                    /* Elements must be used within message definitions or segment definitions */
                    LogMsg(141, lpMSPB->lpCurrELB->lpszName);
                    *lpbMsgSetParse = FAILURE;
                    return;
                }
                lpMSPB->lpCurrELB = NULL;
                break;
            case P_SEGDEF:
                /* Segment Definitions with duplicate types are not permitted */
                lpCheckSDB = (SegmentDefBlk*)FindMsgSetLinkedListBlk((MsgSetLinkedListBlk*)kpMCB->lpMSPB->lpCurrMSH->lpSDBHead,
                                                                     lpMSPB->lpCurrSDB->lpszType,
                                                                     lpMSPB->lpCurrSDB->lTypeLen);
                if(lpCheckSDB)
                {
                    LogMsg(143, lpMSPB->lpCurrSDB->lpszType);
                    *lpbMsgSetParse = FAILURE;
                    return;
                }
                /* Seems ok so add the segment definition to the chain */
                AddToTailOfList((LinkedListBlk**)&kpMCB->lpMSPB->lpCurrMSH->lpSDBHead,
                                (LinkedListBlk**)&kpMCB->lpMSPB->lpCurrMSH->lpSDBTail,
                                (LinkedListBlk*)lpMSPB->lpCurrSDB);
                lpMSPB->lpCurrSDB = NULL;
                break;
            case P_SEGMENT:
                /* Are we adding this element/segment to a message definition or a segment definiton */
                if(lpMSPB->lpCurrMDB)
                {
                    /* Segments with duplicate types are not permitted */
                    lpCheckSGB = (SegmentBlk*)FindMsgSetLinkedListBlk((MsgSetLinkedListBlk*)lpMSPB->lpCurrMDB->lpMSLLHead,
                                                                      lpMSPB->lpCurrSGB->lpszType,
                                                                      lpMSPB->lpCurrSGB->lTypeLen);
                    if(lpCheckSGB)
                    {
                        LogMsg(145, lpMSPB->lpCurrSGB->lpszType, kpszMSGDEF, lpMSPB->lpCurrMDB->lpszType);
                        *lpbMsgSetParse = FAILURE;
                        return;
                    }
                    /* Seems ok so add the Segment to the MsgDef */
                    AddToTailOfList((LinkedListBlk**)&lpMSPB->lpCurrMDB->lpMSLLHead,
                                    (LinkedListBlk**)&lpMSPB->lpCurrMDB->lpMSLLTail,
                                    (LinkedListBlk*)lpMSPB->lpCurrSGB);
                }
                else if(lpMSPB->lpCurrSDB)
                {
                    /* Segments with duplicate types are not permitted */
                    lpCheckSGB = (SegmentBlk*)FindMsgSetLinkedListBlk(lpMSPB->lpCurrSDB->lpMSLLHead,
                                                                      lpMSPB->lpCurrSGB->lpszType,
                                                                      lpMSPB->lpCurrSGB->lTypeLen);
                    if(lpCheckSGB)
                    {
                        LogMsg(145, lpMSPB->lpCurrSGB->lpszType, kpszSEGDEF, lpMSPB->lpCurrSDB->lpszType);
                        *lpbMsgSetParse = FAILURE;
                        return;
                    }
                    /* Seems ok so add the Segment to the SegDef */
                    AddToTailOfList((LinkedListBlk**)&lpMSPB->lpCurrSDB->lpMSLLHead,
                                    (LinkedListBlk**)&lpMSPB->lpCurrSDB->lpMSLLTail,
                                    (LinkedListBlk*)lpMSPB->lpCurrSGB);
                }
                else
                {
                    /* Segments must be used within message definitions or segment definitions */
                    LogMsg(144, lpMSPB->lpCurrSGB->lpszType);
                    *lpbMsgSetParse = FAILURE;
                    return;
                }
                lpMSPB->lpCurrSGB = NULL;
                break;
            case P_MSGDEF:
                /* MessageTemplates with duplicate types are not permitted */
                lpCheckMDB = (MsgDefBlk*)FindMsgSetLinkedListBlk((MsgSetLinkedListBlk*)kpMCB->lpMSPB->lpCurrMSH->lpMDBHead,
                                                                 lpMSPB->lpCurrMDB->lpszType,
                                                                 lpMSPB->lpCurrMDB->lTypeLen);
                if(lpCheckMDB)
                {
                    LogMsg(146, lpMSPB->lpCurrMDB->lpszType);
                    *lpbMsgSetParse = FAILURE;
                    return;
                }
                /* Seems ok so add the Message definition to the chain */
                AddToTailOfList((LinkedListBlk**)&kpMCB->lpMSPB->lpCurrMSH->lpMDBHead,
                                (LinkedListBlk**)&kpMCB->lpMSPB->lpCurrMSH->lpMDBTail,
                                (LinkedListBlk*)lpMSPB->lpCurrMDB);
                lpMSPB->lpCurrMDB = NULL;
                break;
            default:
                PMSDisplayParseError(139);
                *lpbMsgSetParse = FAILURE;
                return;
            }
            break;
        default:
            LogMsg(4, "Unexpected element type not <MessageTemplates>", lpszElementNameBuff, __LINE__, __FILE__);
            break;
        }
        break;
    default:
        PMSDisplayParseError(139);
        LogMsg(4, "Invalid parse depth", lpszElementNameBuff, __LINE__, __FILE__);
        *lpbMsgSetParse = FAILURE;
        return;
    }

    /* Reset the ParseCharData buffer */
    ClearPCDB();

    /* Clear the element name buffer and element type entries */
    if(0 <= lpMSPB->lMsgSetParseDepth)
    {
        memset(lpMSPB->lplpszElementName[lpMSPB->lMsgSetParseDepth], 0x00, NAME_BUFF_LEN + 4);
        lpMSPB->eMsgSetType[lpMSPB->lMsgSetParseDepth] = P_UNKNOWN;
    }
}

/*--------------------------------------------------------------------------*/
static void XMLCALL
PMSParseCharData(void* vpUserData, const XML_Char* kpszCD, int lLen)
{
    long            lPCDBuffLen;
    long*           lpbMsgSetParse;

    /* Get pointer to MsgSet parse success or failure value */
    lpbMsgSetParse = vpUserData;

    /* Already FAILED? Don't do anything else */
    if(FAILURE == *lpbMsgSetParse)
    {
        return;
    }

    /* ???TA Need to handle multiple calls as according to doco */
    /* ???TA we may not get everything at once                  */

    /* Got some data so allocate and initialise an area to copy it to */
    lPCDBuffLen = (((lLen + 8) / 8) + 1) * 8;
    kpMCB->lParseCharDataLen = lLen;
    kpMCB->lpszParseCharDataBuff = malloc(lPCDBuffLen);
    if(! kpMCB->lpszParseCharDataBuff)
    {
        kpMCB->lParseCharDataLen = 0;
        kpMCB->lpszParseCharDataBuff = NULL;
        LogMsg(3, lPCDBuffLen, "ParseCharData", __LINE__, __FILE__);
        *lpbMsgSetParse = FAILURE;
        return;
    }
    memset(kpMCB->lpszParseCharDataBuff, 0x00, lPCDBuffLen);

    /* Copy the data contained in the element to our new buffer */
    memcpy(kpMCB->lpszParseCharDataBuff, kpszCD, lLen);
}

/*--------------------------------------------------------------------------*/
static void XMLCALL
PMSComments(void* vpUserData, const XML_Char* kpszComment)
{
    /* Don't do anything with comments at the moment */

    /* Pretend to do something so dont get level 4 warnings */
    /* about unreferenced formal parameters                 */
    void*  vpUD;
    char*  kpszC;
    
    vpUD = vpUserData;
    kpszC = (char*)kpszComment;
}

/*--------------------------------------------------------------------------*/
void
PMSDisplayParseError(long lMsgNum)
{
    long  lIdx;
    char  lpszExtraInfo[EXTRA_MSG_INFO_LEN + 4];
    char* lpszExtraBuff;
    
    lpszExtraBuff = lpszExtraInfo;
    memset(lpszExtraBuff, 0x00, EXTRA_MSG_INFO_LEN + 4);

    for(lIdx = 0; lIdx <= kpMCB->lpMSPB->lMsgSetParseDepth; lIdx++)
    {
        sprintf(lpszExtraBuff, "%s.", kpMCB->lpMSPB->lplpszElementName[lIdx]);
        lpszExtraBuff += StrlenNull(kpMCB->lpMSPB->lplpszElementName[lIdx]) + 1;
    }

    LogMsg(lMsgNum, lpszExtraInfo);
}

/*--------------------------------------------------------------------------*/
enum eMSLLBlockEnum
PMSGetElementType(const char *kpszElementName)
{
    enum eMSLLBlockEnum eMSLLBlockIs;

    /****** Ok which element is it ? ******/
    if(0 == memcmp(kpszMSGSET, kpszElementName, strlen(kpszMSGSET) + 1))
    {
        eMSLLBlockIs = P_MSGSET;
    }
    else if(0 == memcmp(kpszMSGTEMPLATES, kpszElementName, strlen(kpszMSGTEMPLATES) + 1))
    {
        eMSLLBlockIs = P_MSGTEMPLATES;
    }
    else if(0 == memcmp(kpszMESSAGES, kpszElementName, strlen(kpszMESSAGES) + 1))
    {
        eMSLLBlockIs = P_MESSAGES;
    }
    else if(0 == memcmp(kpszMSG, kpszElementName, strlen(kpszMSG) + 1))
    {
        eMSLLBlockIs = P_MSG;
    }
    else if(0 == memcmp(kpszELEMENT, kpszElementName, strlen(kpszELEMENT) + 1))
    {
        eMSLLBlockIs = P_ELEMENT;
    }
    else if(0 == memcmp(kpszSEGDEF, kpszElementName, strlen(kpszSEGDEF) + 1))
    {
        eMSLLBlockIs = P_SEGDEF;
    }
    else if(0 == memcmp(kpszMSGDEF, kpszElementName, strlen(kpszMSGDEF) + 1))
    {
        eMSLLBlockIs = P_MSGDEF;
    }
    else if(0 == memcmp(kpszSEGMENT, kpszElementName, strlen(kpszSEGMENT) + 1))
    {
        eMSLLBlockIs = P_SEGMENT;
    }
    else
    {
        eMSLLBlockIs = P_UNKNOWN;
    }
    return eMSLLBlockIs;
}

/*--------------------------------------------------------------------------*/
ElementBlk*
CreateElementBlk()
{
    ElementBlk* lpELB;

    lpELB = malloc(sizeof(ElementBlk));
    if(lpELB)
    {
        memset(lpELB, 0x00, sizeof(ElementBlk));
        lpELB->lpELBNext = NULL;
        lpELB->lpELBPrev = NULL;
        lpELB->eMSLLBlockType = P_ELEMENT;
        lpELB->lpszName = CloneString(kpMCB->lpszCurrNameAttr);
        lpELB->lNameLen = StrlenNull(lpELB->lpszName);
        lpELB->lpszElementType = CloneString(kpMCB->lpszCurrTypeAttr);
        lpELB->eElementType = T_UNKNOWN;
        lpELB->lpszElementSep = CloneStringNull(kpMCB->lpszCurrElementSepAttr);
        lpELB->lpszLen = CloneStringNull(kpMCB->lpszCurrLenAttr);
        if(0 == StrlenNull(lpELB->lpszLen))
        {
            lpELB->lpszLen = CloneString(kpszMINUSONE);
        }
        lpELB->lpszPos = CloneStringNull(kpMCB->lpszCurrPosAttr);
        if(0 == StrlenNull(lpELB->lpszPos))
        {
            lpELB->lpszPos = CloneString(kpszMINUSONE);
        }
        lpELB->lpszOffset = CloneStringNull(kpMCB->lpszCurrOffsetAttr);
        if(0 == StrlenNull(lpELB->lpszOffset))
        {
            lpELB->lpszOffset = CloneString(kpszMINUSONE);
        }
        lpELB->lpszJustify = CloneStringNull(kpMCB->lpszCurrJustifyAttr);
        if(0 == StrlenNull(lpELB->lpszJustify))
        {
            lpELB->lpszJustify = CloneString(kpszLEFT);
        }
        if((0 != memcmp(kpszLEFT, lpELB->lpszJustify, strlen(kpszLEFT) + 1)) &&
           (0 != memcmp(kpszRIGHT, lpELB->lpszJustify, strlen(kpszRIGHT) + 1)))
        {
            LogMsg(230, lpELB->lpszName, lpELB->lpszJustify);
            kpMCB->lProcessing = FAILURE;
            FreeELB(&lpELB);
            return NULL;
        }
        lpELB->lpszPadChar = CloneStringNull(kpMCB->lpszCurrPadCharAttr);
        lpELB->lpszEndAlign = CloneStringNull(kpMCB->lpszCurrEndAlignAttr);
        if(0 == StrlenNull(lpELB->lpszEndAlign))
        {
            lpELB->lpszEndAlign = CloneString(kpszONE);
        }
        lpELB->bCmqValues = FALSE;
        if(0 == memcmp(kpszYES, kpMCB->lpszCurrCmqValuesAttr, strlen(kpszYES) + 1))
        {
            lpELB->bCmqValues = TRUE;
        }
        lpELB->lpszMinOccurs = CloneStringNull(kpMCB->lpszCurrMinOccursAttr);
        if(0 == StrlenNull(lpELB->lpszMinOccurs))
        {
            lpELB->lpszMinOccurs = CloneString(kpszONE);
        }
        lpELB->lpszMaxOccurs = CloneStringNull(kpMCB->lpszCurrMaxOccursAttr);
        if(0 == StrlenNull(lpELB->lpszMaxOccurs))
        {
            lpELB->lpszMaxOccurs = CloneString(kpszONE);
        }
        else
        {
            if(0 == memcmp(kpszUNBOUNDED, lpELB->lpszMaxOccurs, strlen(kpszUNBOUNDED) + 1))
            {
                lpELB->lpszMaxOccurs = CloneString(kpszMINUSONE);
            }
        }
        lpELB->lpszDefaultValue = NULL;

        /* Convert values */
        lpELB->lLen = ConvNum(NULL, lpELB->lpszLen, kpszLEN);
        lpELB->lPos = ConvNum(NULL, lpELB->lpszPos, kpszPOS);
        lpELB->lOffset = ConvNum(NULL, lpELB->lpszOffset, kpszOFFSET);
        lpELB->lEndAlign = ConvNum(NULL, lpELB->lpszEndAlign, kpszENDALIGN);
        lpELB->lMinOccurs = ConvNum(NULL, lpELB->lpszMinOccurs, kpszMINOCCURS);
        lpELB->lMaxOccurs = ConvNum(NULL, lpELB->lpszMaxOccurs, kpszMAXOCCURS);
        if(FAILURE == kpMCB->lProcessing)
        {
            FreeELB(&lpELB);
            return NULL;
        }

        /* Determine the element type */
        if(0 == memcmp(kpszCHAR, lpELB->lpszElementType, strlen(kpszCHAR) + 1))
        {
            lpELB->eElementType = T_CHAR;
        }
        else if(0 == memcmp(kpszHEX, lpELB->lpszElementType, strlen(kpszHEX) + 1))
        {
            lpELB->eElementType = T_HEX;
        }
        else if(0 == memcmp(kpszINT8, lpELB->lpszElementType, strlen(kpszINT8) + 1))
        {
            lpELB->eElementType = T_INT8;
        }
        else if(0 == memcmp(kpszINT16, lpELB->lpszElementType, strlen(kpszINT16) + 1))
        {
            lpELB->eElementType = T_INT16;
        }
        else if(0 == memcmp(kpszINT32, lpELB->lpszElementType, strlen(kpszINT32) + 1))
        {
            lpELB->eElementType = T_INT32;
        }
        else if(0 == memcmp(kpszINT64, lpELB->lpszElementType, strlen(kpszINT64) + 1))
        {
            lpELB->eElementType = T_INT64;
        }
        else if(0 == memcmp(kpszPD, lpELB->lpszElementType, strlen(kpszPD) + 1))
        {
            lpELB->eElementType = T_PD;
        }
        else
        {
            LogMsg(155, lpELB->lpszName, lpELB->lpszElementType);
            kpMCB->lProcessing = FAILURE;
            FreeELB(&lpELB);
            return NULL;
        }
        /* For fields that can have a derived length it is illegal to specify CmqValues="Yes" */
        if((T_CHAR == lpELB->eElementType) ||
           (T_HEX == lpELB->eElementType) ||
           (T_PD == lpELB->eElementType))
        {
            if((klUNDEFINED == lpELB->lLen) &&
               (TRUE == lpELB->bCmqValues))
            {
                LogMsg(237, lpELB->lpszName);
                kpMCB->lProcessing = FAILURE;
                FreeELB(&lpELB);
                return NULL;
            }
        }

    }
    else
    {
        ConsoleMsg(3, sizeof(ElementBlk), "ElementBlk", __LINE__, __FILE__);
        kpMCB->lProcessing = FAILURE;
        return NULL;
    }

    return lpELB;
}

/*--------------------------------------------------------------------------*/
SegmentDefBlk*
CreateSegmentDefBlk()
{
    SegmentDefBlk* lpSDB;

    lpSDB = malloc(sizeof(SegmentDefBlk));
    if(lpSDB)
    {
        memset(lpSDB, 0x00, sizeof(SegmentDefBlk));
        lpSDB->lpSDBNext = NULL;
        lpSDB->lpSDBPrev = NULL;
        lpSDB->eMSLLBlockType = P_SEGDEF;
        lpSDB->lpszType = CloneString(kpMCB->lpszCurrTypeAttr);
        lpSDB->lTypeLen = StrlenNull(lpSDB->lpszType);
        lpSDB->lpszOutputFormat = CloneStringNull(kpMCB->lpszCurrOutputFormatAttr);
        lpSDB->eOutputFormat = O_UNKNOWN;
        lpSDB->lpszElementSep = CloneStringNull(kpMCB->lpszCurrElementSepAttr);
        lpSDB->lpszSegSep = CloneStringNull(kpMCB->lpszCurrSegSepAttr);
        lpSDB->lpszCCSID = CloneStringNull(kpMCB->lpszCurrCCSIDAttr);
        lpSDB->lpszEncoding = CloneStringNull(kpMCB->lpszCurrEncodingAttr);
        lpSDB->lpMSLLHead = NULL;
        lpSDB->lpMSLLTail = NULL;
    }
    else
    {
        ConsoleMsg(3, sizeof(SegmentDefBlk), "SegmentDefBlk", __LINE__, __FILE__);
        kpMCB->lProcessing = FAILURE;
        return NULL;
    }

    return lpSDB;
}

/*--------------------------------------------------------------------------*/
SegmentBlk*
CreateSegmentBlk()
{
    SegmentBlk* lpSGB;

    lpSGB = malloc(sizeof(SegmentBlk));
    if(lpSGB)
    {
        memset(lpSGB, 0x00, sizeof(SegmentBlk));
        lpSGB->lpSGBNext = NULL;
        lpSGB->lpSGBPrev = NULL;
        lpSGB->eMSLLBlockType = P_SEGMENT;
        lpSGB->lpszType = CloneString(kpMCB->lpszCurrTypeAttr);
        lpSGB->lTypeLen = StrlenNull(lpSGB->lpszType);
        lpSGB->lpszOutputFormat = CloneStringNull(kpMCB->lpszCurrOutputFormatAttr);
        lpSGB->eOutputFormat = O_UNKNOWN;
        lpSGB->lpszElementSep = CloneStringNull(kpMCB->lpszCurrElementSepAttr);
        lpSGB->lpszSegSep = CloneStringNull(kpMCB->lpszCurrSegSepAttr);
        lpSGB->lpszMinOccurs = CloneStringNull(kpMCB->lpszCurrMinOccursAttr);
        if(0 == StrlenNull(lpSGB->lpszMinOccurs))
        {
            lpSGB->lpszMinOccurs = CloneString(kpszONE);
        }
        lpSGB->lpszMaxOccurs = CloneStringNull(kpMCB->lpszCurrMaxOccursAttr);
        if(0 == StrlenNull(lpSGB->lpszMaxOccurs))
        {
            lpSGB->lpszMaxOccurs = CloneString(kpszONE);
        }
        else
        {
            if(0 == memcmp(kpszUNBOUNDED, lpSGB->lpszMaxOccurs, strlen(kpszUNBOUNDED) + 1))
            {
                lpSGB->lpszMaxOccurs = CloneString(kpszMINUSONE);
            }
        }
        lpSGB->lpszCCSID = CloneStringNull(kpMCB->lpszCurrCCSIDAttr);
        lpSGB->lpszEncoding = CloneStringNull(kpMCB->lpszCurrEncodingAttr);
        lpSGB->lpSDB = NULL;

        /* Convert values */
        lpSGB->lMinOccurs = ConvNum(NULL, lpSGB->lpszMinOccurs, kpszMINOCCURS);
        lpSGB->lMaxOccurs = ConvNum(NULL, lpSGB->lpszMaxOccurs, kpszMAXOCCURS);
    }
    else
    {
        ConsoleMsg(3, sizeof(SegmentBlk), "SegmentBlk", __LINE__, __FILE__);
        kpMCB->lProcessing = FAILURE;
        return NULL;
    }

    return lpSGB;
}

/*--------------------------------------------------------------------------*/
MsgDefBlk*
CreateMsgDefBlk()
{
    MsgDefBlk* lpMDB;

    lpMDB = malloc(sizeof(MsgDefBlk));
    if(lpMDB)
    {
        memset(lpMDB, 0x00, sizeof(MsgDefBlk));
        lpMDB->lpMDBNext = NULL;
        lpMDB->lpMDBPrev = NULL;
        lpMDB->eMSLLBlockType = P_MSGDEF;
        lpMDB->lpszType = CloneString(kpMCB->lpszCurrTypeAttr);
        lpMDB->lTypeLen = StrlenNull(lpMDB->lpszType);
        lpMDB->lpszOutputFormat = CloneStringNull(kpMCB->lpszCurrOutputFormatAttr);
        lpMDB->eOutputFormat = O_UNKNOWN;
        lpMDB->lpszElementSep = CloneStringNull(kpMCB->lpszCurrElementSepAttr);
        lpMDB->lpszSegSep = CloneStringNull(kpMCB->lpszCurrSegSepAttr);
        lpMDB->lpszMsgSep = CloneStringNull(kpMCB->lpszCurrMsgSepAttr);
        lpMDB->bTruncate = FALSE;
        if(0 == memcmp(kpszYES, kpMCB->lpszCurrTruncateAttr, strlen(kpszYES) + 1))
        {
            lpMDB->bTruncate = TRUE;
        }
        lpMDB->lpszCCSID = CloneStringNull(kpMCB->lpszCurrCCSIDAttr);
        lpMDB->lpszEncoding = CloneStringNull(kpMCB->lpszCurrEncodingAttr);
        lpMDB->lpMSLLHead = NULL;
        lpMDB->lpMSLLTail = NULL;
    }
    else
    {
        ConsoleMsg(3, sizeof(MsgDefBlk), "MsgDefBlk", __LINE__, __FILE__);
        kpMCB->lProcessing = FAILURE;
        return NULL;
    }

    return lpMDB;
}

/*--------------------------------------------------------------------------*/
MsgSetTreeNode*
CreateMsgSetTreeNode(enum eMsgSetTreeNodeEnum eMsgSetTreeNodeType,
                     const char* const kpszNodeName, const char* const kpszNodeValue,
                     MsgSetLinkedListBlk* lpMSLL)
{
    MsgSetTreeNode* lpMSTN;

    lpMSTN = malloc(sizeof(MsgSetTreeNode));
    if(lpMSTN)
    {
        memset(lpMSTN, 0x00, sizeof(MsgSetTreeNode));
        lpMSTN->lpParent = NULL;
        lpMSTN->lpNextSibling = NULL;
        lpMSTN->lpPrevSibling = NULL;
        lpMSTN->lpFirstChild = NULL;
        lpMSTN->lpLastChild = NULL;
        lpMSTN->lpNextSiblingSameName = NULL;
        lpMSTN->lpPrevSiblingSameName = NULL;
        lpMSTN->lIndexSameName = 0;
        lpMSTN->eMsgSetTreeNodeType = eMsgSetTreeNodeType;
        lpMSTN->lpszNodeName = CloneStringNull(kpszNodeName);
        lpMSTN->lNodeNameLen = StrlenNull(lpMSTN->lpszNodeName);
        lpMSTN->lpszNodeValue = CloneStringNull(kpszNodeValue);
        lpMSTN->lNodeValueLen = StrlenNull(lpMSTN->lpszNodeValue);
        lpMSTN->lNodeDepth = -1;
        lpMSTN->lpMDB = NULL;
        lpMSTN->lpSDB = NULL;
        lpMSTN->lpSGB = NULL;
        lpMSTN->lpMSLL = lpMSLL;
        lpMSTN->lpNextRequiredMSLL = NULL;
        lpMSTN->lFlags = MSTN_NONE;
        lpMSTN->lOutputDataLen = 0;
        lpMSTN->lpszBuffLoc = NULL;
    }
    else
    {
        ConsoleMsg(3, sizeof(MsgSetTreeNode), "MsgSetTreeNode", __LINE__, __FILE__);
        kpMCB->lProcessing = FAILURE;
        return NULL;
    }

    return lpMSTN;
}

/*--------------------------------------------------------------------------*/
MsgSetTreeNode*
AddLastChild(MsgSetTreeNode* lpTNParent,
             enum eMsgSetTreeNodeEnum eMsgSetTreeNodeType,
             const char* const kpszNodeName,
             const char* const kpszNodeValue,
             MsgSetLinkedListBlk* lpMSLL)
{
    MsgSetTreeNode* lpTNNew;
    MsgSetTreeNode* lpTNSibling;

    lpTNNew = CreateMsgSetTreeNode(eMsgSetTreeNodeType, kpszNodeName,
                                     kpszNodeValue, lpMSLL);
    if(lpTNNew)
    {
        lpTNNew->lpParent = (struct MsgSetTreeNode*)lpTNParent;
        lpTNSibling = (MsgSetTreeNode*)lpTNParent->lpLastChild;
        if(NULL == lpTNSibling)
        {
            /* No sibling so just make ourselves the only child */
            lpTNParent->lpFirstChild = (struct MsgSetTreeNode*)lpTNNew;
            lpTNParent->lpLastChild = (struct MsgSetTreeNode*)lpTNNew;
        }
        else
        {
            /* Make ourselves the last child */
            lpTNParent->lpLastChild = (struct MsgSetTreeNode*)lpTNNew;
            lpTNSibling->lpNextSibling = (struct MsgSetTreeNode*)lpTNNew;
            lpTNNew->lpPrevSibling = (struct MsgSetTreeNode*)lpTNSibling;
        }
    }

    return lpTNNew;
}


/*--------------------------------------------------------------------------*/
MsgSetTreeNode*
AddFirstChild(MsgSetTreeNode* lpTNParent,
             enum eMsgSetTreeNodeEnum eMsgSetTreeNodeType,
             const char* const kpszNodeName,
             const char* const kpszNodeValue,
             MsgSetLinkedListBlk* lpMSLL)
{
    MsgSetTreeNode* lpTNNew;
    MsgSetTreeNode* lpTNSibling;

    lpTNNew = CreateMsgSetTreeNode(eMsgSetTreeNodeType, kpszNodeName,
                                     kpszNodeValue, lpMSLL);
    if(lpTNNew)
    {
        lpTNNew->lpParent = (struct MsgSetTreeNode*)lpTNParent;
        lpTNSibling = (MsgSetTreeNode*)lpTNParent->lpFirstChild;
        if(NULL == lpTNSibling)
        {
            /* No sibling so just make ourselves the only child */
            lpTNParent->lpFirstChild = (struct MsgSetTreeNode*)lpTNNew;
            lpTNParent->lpLastChild = (struct MsgSetTreeNode*)lpTNNew;
        }
        else
        {
            /* Make ourselves the first child */
            lpTNParent->lpFirstChild = (struct MsgSetTreeNode*)lpTNNew;
            lpTNSibling->lpPrevSibling = (struct MsgSetTreeNode*)lpTNNew;
            lpTNNew->lpNextSibling = (struct MsgSetTreeNode*)lpTNSibling;
        }
    }

    return lpTNNew;
}


/*--------------------------------------------------------------------------*/
MsgSetTreeNode*
AddPrevSibling(MsgSetTreeNode* lpTNCurr,
             enum eMsgSetTreeNodeEnum eMsgSetTreeNodeType,
             const char* const kpszNodeName,
             const char* const kpszNodeValue,
             MsgSetLinkedListBlk* lpMSLL)
{
    MsgSetTreeNode* lpTNNew;
    MsgSetTreeNode* lpTNParent;
    MsgSetTreeNode* lpTNPrevSibling;

    lpTNNew = CreateMsgSetTreeNode(eMsgSetTreeNodeType, kpszNodeName,
                                     kpszNodeValue, lpMSLL);
    if(lpTNNew)
    {
        lpTNParent = (MsgSetTreeNode*)lpTNCurr->lpParent;
        lpTNNew->lpParent = lpTNCurr->lpParent;
        lpTNPrevSibling = (MsgSetTreeNode*)lpTNCurr->lpPrevSibling;
        if(NULL == lpTNPrevSibling)
        {
            /* No sibling so just make ourselves the first child */
            lpTNParent->lpFirstChild = (struct MsgSetTreeNode*)lpTNNew;
        }
        else
        {
            /* Insert before the current node and after its previous sibling */
            lpTNPrevSibling->lpNextSibling = (struct MsgSetTreeNode*)lpTNNew;
        }
        lpTNNew->lpNextSibling = (struct MsgSetTreeNode*)lpTNCurr;
        lpTNNew->lpPrevSibling = (struct MsgSetTreeNode*)lpTNPrevSibling;
        lpTNCurr->lpPrevSibling = (struct MsgSetTreeNode*)lpTNNew;
    }

    return lpTNNew;
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     ParseNewLineMsgSet                                                   */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Parse the contents of a newline delimited text message set           */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     N/A                                                                  */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     SUCCESS or FAILURE                                                   */
/*==========================================================================*/
long
ParseNewLineMsgSet()
{
    char* lpszStartOfLine;
    char* lpszCurrChar;
    char* lpszMax;
    long  lCurrLineLen;

    /* Initialise various pointers and counters */
    lpszStartOfLine = kpMCB->lpMSPB->lpszMsgSetBuff;
    lpszCurrChar = kpMCB->lpMSPB->lpszMsgSetBuff;
    lpszMax = kpMCB->lpMSPB->lpszMsgSetBuff + kpMCB->lpMSPB->lMsgSetDataLen;
    lCurrLineLen = 0;

    /* Process the buffer character by character until we run out */
    while(lpszCurrChar < lpszMax)
    {
        /* Got a CR or a LF then we have reached the end of the line */
        if(('\r' == *lpszCurrChar) ||
           ('\n' == *lpszCurrChar))
        {
            *lpszCurrChar = '\0';
            if(0 < lCurrLineLen)
            {
                AddSingleElementMsg(kpszMLINEDATA, kpszLINEDATA, lpszStartOfLine);
            }
            lpszStartOfLine = lpszCurrChar + 1;
            lCurrLineLen = 0;
        }
        else
        {
            lCurrLineLen++;
        }
        lpszCurrChar++;
    }
    if(0 < lCurrLineLen)
    {
        AddSingleElementMsg(kpszMLINEDATA, kpszLINEDATA, lpszStartOfLine);
    }


    return SUCCESS;
}

/*--------------------------------------------------------------------------*/
void
AddSingleElementMsg(const char* const kpszMsgType,
                    const char* const kpszElementType,
                    const char* const kpszElementData)
{
    MsgSetTreeNode* lpMSTN;

    lpMSTN = AddLastChild(kpMCB->lpMSPB->lpCurrMSH->lpMSFRootNode,
                          N_MSG,
                          kpszMSG,
                          kpszMsgType,
                          NULL);
    lpMSTN = AddLastChild(lpMSTN,
                          N_ELEMENT,
                          kpszElementType,
                          kpszElementData,
                          NULL);
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     PrepareMsgSets                                                       */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Having successfully parsed the message set files go through the      */
/*     linked lists and perform further verification and initialisation     */
/*     Also determine the values for LocalCodedCharSetId and LocalEncoding  */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     N/A                                                                  */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     SUCCESS or FAILURE                                                   */
/*==========================================================================*/
long
PrepareMsgSets()
{
    MsgSetLinkedListBlk* lpMSLL;
    SegmentDefBlk*       lpSDB;
    SegmentBlk*          lpSGB;
    MsgDefBlk*           lpMDB;
    MsgSetHdr*           lpMSH;
    Int16                lEncoding;

    /* Determine the LocalCodedCharSetId value */
    if(0 < StrlenNull(kpMCB->lpCIB->lpszLocalCodedCharSetId))
    {
        kpMCB->lpCIB->lLocalCodedCharSetId = 
            ConvNumFailReturnsZero(kpMCB->lpCIB->lpszLocalCodedCharSetId);
        if(0 == kpMCB->lpCIB->lLocalCodedCharSetId)
        {
            LogMsg(221, kpMCB->lpCIB->lpszLocalCodedCharSetId);
            return FAILURE;
        }
        else
        {
            LogMsg(220, kpMCB->lpCIB->lLocalCodedCharSetId);
        }
    }
    else
    {
        kpMCB->lpCIB->lLocalCodedCharSetId = GetLocalCCSID();
        /* If unable to get the LocalCodedCharSetId value stop processing */
        if(0 == kpMCB->lpCIB->lLocalCodedCharSetId)
        {
            LogMsg(219);
            return FAILURE;
        }
        /* Message 218 issued in GetLocalCCSID() if successful */
    }

    /* Determine the LocalEncoding value */
    if(0 < StrlenNull(kpMCB->lpCIB->lpszLocalEncoding))
    {
        if(0 == memcmp(kpMCB->lpCIB->lpszLocalEncoding, "0x", 2))
        {
            HexToLong(&kpMCB->lpCIB->lLocalEncoding,
                      kpMCB->lpCIB->lpszLocalEncoding + 2);
        }
        else
        {
            kpMCB->lpCIB->lLocalEncoding = 
                ConvNumFailReturnsZero(kpMCB->lpCIB->lpszLocalEncoding);
            if(0 == kpMCB->lpCIB->lLocalEncoding)
            {
                LogMsg(224, kpMCB->lpCIB->lpszLocalEncoding);
                return FAILURE;
            }
        }
        lEncoding = (Int16)kpMCB->lpCIB->lLocalEncoding;
        LogMsg(223, lEncoding);
    }
    else
    {
        kpMCB->lpCIB->lLocalEncoding = MQENC_NATIVE;
        lEncoding = (Int16)kpMCB->lpCIB->lLocalEncoding;
        LogMsg(222, lEncoding);
    }

    /* Process each of the message sets */
    lpMSH = kpMCB->lpMSHHead;
    while(NULL != lpMSH)
    {
        /* Check that all the segments defined in a message definition block */
        /* have corresponding segment definition blocks                      */
        lpMDB = lpMSH->lpMDBHead;
        while(NULL != lpMDB)
        {
            lpMSLL = lpMDB->lpMSLLHead;
            while(NULL != lpMSLL)
            {
                if(P_SEGMENT == lpMSLL->eMSLLBlockType)
                {
                    lpSGB = (SegmentBlk*)lpMSLL;
                    /* Find the associated segment definition for this segment */
                    lpSGB->lpSDB = (SegmentDefBlk*)FindMsgSetLinkedListBlk((MsgSetLinkedListBlk*)lpMSH->lpSDBHead,
                                                                           lpSGB->lpszType,
                                                                           lpSGB->lTypeLen);
                    if(! lpSGB->lpSDB)
                    {
                        LogMsg(147, lpSGB->lpszType, lpSGB->lpszType, lpMDB->lpszType);
                        return FAILURE;
                    }
                }
                lpMSLL = (MsgSetLinkedListBlk*)lpMSLL->lpMSLLNext;
            }
            lpMDB = (MsgDefBlk*)lpMDB->lpMDBNext;
        }

        /* Now walk the list of segment definition blocks making sure that any */
        /* segments used therein have their own segment definition block       */
        lpSDB = lpMSH->lpSDBHead;
        while(NULL != lpSDB)
        {
            lpMSLL = lpSDB->lpMSLLHead;
            while(NULL != lpMSLL)
            {
                if(P_SEGMENT == lpMSLL->eMSLLBlockType)
                {
                    lpSGB = (SegmentBlk*)lpMSLL;
                    /* Find the associated segment definition for this segment */
                    lpSGB->lpSDB = (SegmentDefBlk*)FindMsgSetLinkedListBlk((MsgSetLinkedListBlk*)lpMSH->lpSDBHead,
                                                                           lpSGB->lpszType,
                                                                           lpSGB->lTypeLen);
                    if(! lpSGB->lpSDB)
                    {
                        LogMsg(148, lpSGB->lpszType, lpSGB->lpszType);
                        return FAILURE;
                    }
                }
                lpMSLL = (MsgSetLinkedListBlk*)lpMSLL->lpMSLLNext;
            }
            lpSDB = (SegmentDefBlk*)lpSDB->lpSDBNext;
        }

        /* Recursively walk the tree assigning message set linked list block pointers */
        /* for the ElementBlk, SegmentDefBlk, SegmentBlk, MsgDefBlk control blocks    */
        /* used by the <Msg> segments in the <Messages> tree                          */
        SetMsgTreeMSLLs(lpMSH, lpMSH->lpMSFRootNode, NULL, NULL, NULL, 0, 0);
        if(FAILURE == kpMCB->lProcessing)
        {
            return FAILURE;
        }

        /* Recursively walk the tree linking MsgSetTreeNode siblings that have the */
        /* same name or type                                                       */
        SetMsgTreeSameNameLinks(lpMSH->lpMSFRootNode);
        if(FAILURE == kpMCB->lProcessing)
        {
            return FAILURE;
        }

        /* DisplayMsgTree("SMTSNL", lpMSH->lpMSFRootNode, 0, 0); */

        /* Having processed the MsgSetFile tree from its root node MSFRootNode */
        /* to generate the required pointers it is now time to create the      */
        /* actual output generation tree                                       */
        CreateOutMsgTreeNode(lpMSH->lpMSFRootNode, lpMSH->lpOutRootNode, NULL, NULL, NULL, 0, 0);
        if(FAILURE == kpMCB->lProcessing)
        {
            return FAILURE;
        }

        /* DisplayMsgTree("COMTN", lpMSH->lpOutRootNode, 0, 0); */

        /* Next message set */
        lpMSH = (MsgSetHdr*)lpMSH->lpMSHNext;
    }

    return SUCCESS;
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     SetMsgTreeMSLLs                                                      */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Recursive function to find and set the MsgSetLinkedListBlk pointer   */
/*     for this element of the tree before recursing to any children and or */
/*     siblings                                                             */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpMSH               Pointer to message set header                    */
/*     lpTNCall            Pointer to a message set tree node               */
/*     lpRecursionMSSLHead Pointer to head of list containing elements      */
/*                         and segments to be searched for a match          */
/*     lRecursionDepth     Depth the recursive calls have reached           */
/*     lMsgNumber          Number of the <Msg> being processed              */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
SetMsgTreeMSLLs(MsgSetHdr* lpMSH,
                MsgSetTreeNode* lpTNCall,
                MsgSetLinkedListBlk* lpRecursionMSLLHead,
                MsgDefBlk* lpRecursionMDB,
                SegmentBlk* lpRecursionSGB,
                long lRecursionDepth,
                long lMsgNumber)
{
    MsgSetLinkedListBlk* lpMSLLFind;
    MsgDefBlk*           lpMDBFind;
    SegmentBlk*          lpSGBFind;
    MsgSetLinkedListBlk* lpMSLLHeadForChild;

    /* If we've already failed then go no further */
    if(FAILURE == kpMCB->lProcessing)
    {
        return;
    }

    /* Initialise local variables */
    lpMSLLHeadForChild = NULL;

    /* Set the depth value for the current node */
    lpTNCall->lNodeDepth = lRecursionDepth;

    /* Find the appropriate message set linked list block depending */
    /* on the current recursion depth                               */
    switch(lRecursionDepth)
    {
    case 0:
        /* At a recursion depth of zero we are processing */
        /* for the root node so no searching is required  */
        break;
    case 1:
        /* At a recursion depth of one we are dealing with message */
        /* set definition blocks, try to find a matching one       */
        lpMDBFind = (MsgDefBlk*)FindMsgSetLinkedListBlk((MsgSetLinkedListBlk*)lpMSH->lpMDBHead,
                                                        lpTNCall->lpszNodeValue,
                                                        lpTNCall->lNodeValueLen);
        if(! lpMDBFind)
        {
            LogMsg(149, lpTNCall->lpszNodeValue, lMsgNumber);
            kpMCB->lProcessing = FAILURE;
            return;
        }
        lpTNCall->eMsgSetTreeNodeType = N_MSG;
        lpTNCall->lpMDB = lpMDBFind;
        lpTNCall->lpSDB = NULL;
        lpTNCall->lpSGB = NULL;
        lpTNCall->lpMSLL = (MsgSetLinkedListBlk*)lpMDBFind;
        lpTNCall->lpNextRequiredMSLL = NULL; /* Messages are always required */
        lpMSLLHeadForChild = lpMDBFind->lpMSLLHead;
        lpRecursionMDB = lpMDBFind;
        break;
    default:
        /* At a recursion depths greater than one we are looking */
        /* for either an element or segment block that matches   */
        /* this nodes name. Using the current segment block ptr  */
        /* try to locate a match for this nodes name             */
        lpMSLLFind = FindMsgSetLinkedListBlk(lpRecursionMSLLHead,
                                             lpTNCall->lpszNodeName,
                                             lpTNCall->lNodeNameLen);
        if(NULL == lpMSLLFind)
        {
            LogMsg(151, lpTNCall->lpszNodeName, lMsgNumber, lpRecursionMDB->lpszType);
            kpMCB->lProcessing = FAILURE;
            return;
        }
        switch(lpMSLLFind->eMSLLBlockType)
        {
        case P_ELEMENT:
            lpTNCall->eMsgSetTreeNodeType = N_ELEMENT;
            lpTNCall->lpMDB = lpRecursionMDB;
            /* Messages can directly contain elements so no segment will be available */
            if(NULL != lpRecursionSGB)
            {
                lpTNCall->lpSDB = lpRecursionSGB->lpSDB;
                lpTNCall->lpSGB = lpRecursionSGB;
            }
            lpTNCall->lpMSLL = lpMSLLFind;
            lpTNCall->lpNextRequiredMSLL = FindNextRequiredMSLL(lpTNCall->lpMSLL);
            break;
        case P_SEGMENT:
            lpTNCall->eMsgSetTreeNodeType = N_SEGMENT;
            lpTNCall->lpMSLL = lpMSLLFind;
            lpTNCall->lpNextRequiredMSLL = FindNextRequiredMSLL(lpTNCall->lpMSLL);
            lpSGBFind = (SegmentBlk*)lpMSLLFind;
            lpTNCall->lpMDB = lpRecursionMDB;
            lpTNCall->lpSDB = lpSGBFind->lpSDB;
            lpTNCall->lpSGB = lpSGBFind;
            lpMSLLHeadForChild = (MsgSetLinkedListBlk*)lpSGBFind->lpSDB->lpMSLLHead;
            lpRecursionSGB = lpSGBFind;
            break;
        }
        break;
    }

    /* Recursively call this routine for the first child of this element */
    if(NULL != lpTNCall->lpFirstChild)
    {
        /* Calling a child element so the recursion depth increases */
        SetMsgTreeMSLLs(lpMSH,
                        (MsgSetTreeNode*)lpTNCall->lpFirstChild,
                        lpMSLLHeadForChild,
                        lpRecursionMDB,
                        lpRecursionSGB,
                        lRecursionDepth + 1,
                        lMsgNumber);
    }

    /* Recursively call this routine for our next sibling                */
    if(NULL != lpTNCall->lpNextSibling)
    {
        /* If its a <Msg> sibling then increase the message number */
        if((1 == lRecursionDepth) &&
           (0 == memcmp(kpszMSG, lpTNCall->lpszNodeName, strlen(kpszMSG) + 1)))
        {
            lMsgNumber++;
        }
        SetMsgTreeMSLLs(lpMSH,
                        (MsgSetTreeNode*)lpTNCall->lpNextSibling,
                        lpRecursionMSLLHead,
                        lpRecursionMDB,
                        lpRecursionSGB,
                        lRecursionDepth,
                        lMsgNumber);
    }
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     SetMsgTreeSameNameLinks                                              */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Link nodes that have the same name and set their same name index     */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTNCall           Pointer to a message set tree node                */
/*     lRecursionDepth    Depth the recursive calls have reached            */
/*     lMsgNumber         Number of the <Msg> being processed               */
/*     eRecursionCallEnum Type of recursive call for CHILD or SIBLING       */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
SetMsgTreeSameNameLinks(MsgSetTreeNode* lpTNCall)
{
    MsgSetTreeNode* lpTNWork;
    MsgSetTreeNode* lpTNCheck;
    SegmentBlk*     lpCurrSGB;
    ElementBlk*     lpCurrELB;
    long            lNameListLen;

    /* If we've already failed then go no further */
    if(FAILURE == kpMCB->lProcessing)
    {
        return;
    }

    /* Look for nodes with the same name. Only need to process */
    /* nodes that have not been linked yet                     */
    if((NULL == lpTNCall->lpNextSiblingSameName) &&
       (NULL == lpTNCall->lpPrevSiblingSameName))
    {
        lpTNWork = lpTNCall;
        lpTNCheck = (MsgSetTreeNode*)lpTNWork->lpNextSibling;
        while(NULL != lpTNCheck)
        {
            if(lpTNWork->lNodeNameLen == lpTNCheck->lNodeNameLen)
            {
                if(0 == memcmp(lpTNWork->lpszNodeName, lpTNCheck->lpszNodeName, lpTNWork->lNodeNameLen))
                {
                    lpTNWork->lpNextSiblingSameName = (struct MsgSetTreeNode*)lpTNCheck;
                    lpTNCheck->lpPrevSiblingSameName = (struct MsgSetTreeNode*)lpTNWork;
                    lpTNCheck->lIndexSameName = lpTNWork->lIndexSameName + 1;
                    /* Check that the MaxOccurs count has not been exceeded for segments and elements */
                    switch(lpTNCheck->lpMSLL->eMSLLBlockType)
                    {
                    case P_ELEMENT:
                        lpCurrELB = (ElementBlk*)lpTNCheck->lpMSLL;
                        if((klUNBOUNDED != lpCurrELB->lMaxOccurs) &&
                           (lpTNCheck->lIndexSameName >= lpCurrELB->lMaxOccurs))
                        {
                            lNameListLen = 0;
                            memset(kpMCB->lpszErrorWorkBuff, 0x00, ERROR_WORK_BUFF_LEN + 4);
                            NameListInErrorWorkBuff(lpTNCheck, &lNameListLen);
                            LogMsg(229, kpszELEMENT, kpMCB->lpszErrorWorkBuff, lpCurrELB->lMaxOccurs);
                            kpMCB->lProcessing = FAILURE;
                        }
                        break;
                    case P_SEGMENT:
                        lpCurrSGB = (SegmentBlk*)lpTNCheck->lpMSLL;
                        if((klUNBOUNDED != lpCurrSGB->lMaxOccurs) &&
                           (lpTNCheck->lIndexSameName >= lpCurrSGB->lMaxOccurs))
                        {
                            lNameListLen = 0;
                            memset(kpMCB->lpszErrorWorkBuff, 0x00, ERROR_WORK_BUFF_LEN + 4);
                            NameListInErrorWorkBuff(lpTNCheck, &lNameListLen);
                            LogMsg(229, kpszSEGMENT, kpMCB->lpszErrorWorkBuff, lpCurrSGB->lMaxOccurs);
                            kpMCB->lProcessing = FAILURE;
                        }
                        break;
                    }
                    lpTNWork = lpTNCheck;
                }
            }
            lpTNCheck = (MsgSetTreeNode*)lpTNCheck->lpNextSibling;
        }
    }

    /* Recursively call this routine for the first child of this element */
    if(NULL != lpTNCall->lpFirstChild)
    {
        SetMsgTreeSameNameLinks((MsgSetTreeNode*)lpTNCall->lpFirstChild);
    }

    /* Recursively call this routine for our next sibling                */
    if(NULL != lpTNCall->lpNextSibling)
    {
        SetMsgTreeSameNameLinks((MsgSetTreeNode*)lpTNCall->lpNextSibling);
    }
}

/*==========================================================================*/
/* FUNCTION:                                                                */
/*     CreateOutMsgTreeNode                                                 */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Create the output message tree from the message set file tree        */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTNMsgSetFile     Pointer to a message set file tree node           */
/*     lpTNOutParent      Parent tree node of the current output tree node  */
/*     lRecursionDepth    Depth the recursive calls have reached            */
/*     lMsgNumber         Number of the <Msg> being processed               */
/*     eRecursionCallEnum Type of recursive call for CHILD or SIBLING       */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
CreateOutMsgTreeNode(MsgSetTreeNode* lpTNMsgSetFile,
                     MsgSetTreeNode* lpTNOutParent, 
                     MsgSetTreeNode* lpTNOut,
                     MsgDefBlk* lpRecursionMDB,
                     SegmentBlk* lpRecursionSGB,
                     long lRecursionDepth, long lMsgNumber)
{
    MsgSetTreeNode* lpTNWork;
    MsgSetTreeNode* lpTNCurrOutParent;
    MsgSetTreeNode* lpTNCurrOut;
    MsgSetTreeNode* lpTNInsert;
    MsgSetTreeNode* lpTNFind;
    MsgSetTreeNode* lpTNOutLastWithSameName;
    ElementBlk*     lpCurrELB;
    long            lElementNodeValueCheckLen;
    long            lNameListLen;
    
    /* If we've already failed then go no further */
    if(FAILURE == kpMCB->lProcessing)
    {
        return;
    }

    /* Initialise local variables */
    lpTNWork = NULL;
    lpTNCurrOut = lpTNOut;
    lpTNCurrOutParent = lpTNOutParent;

    /* Find the appropriate message set linked list block depending */
    /* on the current recursion depth                               */
    switch(lRecursionDepth)
    {
    case 0:
        /* At a recursion depth of zero we are processing */
        /* for the root node so no searching is required  */
        break;
    default:
        /* Try to find a child with a matching name */
        lpTNOutLastWithSameName = NULL;
        lpTNFind = FindMatchingChild(lpTNCurrOutParent,
                                     lpTNMsgSetFile->lpszNodeName,
                                     lpTNMsgSetFile->lNodeNameLen,
                                     lpTNMsgSetFile->lIndexSameName,
                                     &lpTNOutLastWithSameName);
        
        /* If not found then we need to create one */
        /* otherwise just set its value            */
        if(NULL == lpTNFind)
        {
            lpTNInsert = NULL;
            if(NULL != lpTNCurrOutParent->lpFirstChild)
            {
                if(NULL != lpTNMsgSetFile->lpNextRequiredMSLL)
                {
                    /* ???TA set NULL below for better scanning performance */
                    lpTNInsert = FindMsgTreeNode(lpTNCurrOutParent,
                                                 NULL,
                                                 lpTNMsgSetFile->lpNextRequiredMSLL->lpszNameOrType,
                                                 lpTNMsgSetFile->lpNextRequiredMSLL->lNameOrTypeLen);
                }
            }
            switch(lpTNMsgSetFile->eMsgSetTreeNodeType)
            {
            case N_MSG:
                lpTNCurrOut = CreateTreeNodeFromMsgDefBlk(lpTNCurrOutParent,
                                                          lpTNOutLastWithSameName,
                                                          (MsgDefBlk*)lpTNMsgSetFile->lpMSLL,
                                                          lRecursionDepth);
                lpRecursionMDB = (MsgDefBlk*)lpTNMsgSetFile->lpMSLL;
                break;
            case N_SEGMENT:
                lpTNCurrOut = CreateTreeNodeFromSegmentBlk(lpTNCurrOutParent,
                                                           lpTNInsert,
                                                           lpTNOutLastWithSameName,
                                                           lpRecursionMDB,
                                                           (SegmentBlk*)lpTNMsgSetFile->lpMSLL,
                                                           lRecursionDepth);
                lpRecursionSGB = (SegmentBlk*)lpTNMsgSetFile->lpMSLL;
                break;
            case N_ELEMENT:
                lpTNCurrOut = CreateTreeNodeFromElementBlk(lpTNCurrOutParent,
                                                           lpTNInsert,
                                                           lpTNOutLastWithSameName,
                                                           lpRecursionMDB,
                                                           lpRecursionSGB,
                                                           (ElementBlk*)lpTNMsgSetFile->lpMSLL,
                                                           lRecursionDepth);
                break;
            }
            lpTNCurrOut->lIndexSameName = lpTNMsgSetFile->lIndexSameName;
            lpTNCurrOut->lpszNodeValue = CloneStringNull(lpTNMsgSetFile->lpszNodeValue);
            lpTNCurrOut->lNodeValueLen = StrlenNull(lpTNCurrOut->lpszNodeValue);
            lpTNCurrOut->lpMDB = lpRecursionMDB;
            if(NULL != lpRecursionSGB)
            {
                lpTNCurrOut->lpSDB = lpRecursionSGB->lpSDB;
                lpTNCurrOut->lpSGB = lpRecursionSGB;
            }
            lpTNCurrOutParent = lpTNCurrOut;
        }
        else
        {
            lpTNCurrOut = lpTNFind;
            lpRecursionMDB = lpTNMsgSetFile->lpMDB;
            lpRecursionSGB = lpTNMsgSetFile->lpSGB;
            lpTNCurrOut->lIndexSameName = lpTNMsgSetFile->lIndexSameName;
            lpTNCurrOut->lpszNodeValue = CloneStringNull(lpTNMsgSetFile->lpszNodeValue);
            lpTNCurrOut->lNodeValueLen = StrlenNull(lpTNCurrOut->lpszNodeValue);
            lpTNCurrOut->lpMDB = lpRecursionMDB;
            if(NULL != lpRecursionSGB)
            {
                lpTNCurrOut->lpSDB = lpRecursionSGB->lpSDB;
                lpTNCurrOut->lpSGB = lpRecursionSGB;
            }
            lpTNCurrOutParent = lpTNCurrOut;
        }

        /* Validate Element output nodes */
        if(N_ELEMENT == lpTNCurrOut->eMsgSetTreeNodeType)
        {
            lpCurrELB = (ElementBlk*)lpTNCurrOut->lpMSLL;
            lElementNodeValueCheckLen = lpCurrELB->lLen;
            if(klUNDEFINED == lElementNodeValueCheckLen)
            {
                lElementNodeValueCheckLen = lpTNCurrOut->lNodeValueLen;
            }
            if(0 >= lElementNodeValueCheckLen)
            {
                lNameListLen = 0;
                memset(kpMCB->lpszErrorWorkBuff, 0x00, ERROR_WORK_BUFF_LEN + 4);
                NameListInErrorWorkBuff(lpTNCurrOut, &lNameListLen);
                LogMsg(232, kpMCB->lpszErrorWorkBuff, lpCurrELB->lpszName, lpCurrELB->lpszElementType);
                kpMCB->lProcessing = FAILURE;
                return;
            }
        }
    }
    
    /* Recursively call this routine for the first child of this element */
    if(NULL != lpTNMsgSetFile->lpFirstChild)
    {
        /* If called for a child element then the recursion depth has increased */
        CreateOutMsgTreeNode((MsgSetTreeNode*)lpTNMsgSetFile->lpFirstChild,
                             (MsgSetTreeNode*)lpTNCurrOutParent,
                             lpTNCurrOut,
                             lpRecursionMDB,
                             lpRecursionSGB,
                             lRecursionDepth + 1,
                             lMsgNumber);
    }

    /* Recursively call this routine for our next sibling                */
    if(NULL != lpTNMsgSetFile->lpNextSibling)
    {
        /* If its a <Msg> sibling then increase the message number */
        if((1 == lRecursionDepth) &&
           (0 == memcmp(kpszMSG, lpTNMsgSetFile->lpszNodeName, strlen(kpszMSG) + 1)))
        {
            lMsgNumber++;
        }

        CreateOutMsgTreeNode((MsgSetTreeNode*)lpTNMsgSetFile->lpNextSibling,
                             (MsgSetTreeNode*)lpTNCurrOutParent->lpParent,
                             lpTNCurrOut,
                             lpRecursionMDB,
                             lpRecursionSGB,
                             lRecursionDepth,
                             lMsgNumber);
    }
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     CreateTreeNodeFromMsgDefBlk                                          */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Create the output message tree elements from a message definition    */
/*     block                                                                */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTNMsgSetFile     Pointer to a message set file tree node from      */
/*                        which the new node is to be created               */
/*     lpMDB              Pointer to the message definition block           */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     MsgSetTreeNode*    Pointer to the node created by this routine       */
/*==========================================================================*/
MsgSetTreeNode*
CreateTreeNodeFromMsgDefBlk(MsgSetTreeNode* lpTNCurrOutParent,
                            MsgSetTreeNode* lpTNLastWithSameName,
                            MsgDefBlk* lpMDB,
                            long lRecursionDepth)
{
    MsgSetTreeNode*      lpTNNewMsg;
    MsgSetLinkedListBlk* lpMSLL;
    SegmentBlk*          lpCurrSGB;
    ElementBlk*          lpCurrELB;

    lpTNNewMsg = AddLastChild(lpTNCurrOutParent,
                              N_MSG,
                              kpszMSG,
                              lpMDB->lpszType,
                              (MsgSetLinkedListBlk*)lpMDB);
    lpTNNewMsg->lNodeDepth = lRecursionDepth;
    lpTNNewMsg->lpMDB = lpMDB;
    lpTNNewMsg->lpSDB = NULL;
    lpTNNewMsg->lpSGB = NULL;

    if(NULL != lpTNLastWithSameName)
    {
        lpTNNewMsg->lpPrevSiblingSameName = (struct MsgSetTreeNode*)lpTNLastWithSameName;
        lpTNLastWithSameName->lpNextSiblingSameName = (struct MsgSetTreeNode*)lpTNNewMsg;
    }

    /* Generate element and segment nodes for the current message definition */
    lpMSLL = lpMDB->lpMSLLHead;
    while(NULL != lpMSLL)
    {
        switch(lpMSLL->eMSLLBlockType)
        {
        case P_SEGMENT:
            lpCurrSGB = (SegmentBlk*)lpMSLL;
            /* Segments with a MinOccurs value of zero are optional and will */
            /* only be created when they are actually declared in a message  */
            if(0 < lpCurrSGB->lMinOccurs)
            {
                CreateTreeNodeFromSegmentBlk(lpTNNewMsg, NULL, NULL, lpMDB, lpCurrSGB, lRecursionDepth + 1);
            }
            break;
        case P_ELEMENT:
            lpCurrELB = (ElementBlk*)lpMSLL;
            /* Elements with a MinOccurs value of zero are optional and will */
            /* only be created when they are actually declared in a message  */
            if(0 < lpCurrELB->lMinOccurs)
            {
                CreateTreeNodeFromElementBlk(lpTNNewMsg, NULL, NULL, lpMDB, NULL, lpCurrELB, lRecursionDepth + 1);
            }
            break;
        }
        lpMSLL = (MsgSetLinkedListBlk*)lpMSLL->lpMSLLNext;
    }

    return lpTNNewMsg;
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     CreateTreeNodeFromSegmentBlk                                         */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Create the output message tree elements from a segment               */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTNCurrOutParent  Pointer to the output tree node that will become  */
/*                        the parent of this message                        */
/*     lpSGB              Pointer to the segment block                      */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     MsgSetTreeNode*    Pointer to the node created by this routine       */
/*==========================================================================*/
MsgSetTreeNode*
CreateTreeNodeFromSegmentBlk(MsgSetTreeNode* lpTNCurrOutParent,
                             MsgSetTreeNode* lpTNCurrOut,
                             MsgSetTreeNode* lpTNLastWithSameName,
                             MsgDefBlk* lpMDB,
                             SegmentBlk* lpSGB,
                             long lRecursionDepth)
{
    MsgSetTreeNode*      lpTNNewSegment;
    SegmentDefBlk*       lpSDB;
    MsgSetLinkedListBlk* lpMSLL;
    SegmentBlk*          lpCurrSGB;
    ElementBlk*          lpCurrELB;

    if(NULL == lpTNCurrOut)
    {
        lpTNNewSegment = AddLastChild(lpTNCurrOutParent,
                                      N_SEGMENT,
                                      lpSGB->lpszType,
                                      NULL,
                                      (MsgSetLinkedListBlk*)lpSGB);
    }
    else
    {
        lpTNNewSegment = AddPrevSibling(lpTNCurrOut,
                                        N_SEGMENT,
                                        lpSGB->lpszType,
                                        NULL,
                                        (MsgSetLinkedListBlk*)lpSGB);
    }
    lpTNNewSegment->lNodeDepth = lRecursionDepth;
    lpTNNewSegment->lpMDB = lpMDB;
    lpTNNewSegment->lpSDB = lpSGB->lpSDB;
    lpTNNewSegment->lpSGB = lpSGB;

    if(NULL != lpTNLastWithSameName)
    {
        lpTNNewSegment->lpPrevSiblingSameName = (struct MsgSetTreeNode*)lpTNLastWithSameName;
        lpTNLastWithSameName->lpNextSiblingSameName = (struct MsgSetTreeNode*)lpTNNewSegment;
    }

    /* Generate element and segment nodes for the current segment definition */
    lpSDB = (SegmentDefBlk*)lpSGB->lpSDB;
    lpMSLL = lpSDB->lpMSLLHead;
    while(NULL != lpMSLL)
    {
        switch(lpMSLL->eMSLLBlockType)
        {
        case P_SEGMENT:
            lpCurrSGB = (SegmentBlk*)lpMSLL;
            /* Segments with a MinOccurs value of zero are optional and will */
            /* only be created when they are actually declared in a message  */
            if(0 < lpCurrSGB->lMinOccurs)
            {
                CreateTreeNodeFromSegmentBlk(lpTNNewSegment, NULL, NULL, lpMDB, lpCurrSGB, lRecursionDepth + 1);
            }
            break;
        case P_ELEMENT:
            lpCurrELB = (ElementBlk*)lpMSLL;
            /* Elements with a MinOccurs value of zero are optional and will */
            /* only be created when they are actually declared in a message  */
            if(0 < lpCurrELB->lMinOccurs)
            {
                CreateTreeNodeFromElementBlk(lpTNNewSegment, NULL, NULL, lpMDB, lpSGB, lpCurrELB, lRecursionDepth + 1);
            }
            break;
        }
        lpMSLL = (MsgSetLinkedListBlk*)lpMSLL->lpMSLLNext;
    }

    return lpTNNewSegment;
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     CreateTreeNodeFromElementBlk                                         */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Create the output message tree elements from a Element               */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTNCurrOutParent  Pointer to the output tree node that will become  */
/*                        the parent of this element                        */
/*     lpELB              Pointer to the Element block                      */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     MsgSetTreeNode*    Pointer to the node created by this routine       */
/*==========================================================================*/
MsgSetTreeNode*
CreateTreeNodeFromElementBlk(MsgSetTreeNode* lpTNCurrOutParent,
                             MsgSetTreeNode* lpTNCurrOut,
                             MsgSetTreeNode* lpTNLastWithSameName,
                             MsgDefBlk* lpMDB,
                             SegmentBlk* lpSGB,
                             ElementBlk* lpELB,
                             long lRecursionDepth)
{
    MsgSetTreeNode* lpTNNewElement;

    if(NULL == lpTNCurrOut)
    {
        lpTNNewElement = AddLastChild(lpTNCurrOutParent,
                                      N_ELEMENT,
                                      lpELB->lpszName,
                                      lpELB->lpszDefaultValue,
                                      (MsgSetLinkedListBlk*)lpELB);
    }
    else
    {
        lpTNNewElement = AddPrevSibling(lpTNCurrOut,
                                        N_ELEMENT,
                                        lpELB->lpszName,
                                        lpELB->lpszDefaultValue,
                                        (MsgSetLinkedListBlk*)lpELB);
    }
    lpTNNewElement->lNodeDepth = lRecursionDepth;
    lpTNNewElement->lpMDB = lpMDB;
    if(NULL != lpSGB)
    {
        lpTNNewElement->lpSDB = lpSGB->lpSDB;
        lpTNNewElement->lpSGB = lpSGB;
    }

    if(NULL != lpTNLastWithSameName)
    {
        lpTNNewElement->lpPrevSiblingSameName = (struct MsgSetTreeNode*)lpTNLastWithSameName;
        lpTNLastWithSameName->lpNextSiblingSameName = (struct MsgSetTreeNode*)lpTNNewElement;
    }


    return lpTNNewElement;
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     FindMatchingChild                                                    */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Find a matching child                                                */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTNParent          Pointer to the parent node                       */
/*     lpszChildName       Name of the child we are searching for           */
/*     lChildNameLen       Length of the child name                         */
/*     lChildIndexSameName Index value of the child we are searching for    */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     MsgSetTreeNode*    Pointer to a matching node or NULL                */
/*==========================================================================*/
MsgSetTreeNode*
FindMatchingChild(MsgSetTreeNode* lpTNParent, char* lpszChildName,
                  long lChildNameLen, long lChildIndexSameName,
                  MsgSetTreeNode** lplpTNLastWithSameName)
{
    /* If we have no children we can't find a match */
    if(NULL == lpTNParent->lpFirstChild)
    {
        return NULL;
    }

    return FindMatchingSibling((MsgSetTreeNode*)lpTNParent->lpFirstChild,
                               lpszChildName, lChildNameLen,
                               lChildIndexSameName,
                               lplpTNLastWithSameName);
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     FindMatchingSibling                                                  */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Find a matching sibling                                              */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTNSibling           Pointer to sibling node to start searching at  */
/*     lpszSiblingName       Name of the sibling we are searching for       */
/*     lSiblingNameLen       Length of the sibling name                     */
/*     lSiblingIndexSameName Index value of sibling we are searching for    */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     MsgSetTreeNode*    Pointer to a matching node or NULL                */
/*==========================================================================*/
MsgSetTreeNode*
FindMatchingSibling(MsgSetTreeNode* lpTNSibling, char* lpszSiblingName,
                    long lSiblingNameLen, long lSiblingIndexSameName,
                    MsgSetTreeNode** lplpTNLastWithSameName)
{
    MsgSetTreeNode* lpTNWorkSibling;

    lpTNWorkSibling = lpTNSibling;
    while(NULL != lpTNWorkSibling)
    {
        if(lSiblingNameLen == lpTNWorkSibling->lNodeNameLen)
        {
            if(0 == memcmp(lpszSiblingName, lpTNWorkSibling->lpszNodeName, lSiblingNameLen))
            {
                /* Now transition to using the same name pointers */
                while(NULL != lpTNWorkSibling)
                {
                    *lplpTNLastWithSameName = lpTNWorkSibling;
                    if(lSiblingIndexSameName == lpTNWorkSibling->lIndexSameName)
                    {
                        return lpTNWorkSibling;
                    }
                    lpTNWorkSibling = (MsgSetTreeNode*)lpTNWorkSibling->lpNextSiblingSameName;
                }
                return NULL;
            }
        }
        lpTNWorkSibling = (MsgSetTreeNode*)lpTNWorkSibling->lpNextSibling;
    }
    
    return NULL;
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     CheckAndAddMSH                                                       */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Check to see if a MsgSetHdr exists if so just return it. If not      */
/*     create one, initialise it and add it to the linked list.             */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     kpszMsgSetName   Name of the message set header to find              */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     Pointer to a MsgSetHdr                                               */
/*==========================================================================*/
MsgSetHdr*
CheckAndAddMSH(const char* const kpszMsgSetName)
{
    MsgSetHdr* lpMSH;

    /* If no name is supplied then fail and return NULL */
    if(NULL == kpszMsgSetName)
    {
        kpMCB->lProcessing = FAILURE;
        return NULL;
    }

    /* If its an empty string then fail and return NULL */
    if(0 == StrlenNull(kpszMsgSetName))
    {
        kpMCB->lProcessing = FAILURE;
        return NULL;
    }

    /* If we can find it just return the pointer */
    lpMSH = FindMsgSetHdr(kpszMsgSetName);
    if(lpMSH)
    {
        return lpMSH;
    }

    /* Ok can't find a message set header with this name lets make one */
    lpMSH = malloc(sizeof(MsgSetHdr));
    if(lpMSH)
    {
        memset(lpMSH, 0x00, sizeof(MsgSetHdr));
        lpMSH->lpMSHNext = NULL;
        lpMSH->lpMSHPrev = NULL;
        lpMSH->lpMDBHead = NULL;
        lpMSH->lpMDBTail = NULL;
        lpMSH->lpSDBHead = NULL;
        lpMSH->lpSDBTail = NULL;
        lpMSH->lpMSFRootNode = CreateMsgSetTreeNode(N_ROOT, kpszROOT, kpszROOT, &kpMCB->oDummyRootMSLL);
        lpMSH->lpMSFRootNode->lNodeDepth = 0;
        lpMSH->lpOutRootNode = CreateMsgSetTreeNode(N_ROOT, kpszROOT, kpszROOT, &kpMCB->oDummyRootMSLL);
        lpMSH->lpOutRootNode->lNodeDepth = 0;
        lpMSH->lpszMsgSetName = CloneString(kpszMsgSetName);
        /* Add it to the MsgSetHdr linked list */
        AddToTailOfList((LinkedListBlk**)&kpMCB->lpMSHHead,
                        (LinkedListBlk**)&kpMCB->lpMSHTail,
                        (LinkedListBlk*)lpMSH);
    }
    else
    {
        LogMsg(3, sizeof(MsgSetHdr), "MsgSetHdr", __LINE__, __FILE__);
        kpMCB->lProcessing = FAILURE;
        return NULL;
    }

    return lpMSH;
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     FindMsgSetHdr                                                        */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Find a MsgSetHeader in the linked list                               */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpMSH            Pointer to first MsgSetHdr in the chain             */
/*     kpszMsgSetName   Name of the MsgSetHdr to find                       */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     Pointer to the MsgSetHdr, NULL if not found                          */
/*==========================================================================*/
MsgSetHdr*
FindMsgSetHdr(const char* const kpszMsgSetName)
{
    MsgSetHdr* lpMSH;
    long       lMsgSetNameLen;
    long       lMSHNameLen;

    /* If no name is supplied then we can't find it  */
    if(NULL == kpszMsgSetName)
    {
        return NULL;
    }

    /* If its an empty string then we don't allow it */
    lMsgSetNameLen = StrlenNull(kpszMsgSetName);
    if(0 == lMsgSetNameLen)
    {
        return NULL;
    }

    lpMSH = kpMCB->lpMSHHead;
    while(NULL != lpMSH)
    {
        lMSHNameLen = StrlenNull(lpMSH->lpszMsgSetName);
        if(lMsgSetNameLen == lMSHNameLen)
        {
            if(0 == memcmp(kpszMsgSetName, lpMSH->lpszMsgSetName,
                           lMsgSetNameLen))
            {
                return lpMSH;
            }
        }
        lpMSH = (MsgSetHdr*)lpMSH->lpMSHNext;
    }

    return NULL;
}

/*==========================================================================*/
/* FUNCTION:                                                                */
/*     GenerateMsgInBuff                                                    */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Generate the requested message in the designated buffer              */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTIB     Pointer to a test info block                               */
/*     lpBIB     Pointer to the buffer in which to generate the message     */
/*     lpMsgMSTN Pointer to the message set tree node for this message      */
/*     lpVIBMQMD Pointer to var info block for MQMD                         */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
GenerateMsgInBuff(TestInfoBlk* lpTIB,
                  BuffInfoBlk* lpBIB,
                  MsgSetTreeNode* lpMsgMSTN,
                  VarInfoBlk* lpVIBMQMD)
{
    long lIdx;

    /* Clear any previous message data in the buffer */
    if(0 < lpBIB->lDataLen)
    {
        memset(lpBIB->lpszBuff, 0x00, lpBIB->lDataLen);
        lpBIB->lDataLen = 0;
    }

    /* Initialise message set segment lengths array */
    for(lIdx = 0; lIdx < MAX_MSGSET_PARSE_DEPTH +1; lIdx++)
    {
        kpMCB->lMsgSetSegmentLen[lIdx] = 0;
    }

    /* Generate the data from the message tree nodes       */
    /* NOTE. Start segment indexes at 1 as minus 1 is used */
    /* when getting the segment length                     */
    GenerateNodeData(lpTIB,  lpBIB, (MsgSetTreeNode*)lpMsgMSTN->lpFirstChild, lpVIBMQMD, 1);
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     GenerateNodeData                                                     */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Put the data for the current tree node into the buffer.              */
/*     Process all the nodes in the message tree recursively.               */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTIB           Pointer to a test info block                         */
/*     lpBIB           Pointer to the buffer info block                     */
/*     lpTNData        Pointer to tree node to be processed                 */
/*     lpVIBMQMD       Pointer to var info block for MQMD                   */
/*     lSegmentIndex   Index value for segment length array                 */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
GenerateNodeData(TestInfoBlk* lpTIB,
                 BuffInfoBlk* lpBIB,
                 MsgSetTreeNode* lpTNData,
                 VarInfoBlk* lpVIBMQMD,
                 long lSegmentIndex)
{
    MsgSetTreeNode* lpTNParent;
    MsgSetTreeNode* lpTNWork;
    ElementBlk*     lpCurrSGB;
    ElementBlk*     lpCurrELB;
    MQMDInfoBlk*    lpMIB;
    char*           lpszResValue;
    char*           lpszHexValue;
    long            lResValueLen;
    long            lHexValueLen;
    long            lPackedValueLen;
    long            lOutputDataLen;
    long            lHexOutputLen;
    long            lPackedOutputLen;
    long            lNameListLen;
    long            lIdx;
    long            lSetSeglenFlag;
    long            lSeglen;
    char*           lpszNum;

    /* If we've already failed then go no further */
    if((FAILURE == lpTIB->bCmdResult) ||
       (FAILURE == kpMCB->lProcessing))
    {
        return;
    }

    /* Initialise local variables */
    lpszResValue = NULL;
    lResValueLen = 0;
    lOutputDataLen = 0;

    /* Initialise tree node output related variables */
    lpTNData->lFlags = MSTN_NONE;
    lpTNData->lOutputDataLen = 0;
    lpTNData->lpszBuffLoc = lpBIB->lpszBuff + lpBIB->lDataLen;

    /* Process the node according to type */
    switch(lpTNData->eMsgSetTreeNodeType)
    {
    case N_SEGMENT:
        lpCurrSGB = (ElementBlk*)lpTNData->lpMSLL;
        kpMCB->lMsgSetSegmentLen[lSegmentIndex] = 0;
        break;
    case N_ELEMENT:
        lpCurrELB = (ElementBlk*)lpTNData->lpMSLL;
        /* If our parent is an MQMD then populate the fields in the MQMDInfoBlk */
        /* otherwise process the tree nodes data according to element type      */
        lpTNParent = (MsgSetTreeNode*)lpTNData->lpParent;
        if(0 == memcmp(kpszMQMD, lpTNParent->lpszNodeName, StrlenNull(kpszMQMD) + 1))
        {
            lpMIB = (MQMDInfoBlk*)lpVIBMQMD->lpCtrlBlk;
            if(0 == memcmp(kpszSTRUCID, lpTNData->lpszNodeName, strlen(kpszSTRUCID) + 1))
            {
                /* Do nothing */
            }
            else if(0 == memcmp(kpszVERSION, lpTNData->lpszNodeName, strlen(kpszVERSION) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszVersion, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszREPORT, lpTNData->lpszNodeName, strlen(kpszREPORT) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszReport, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszMSGTYPE, lpTNData->lpszNodeName, strlen(kpszMSGTYPE) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszMsgType, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszEXPIRY, lpTNData->lpszNodeName, strlen(kpszEXPIRY) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszExpiry, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszFEEDBACK, lpTNData->lpszNodeName, strlen(kpszFEEDBACK) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszFeedback, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszENCODING, lpTNData->lpszNodeName, strlen(kpszENCODING) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszEncoding, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszCODEDCHARSETID, lpTNData->lpszNodeName, strlen(kpszCODEDCHARSETID) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszCodedCharSetId, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszFORMAT, lpTNData->lpszNodeName, strlen(kpszFORMAT) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszFormat, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszPRIORITY, lpTNData->lpszNodeName, strlen(kpszPRIORITY) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszPriority, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszPERSISTENCE, lpTNData->lpszNodeName, strlen(kpszPERSISTENCE) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszPersistence, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszMSGID, lpTNData->lpszNodeName, strlen(kpszMSGID) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszMsgId, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszCORRELID, lpTNData->lpszNodeName, strlen(kpszCORRELID) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszCorrelId, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszBACKOUTCOUNT, lpTNData->lpszNodeName, strlen(kpszBACKOUTCOUNT) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszBackoutCount, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszREPLYTOQ, lpTNData->lpszNodeName, strlen(kpszREPLYTOQ) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszReplyToQ, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszREPLYTOQMGR, lpTNData->lpszNodeName, strlen(kpszREPLYTOQMGR) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszReplyToQMgr, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszUSERIDENTIFIER, lpTNData->lpszNodeName, strlen(kpszUSERIDENTIFIER) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszUserIdentifier, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszACCOUNTINGTOKEN, lpTNData->lpszNodeName, strlen(kpszACCOUNTINGTOKEN) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszAccountingToken, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszAPPLIDENTITYDATA, lpTNData->lpszNodeName, strlen(kpszAPPLIDENTITYDATA) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszApplIdentityData, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszPUTAPPLTYPE, lpTNData->lpszNodeName, strlen(kpszPUTAPPLTYPE) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszPutApplType, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszPUTAPPLNAME, lpTNData->lpszNodeName, strlen(kpszPUTAPPLNAME) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszPutApplName, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszPUTDATE, lpTNData->lpszNodeName, strlen(kpszPUTDATE) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszPutDate, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszPUTTIME, lpTNData->lpszNodeName, strlen(kpszPUTTIME) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszPutTime, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszAPPLORIGINDATA, lpTNData->lpszNodeName, strlen(kpszAPPLORIGINDATA) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszApplOriginData, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszGROUPID, lpTNData->lpszNodeName, strlen(kpszGROUPID) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszGroupId, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszMSGSEQNUMBER, lpTNData->lpszNodeName, strlen(kpszMSGSEQNUMBER) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszMsgSeqNumber, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszOFFSET, lpTNData->lpszNodeName, strlen(kpszOFFSET) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszOffset, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszMSGFLAGS, lpTNData->lpszNodeName, strlen(kpszMSGFLAGS) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszMsgFlags, &lpTNData->lpszNodeValue);
            }
            else if(0 == memcmp(kpszORIGINALLENGTH, lpTNData->lpszNodeName, strlen(kpszORIGINALLENGTH) + 1))
            {
                ChangeIfStr2Exists(&lpMIB->lpszOriginalLength, &lpTNData->lpszNodeValue);
            }
            else
            {
                /* ??? TA Throw an error ? */
            }
        }
        else
        {
            /* Resolve any variables in the data. Use the tree node value */
            /* if available otherwise use the nodes default value         */
            if(0 == lpTNData->lNodeValueLen)
            {
                ResolveNodeValue(lpTNData, lpTIB, &lpszResValue, &lpCurrELB->lpszDefaultValue, TRUE);
            }
            else
            {
                ResolveNodeValue(lpTNData, lpTIB, &lpszResValue, &lpTNData->lpszNodeValue, TRUE);
            }

            /* Process the data according to the elements type */
            switch(lpCurrELB->eElementType)
            {
            case T_CHAR:
                lResValueLen = StrlenNull(lpszResValue);
                if(TRUE == ResolveCharCmqValue(lpTIB, lpBIB, lpszResValue, lpCurrELB->lLen, lpCurrELB->bCmqValues))
                {
                    lOutputDataLen = lpCurrELB->lLen;
                    lpBIB->lDataLen += lOutputDataLen;
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                else
                {
                    /* Determine the length of the resolved data and the output field */
                    if(klUNDEFINED == lpCurrELB->lLen)
                    {
                        lOutputDataLen = lResValueLen;
                    }
                    else
                    {
                        /* Check for data truncation and only process if its allowed for the message */
                        if(lResValueLen > lpCurrELB->lLen)
                        {
                            if(FALSE == lpTNData->lpMDB->bTruncate)
                            {
                                lNameListLen = 0;
                                memset(kpMCB->lpszErrorWorkBuff, 0x00, ERROR_WORK_BUFF_LEN + 4);
                                NameListInErrorWorkBuff(lpTNData, &lNameListLen);
                                LogMsg(233, lpTIB->lTestNum, lResValueLen, kpMCB->lpszErrorWorkBuff, lpCurrELB->lLen);
                                FreeIfNotNull((void**)&lpszResValue);
                                lpTIB->bCmdResult = FAILURE;
                                return;
                            }
                            /* Set lResValueLen so it truncates */
                            lResValueLen = lpCurrELB->lLen;
                        }
                        lOutputDataLen = lpCurrELB->lLen;
                    }
                    /* If requested take care of end of field alignment */
                    if(1 < lpCurrELB->lEndAlign)
                    {
                        if(0 != (lOutputDataLen % lpCurrELB->lEndAlign))
                        {
                            lOutputDataLen = ((lOutputDataLen / lpCurrELB->lEndAlign) + 1) * lpCurrELB->lEndAlign;
                        }
                    }
                    /* Check the buffer can fit the data if not then extend it */
                    if(FAILURE == CheckAndExtendBuff(lpBIB, lOutputDataLen))
                    {
                        return;
                    }
                    /* Handle any padding requirements */
                    if(lResValueLen == lOutputDataLen)
                    {
                        memcpy(lpBIB->lpszBuff + lpBIB->lDataLen, lpszResValue, lResValueLen);
                    }
                    else
                    {
                        if(0 == memcmp(kpszLEFT, lpCurrELB->lpszJustify, strlen(kpszLEFT) + 1))
                        {
                            memcpy(lpBIB->lpszBuff + lpBIB->lDataLen, lpszResValue, lResValueLen);
                            if(0 < StrlenNull(lpCurrELB->lpszPadChar))
                            {
                                memset(lpBIB->lpszBuff + lpBIB->lDataLen + lResValueLen, *lpCurrELB->lpszPadChar, lOutputDataLen - lResValueLen);
                            }
                        }
                        else
                        {
                            memcpy(lpBIB->lpszBuff + lpBIB->lDataLen + (lOutputDataLen - lResValueLen), lpszResValue, lResValueLen);
                            if(0 < StrlenNull(lpCurrELB->lpszPadChar))
                            {
                                memset(lpBIB->lpszBuff + lpBIB->lDataLen, *lpCurrELB->lpszPadChar, lOutputDataLen - lResValueLen);
                            }
                        }
                    }
                    lpBIB->lDataLen += lOutputDataLen;
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                break;
            case T_HEX:
                /* Get a pointer to the hex value ignoring the leading '0x' bytes if present */
                lpszHexValue = lpszResValue;
                if(0 == memcmp(lpszResValue, "0x", 2))
                {
                    lpszHexValue += 2;
                }

                if(TRUE == ResolveHexCmqValue(lpTIB, lpBIB, lpszHexValue, lpCurrELB->lLen, lpCurrELB->bCmqValues))
                {
                    lOutputDataLen = lpCurrELB->lLen;
                    lpBIB->lDataLen += lOutputDataLen;
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                else
                {
                    if(FAILURE == CheckValueIsHex(lpTNData, lpTIB, lpszHexValue))
                    {
                        return;
                    }

                    /* Determine the length of the resolved data and the output field */
                    lHexValueLen = StrlenNull(lpszHexValue);
                    if(0 == (lHexValueLen % 2))
                    {
                        lHexOutputLen = lHexValueLen / 2;
                    }
                    else
                    {
                        lHexOutputLen = (lHexValueLen / 2) + 1;
                    }
                    if(klUNDEFINED == lpCurrELB->lLen)
                    {
                        lOutputDataLen = lHexOutputLen;
                    }
                    else
                    {
                        /* Check for data truncation and only process if its allowed for the message */
                        if(lHexOutputLen > lpCurrELB->lLen)
                        {
                            if(FALSE == lpTNData->lpMDB->bTruncate)
                            {
                                lNameListLen = 0;
                                memset(kpMCB->lpszErrorWorkBuff, 0x00, ERROR_WORK_BUFF_LEN + 4);
                                NameListInErrorWorkBuff(lpTNData, &lNameListLen);
                                LogMsg(233, lpTIB->lTestNum, lHexOutputLen, kpMCB->lpszErrorWorkBuff, lpCurrELB->lLen);
                                FreeIfNotNull((void**)&lpszResValue);
                                lpTIB->bCmdResult = FAILURE;
                                return;
                            }
                            /* Set lHexOutputLen and lHexValueLen so data is truncated */
                            lHexOutputLen = lpCurrELB->lLen;
                            lHexValueLen = lHexOutputLen * 2;
                        }
                        lOutputDataLen = lpCurrELB->lLen;
                    }
                    /* Check the buffer can fit the data if not then extend it */
                    if(FAILURE == CheckAndExtendBuff(lpBIB, lOutputDataLen))
                    {
                        return;
                    }
                    HexToBuff(lpBIB->lpszBuff + lpBIB->lDataLen, lpszHexValue, lHexOutputLen);
                    lpBIB->lDataLen += lOutputDataLen;
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                break;
            case T_INT8:
                /* Convert the resolved value to an integer in the buffer */
                lOutputDataLen = 1;
                if(TRUE == ResolveIntCmqValue(lpTIB, lpBIB, lpszResValue, lOutputDataLen, lpCurrELB->bCmqValues))
                {
                    lpBIB->lDataLen += lOutputDataLen;
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                else
                {
                    if(FAILURE == CheckValueIsNumeric(lpTNData, lpTIB, lpszResValue))
                    {
                        return;
                    }
                    if(FAILURE == BuffConvertToInt(lpBIB, lpszResValue, lOutputDataLen))
                    {
                        return;
                    }
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                break;
            case T_INT16:
                /* Convert the resolved value to an integer in the buffer */
                lOutputDataLen = 2;
                if(TRUE == ResolveIntCmqValue(lpTIB, lpBIB, lpszResValue, lOutputDataLen, lpCurrELB->bCmqValues))
                {
                    lpBIB->lDataLen += lOutputDataLen;
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                else
                {
                    if(FAILURE == CheckValueIsNumeric(lpTNData, lpTIB, lpszResValue))
                    {
                        return;
                    }
                    if(FAILURE == BuffConvertToInt(lpBIB, lpszResValue, lOutputDataLen))
                    {
                        return;
                    }
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                break;
            case T_INT32:
                /* Convert the resolved value to an integer in the buffer */
                lOutputDataLen = 4;
                if(TRUE == ResolveIntCmqValue(lpTIB, lpBIB, lpszResValue, lOutputDataLen, lpCurrELB->bCmqValues))
                {
                    lpBIB->lDataLen += lOutputDataLen;
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                else
                {
                    if(FAILURE == CheckValueIsNumeric(lpTNData, lpTIB, lpszResValue))
                    {
                        return;
                    }
                    if(FAILURE == BuffConvertToInt(lpBIB, lpszResValue, lOutputDataLen))
                    {
                        return;
                    }
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                break;
            case T_INT64:
                /* Convert the resolved value to an integer in the buffer */
                lOutputDataLen = 8;
                if(TRUE == ResolveIntCmqValue(lpTIB, lpBIB, lpszResValue, lOutputDataLen, lpCurrELB->bCmqValues))
                {
                    lpBIB->lDataLen += lOutputDataLen;
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                else
                {
                    if(FAILURE == CheckValueIsNumeric(lpTNData, lpTIB, lpszResValue))
                    {
                        return;
                    }
                    if(FAILURE == BuffConvertToInt(lpBIB, lpszResValue, lOutputDataLen))
                    {
                        return;
                    }
                    lpTNData->lOutputDataLen = lOutputDataLen;
                }
                break;
            case T_PD:
                if(FAILURE == CheckValueIsHex(lpTNData, lpTIB, lpszResValue))
                {
                    return;
                }

                /* Determine the length of the resolved data and the output field */
                lPackedValueLen = StrlenNull(lpszResValue);
                if(0 == (lPackedValueLen % 2))
                {
                    lPackedOutputLen = lPackedValueLen / 2;
                }
                else
                {
                    lPackedOutputLen = (lPackedValueLen / 2) + 1;
                }
                if(klUNDEFINED == lpCurrELB->lLen)
                {
                    lOutputDataLen = lPackedOutputLen;
                }
                else
                {
                    /* Check for data truncation and only process if its allowed for the message */
                    if(lPackedOutputLen > lpCurrELB->lLen)
                    {
                        if(FALSE == lpTNData->lpMDB->bTruncate)
                        {
                            lNameListLen = 0;
                            memset(kpMCB->lpszErrorWorkBuff, 0x00, ERROR_WORK_BUFF_LEN + 4);
                            NameListInErrorWorkBuff(lpTNData, &lNameListLen);
                            LogMsg(233, lpTIB->lTestNum, lPackedOutputLen, kpMCB->lpszErrorWorkBuff, lpCurrELB->lLen);
                            FreeIfNotNull((void**)&lpszResValue);
                            lpTIB->bCmdResult = FAILURE;
                            return;
                        }
                        /* Set lPackedOutputLen and lPackedValueLen so data is truncated */
                        lPackedOutputLen = lpCurrELB->lLen;
                        lPackedValueLen = lPackedOutputLen * 2;
                    }
                    lOutputDataLen = lpCurrELB->lLen;
                }
                /* Check the buffer can fit the data if not then extend it */
                if(FAILURE == CheckAndExtendBuff(lpBIB, lOutputDataLen))
                {
                    return;
                }
                PackedToBuff(lpBIB->lpszBuff + lpBIB->lDataLen, lpszResValue, lOutputDataLen);
                lpBIB->lDataLen += lOutputDataLen;
                lpTNData->lOutputDataLen = lOutputDataLen;
                break;
            }
            /* Finally update the segment lengths */
            for(lIdx = 0; lIdx < lSegmentIndex; lIdx ++)
            {
                kpMCB->lMsgSetSegmentLen[lIdx] += lOutputDataLen;
            }
        }
        break;
    }

    /* Finished with the resolved value */
    FreeIfNotNull((void**)&lpszResValue);

    /* Recursively call this routine for the first child of this element */
    if(NULL != lpTNData->lpFirstChild)
    {
        GenerateNodeData(lpTIB, lpBIB, (MsgSetTreeNode*)lpTNData->lpFirstChild, lpVIBMQMD, lSegmentIndex + 1);
    }

    /* Recursively call this routine for our next sibling                */
    if(NULL != lpTNData->lpNextSibling)
    {
        GenerateNodeData(lpTIB, lpBIB, (MsgSetTreeNode*)lpTNData->lpNextSibling, lpVIBMQMD, lSegmentIndex);
    }

    /* Scan all the siblings at this level to see if any require their segment length to be set */
    if(NULL == lpTNData->lpNextSibling)
    {
        lpTNWork = lpTNData;
        while(NULL != lpTNWork)
        {
            lSetSeglenFlag = lpTNWork->lFlags & MSTN_SET_SEGLEN_MASK;
            if(0 != lSetSeglenFlag)
            {
                switch(lSetSeglenFlag)
                {
                case MSTN_SET_SEGLEN:
                    lSeglen = kpMCB->lMsgSetSegmentLen[lSegmentIndex-1];
                    break;
                case MSTN_SET_SEGLEN_M1:
                    lSeglen = kpMCB->lMsgSetSegmentLen[lSegmentIndex-1] - 1;
                    break;
                case MSTN_SET_SEGLEN_M2:
                    lSeglen = kpMCB->lMsgSetSegmentLen[lSegmentIndex-1] - 2;
                    break;
                case MSTN_SET_SEGLEN_M4:
                    lSeglen = kpMCB->lMsgSetSegmentLen[lSegmentIndex-1] - 4;
                    break;
                }
                /* Depending on whether we are big or little endian move the */
                /* appropriate bytes into the buffer                         */
                lpszNum = (char*)&lSeglen;
                if(MQENC_INTEGER_NORMAL == kpMCB->lLocalIntegerEncoding)
                {
                    switch(lpTNWork->lOutputDataLen)
                    {
                    case 1:
                        memcpy(lpTNWork->lpszBuffLoc, lpszNum + 3, 1);
                        break;
                    case 2:
                        memcpy(lpTNWork->lpszBuffLoc, lpszNum + 2, 2);
                        break;
                    case 4:
                        memcpy(lpTNWork->lpszBuffLoc, lpszNum, 4);
                        break;
                    }
                }
                else
                {
                    switch(lpTNWork->lOutputDataLen)
                    {
                    case 1:
                        memcpy(lpTNWork->lpszBuffLoc, lpszNum, 1);
                        break;
                    case 2:
                        memcpy(lpTNWork->lpszBuffLoc, lpszNum, 2);
                        break;
                    case 4:
                        memcpy(lpTNWork->lpszBuffLoc, lpszNum, 4);
                        break;
                    }
                }
            }
            lpTNWork = (MsgSetTreeNode*)lpTNWork->lpPrevSibling;
        }
    }
}

/*==========================================================================*/
/* FUNCTION:                                                                */
/*     NameListInErrorWorkBuff                                              */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Walk up the tree to generate the dotted name list for a tree node    */ 
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTN           Pointer to the tree node whose dotted name is to      */
/*                    be generated                                          */
/*     lplNameListLen Pointer to long that gets length of the dotted name   */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
NameListInErrorWorkBuff(MsgSetTreeNode* lpTN, long* lplNameListLen)
{
    if(NULL != lpTN->lpParent)
    {
        NameListInErrorWorkBuff((MsgSetTreeNode*)lpTN->lpParent, lplNameListLen);
    }

    if(N_ROOT != lpTN->eMsgSetTreeNodeType)
    {
        if(N_MSG == lpTN->eMsgSetTreeNodeType)
        {
            *lplNameListLen += sprintf(kpMCB->lpszErrorWorkBuff + *lplNameListLen, "Msg#%d(%s)", lpTN->lIndexSameName, lpTN->lpszNodeValue);
        }
        else
        {
            *lplNameListLen += sprintf(kpMCB->lpszErrorWorkBuff + *lplNameListLen, ".%s[%d]", lpTN->lpszNodeName, lpTN->lIndexSameName);
        }
    }
}

/*==========================================================================*/
/* FUNCTION:                                                                */
/*     CheckValueIsNumeric                                                  */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Check that a value for a tree node is numeric                        */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTN         Pointer to tree node currently being processed          */
/*     lpTIB        Pointer to a test info block                            */
/*     kpszNumStr   Pointer to the string to be tested                      */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     SUCCESS or FAILURE                                                   */
/*==========================================================================*/
long
CheckValueIsNumeric(MsgSetTreeNode* lpTN, TestInfoBlk* lpTIB, const char* const kpszNumStr)
{
    char* lpszCurr;
    long  lNameListLen;
    long  lStrLen;
    long  lIdx;

    /* Establish pointer and length value */
    lpszCurr = (char*)kpszNumStr;
    lStrLen = StrlenNull(kpszNumStr);

    /* If first character is a minus sign skip past it for numeric checking */
    if('-' == *lpszCurr)
    {
        lpszCurr++;
        lStrLen--;
    }
    
    for(lIdx = 0; lIdx < lStrLen; lIdx++)
    {
        if(! isdigit(*lpszCurr))
        {
            lNameListLen = 0;
            memset(kpMCB->lpszErrorWorkBuff, 0x00, ERROR_WORK_BUFF_LEN + 4);
            NameListInErrorWorkBuff(lpTN, &lNameListLen);
            LogMsg(198, lpTIB->lTestNum, kpszNumStr, kpMCB->lpszErrorWorkBuff);
            lpTIB->bCmdResult = FAILURE;
            return FAILURE;
        }
        lpszCurr++;
    }
    return SUCCESS;
}

/*==========================================================================*/
/* FUNCTION:                                                                */
/*     CheckValueIsHex                                                      */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Check that a value for a tree node is valid hexadecimal              */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpTN         Pointer to tree node currently being processed          */
/*     lpTIB        Pointer to a test info block                            */
/*     kpszNumStr   Pointer to the string to be tested                      */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     SUCCESS or FAILURE                                                   */
/*==========================================================================*/
long
CheckValueIsHex(MsgSetTreeNode* lpTN, TestInfoBlk* lpTIB, const char* const kpszHexStr)
{
    char* lpszCurr;
    long  lNameListLen;
    long  lStrLen;
    long  lIdx;

    lpszCurr = (char*)kpszHexStr;
    lStrLen = StrlenNull(kpszHexStr);
    for(lIdx = 0; lIdx < lStrLen; lIdx++)
    {
        if(! isxdigit(*lpszCurr))
        {
            lNameListLen = 0;
            memset(kpMCB->lpszErrorWorkBuff, 0x00, ERROR_WORK_BUFF_LEN + 4);
            NameListInErrorWorkBuff(lpTN, &lNameListLen);
            LogMsg(236, lpTIB->lTestNum, kpszHexStr, kpMCB->lpszErrorWorkBuff);
            lpTIB->bCmdResult = FAILURE;
            return FAILURE;
        }
        lpszCurr++;
    }
    return SUCCESS;
}
void


ResolveNodeValue(MsgSetTreeNode* lpTN, TestInfoBlk* lpTIB,
                 char** lplpszResStr, char** lplpszVarStr,
                 long bDisplayMsg61)
{
    if(NULL == *lplpszVarStr)
    {
        *lplpszResStr = CloneString(NULL);
    }
    else if(0 == memcmp(kpszSEGLEN, *lplpszVarStr, strlen(kpszSEGLEN) + 1))
    {
        lpTN->lFlags = MSTN_SET_SEGLEN;
        *lplpszResStr = CloneString(NULL);
    }
    else if(0 == memcmp(kpszSEGLENM1, *lplpszVarStr, strlen(kpszSEGLENM1) + 1))
    {
        lpTN->lFlags = MSTN_SET_SEGLEN_M1;
        *lplpszResStr = CloneString(NULL);
    }
    else if(0 == memcmp(kpszSEGLENM2, *lplpszVarStr, strlen(kpszSEGLENM2) + 1))
    {
        lpTN->lFlags = MSTN_SET_SEGLEN_M2;
        *lplpszResStr = CloneString(NULL);
    }
    else if(0 == memcmp(kpszSEGLENM4, *lplpszVarStr, strlen(kpszSEGLENM4) + 1))
    {
        lpTN->lFlags = MSTN_SET_SEGLEN_M4;
        *lplpszResStr = CloneString(NULL);
    }
    else
    {
        ResolveValue(lpTIB, lplpszResStr, lplpszVarStr, bDisplayMsg61);
    }
}

/*==========================================================================*/
/* Debugging routines                                                       */
/*==========================================================================*/
void
DisplayMsgTree(char* lpszLabel, MsgSetTreeNode* lpTNOut, 
               long lRecursionDepth,
               long lMsgNumber)
{
    /* If we've already failed then go no further */
    if(FAILURE == kpMCB->lProcessing)
    {
        return;
    }

    DisplayNodeInfo(lpszLabel, lMsgNumber, lRecursionDepth, lpTNOut);
    
    /* Recursively call this routine for the first child of this element */
    if(NULL != lpTNOut->lpFirstChild)
    {
        /* If called for a child element then the recursion depth has increased */
        DisplayMsgTree(lpszLabel, (MsgSetTreeNode*)lpTNOut->lpFirstChild, lRecursionDepth + 1, lMsgNumber);
    }

    /* Recursively call this routine for our next sibling                */
    if(NULL != lpTNOut->lpNextSibling)
    {
        /* If its a <Msg> sibling then increase the message number */
        if((1 == lRecursionDepth) &&
           (0 == memcmp(kpszMSG, lpTNOut->lpszNodeName, strlen(kpszMSG) + 1)))
        {
            lMsgNumber++;
        }
        DisplayMsgTree(lpszLabel, (MsgSetTreeNode*)lpTNOut->lpNextSibling, lRecursionDepth, lMsgNumber);
    }
}

/*--------------------------------------------------------------------------*/
void
DisplayNodeInfo(char* lpszLabel, long lMsgNumber, long lRecursionDepth, MsgSetTreeNode* lpTN)
{
    MsgSetTreeNode* lpTNWork;

    printf("%s-%d:%d:%d:%d:%d", lpszLabel, lMsgNumber, lRecursionDepth, lpTN->lNodeDepth, lpTN->lIndexSameName, lpTN->lpMSLL->eMSLLBlockType);

    DisplayNames(lpTN);

    printf("=%s", lpTN->lpszNodeValue);
     
    if((NULL != lpTN->lpFirstChild) ||
       (NULL != lpTN->lpLastChild) ||
       (NULL != lpTN->lpNextSibling) ||
       (NULL != lpTN->lpPrevSibling) ||
       (NULL != lpTN->lpParent))
    {
        printf("\n");
    }
    if(NULL != lpTN->lpNextRequiredMSLL)
    {
        printf("    NR=%s[%d]", lpTN->lpNextRequiredMSLL->lpszNameOrType);
    }
    if(NULL != lpTN->lpFirstChild)
    {
        lpTNWork = (MsgSetTreeNode*)lpTN->lpFirstChild;
        printf("    FC=%s[%d]", lpTNWork->lpszNodeName, lpTNWork->lIndexSameName);
    }
    if(NULL != lpTN->lpLastChild)
    {
        lpTNWork = (MsgSetTreeNode*)lpTN->lpLastChild;
        printf("    LC=%s[%d]", lpTNWork->lpszNodeName, lpTNWork->lIndexSameName);
    }
    if(NULL != lpTN->lpNextSibling)
    {
        lpTNWork = (MsgSetTreeNode*)lpTN->lpNextSibling;
        printf("    NS=%s[%d]", lpTNWork->lpszNodeName, lpTNWork->lIndexSameName);
    }
    if(NULL != lpTN->lpPrevSibling)
    {
        lpTNWork = (MsgSetTreeNode*)lpTN->lpPrevSibling;
        printf("    PS=%s[%d]", lpTNWork->lpszNodeName, lpTNWork->lIndexSameName);
    }
    if(NULL != lpTN->lpNextSiblingSameName)
    {
        lpTNWork = (MsgSetTreeNode*)lpTN->lpNextSiblingSameName;
        printf("    NSSN=%s[%d]", lpTNWork->lpszNodeName, lpTNWork->lIndexSameName);
    }
    if(NULL != lpTN->lpPrevSiblingSameName)
    {
        lpTNWork = (MsgSetTreeNode*)lpTN->lpPrevSiblingSameName;
        printf("    PSSN=%s[%d]", lpTNWork->lpszNodeName, lpTNWork->lIndexSameName);
    }
    if(NULL != lpTN->lpParent)
    {
        lpTNWork = (MsgSetTreeNode*)lpTN->lpParent;
        printf("    P=%s[%d]", lpTNWork->lpszNodeName, lpTNWork->lIndexSameName);
    }
    printf("\n");

    DisplayBlockInfo(lpTN);
}

/*--------------------------------------------------------------------------*/
void
DisplayNames(MsgSetTreeNode* lpTN)
{
    if(NULL != lpTN->lpParent)
    {
        DisplayNames((MsgSetTreeNode*)lpTN->lpParent);
    }

    printf(".%s[%d]", lpTN->lpszNodeName, lpTN->lIndexSameName);
}

/*--------------------------------------------------------------------------*/
void
DisplayBlockInfo(MsgSetTreeNode* lpTN)
{
    if(NULL != lpTN->lpMDB)
    {
        printf("    MDB=%s    ", lpTN->lpMDB->lpszType);
        if(NULL != lpTN->lpSDB)
        {
            printf("SDB=%s    ", lpTN->lpSDB->lpszType);
        }
        if(NULL != lpTN->lpSGB)
        {
            printf("SGB=%s", lpTN->lpSGB->lpszType);
        }
        printf("\n");
    }
    else
    {
        printf("    No MDB. Should be root\n");
    }
}

