/*==========================================================================*/
/* MsgTestTime.c contains the interval and timing related routines          */
/*==========================================================================*/
/* (C) Copyright Tim Armstrong 2003, 2004, 2005, 2006, 2007, 2008           */
/*==========================================================================*/
/* Includes                                                                 */
/*==========================================================================*/
#ifndef __TJAINC_MSGTEST_H
#include "MsgTest.h"
#endif /* __TJAINC_MSGTEST_H */


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     ReportStats                                                          */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Report interval and message rate statistics                          */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     N/A                                                                  */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
ReportStats()
{
    IntervalInfoBlk* lpIIB;
    IntervalInfoRec* lpIIR;
    char             lpszVarInfo[VAR_INFO_LEN + 4];
    long             lStartOnlyCount;
    long             lEndOnlyCount;
    long             lIntervalCount;
    long             lValidIntervalCount;
    long             lEndBeforeStartCount;
    long             lIdx;
    long             lJdx;
    long             lSortMax;
    long             lSortMin;
    long             lJdxMax;
    long             lMedianPos;
    long             lMedianRem;
    long             lSIQCount;
    long             lSIQRem;
    long             lSIQLower;
    long             lSIQUpper;
    long             lSIQPlusMinus;
    long             lSIQRange;
    double*          dpInterval;
    double           dTemp;
    double           dDiff;
    double           dMin;
    double           dMax;
    double           dMedian;
    double           dAverage;
    double           dStdDev;
    double           dStdDevTotal;
    double           dTotal;
    double           dSIQMin;
    double           dSIQMax;
    double           dSIQMedian;
    double           dSIQAverage;
    double           dSIQStdDev;
    double           dSIQStdDevTotal;
    double           dSIQTotal;
    long             bIsSorted;
    Int64            i64Interval;

    /* Report message rate information */
    if(TRUE == kpMCB->bReportMessageRate)
    {
        ReportMsg(107);
        ReportMsg(128, kpMCB->lMessageRateAchieved + kpMCB->lMessageRateMissed,
                  kpMCB->lMessageRateAchieved, kpMCB->lMessageRateMissed);
    }

    /* Report interval related statistics */
    ReportMsg(107);
    ReportMsg(93);
    memset(lpszVarInfo, 0x00, VAR_INFO_LEN + 4);
    lpIIB = kpMCB->lpIIBHead;
    if(! lpIIB)
    {
        ReportMsg(108);
    }
    else
    {
        while(NULL != lpIIB)
        {
            ReportMsg(97, lpIIB->lpszIntervalName);
            lpIIR = lpIIB->lpIIRHead;
            if(! lpIIR)
            {
                ReportMsg(109);
            }
            else
            {
                /* Initialise counters  and statistical values */
                lStartOnlyCount = 0;
                lEndOnlyCount = 0;
                lIntervalCount = 0;
                lValidIntervalCount = 0;
                lEndBeforeStartCount = 0;
                lSIQCount = 0;
                lSIQPlusMinus = 0;
                lSIQRange = 0;
                dMin = 0;
                dMax = 0;
                dMedian = 0;
                dAverage = 0;
                dStdDev = 0;
                dStdDevTotal = 0;
                dTotal = 0;
                dSIQMin = 0;
                dSIQMax = 0;
                dSIQMedian = 0;
                dSIQStdDev = 0;
                dSIQAverage = 0;
                dSIQStdDevTotal = 0;
                dSIQTotal = 0;
                /* Allocate an array to hold calculated intervals */
                dpInterval = malloc(lpIIB->lIIRCount * sizeof(double));
                /* If able to allocate the calculated intervals array then fill it */
                if(dpInterval)
                {
                    /* Initialise the calculated intervals array */
                    for(lIdx = 0; lIdx < lpIIB->lIIRCount; lIdx++)
                    {
                        dpInterval[lIdx] = 0;
                    }
                    /* Process the interval records into the calculated intervals array */
                    while(NULL != lpIIR)
                    {
                        lIntervalCount++;
                        if((TRUE == lpIIR->bHaveStart) && (FALSE == lpIIR->bHaveEnd))
                        {
                            lStartOnlyCount++;
                        }
                        else if((FALSE == lpIIR->bHaveStart) && (TRUE == lpIIR->bHaveEnd))
                        {
                            lEndOnlyCount++;
                        }
                        else
                        {
                            if(lpIIR->bEndBeforeStart)
                            {
                                lEndBeforeStartCount++;
                                i64Interval = lpIIR->i64IntervalStart - lpIIR->i64IntervalEnd;
                            }
                            else
                            {
                                i64Interval = lpIIR->i64IntervalEnd - lpIIR->i64IntervalStart;
                            }
                            dpInterval[lValidIntervalCount] = ((double)i64Interval) / ((double)kpMCB->i64CountFrequency);
                            dTotal += dpInterval[lValidIntervalCount];
                            lValidIntervalCount++;
                        }
                        lpIIR = (IntervalInfoRec*)lpIIR->lpIIRNext;
                    }
                    /* If more than one valid observation generate various statistics */
                    if(1 < lValidIntervalCount)
                    {
                        /* Sort the values so we can get the median */
                        /* and semi-interquartile based statistics  */
                        lSortMin = 0;
                        lSortMax = lValidIntervalCount;
                        for(lIdx = lSortMin; lIdx < lValidIntervalCount; lIdx++)
                        {
                            /* Move the largest value to the end of the */
                            bIsSorted = TRUE;
                            lJdxMax = lSortMax - 1;
                            for(lJdx = lSortMin; lJdx < lJdxMax; lJdx++)
                            {
                                if(dpInterval[lJdx] > dpInterval[lJdx + 1])
                                {
                                    dTemp = dpInterval[lJdx + 1];
                                    dpInterval[lJdx + 1] = dpInterval[lJdx];
                                    dpInterval[lJdx] = dTemp;
                                    bIsSorted = FALSE;
                                }
                            }
                            /* If the array is already sorted thats it */
                            if(bIsSorted)
                            {
                                break; /* for(lIdx... */
                            }
                            /* Reduce upper limit as max value now occupies the end */
                            lSortMax--;
                            /* Move the smallest value to the start */
                            bIsSorted = TRUE;
                            for(lJdx = lSortMax - 1; lJdx > lSortMin; lJdx--)
                            {
                                if(dpInterval[lJdx] < dpInterval[lJdx-1])
                                {
                                    dTemp = dpInterval[lJdx - 1];
                                    dpInterval[lJdx - 1] = dpInterval[lJdx];
                                    dpInterval[lJdx] = dTemp;
                                    bIsSorted = FALSE;
                                }
                            }
                            /* If the array is already sorted thats it */
                            if(bIsSorted)
                            {
                                break; /* for(lIdx... */
                            }
                            /* Increase lower limit as min value now occupies the start */
                            lSortMin++;
                        }
                        /* Get the minimum and maximum values */
                        dMin = dpInterval[0];
                        dMax = dpInterval[lValidIntervalCount - 1];
                        /* If we have an even number of intervals the median is */
                        /* an average of the two intervals either side of a     */
                        /* theoretical centre. Otherwise it is the centre value */
                        lMedianPos = lValidIntervalCount / 2;
                        lMedianRem = lValidIntervalCount % 2;
                        if(0 == lMedianRem)
                        {
                            dMedian = (dpInterval[lMedianPos-1] + dpInterval[lMedianPos]) / 2;
                        }
                        else
                        {
                            dMedian = dpInterval[lMedianPos];
                        }
                        /* Calculate the average */
                        dAverage = dTotal / ((double)lValidIntervalCount);
                        /* Now we have the average calculate the standard deviation */
                        for(lIdx = 0; lIdx < lValidIntervalCount; lIdx++)
                        {
                            dDiff = dAverage - dpInterval[lIdx];
                            dStdDevTotal += dDiff * dDiff;
                        }
                        dStdDev = sqrt((dStdDevTotal / ((double)lValidIntervalCount)));
                        /* If we have four or more valid intervals we can calculate */
                        /* statistics for the Semi Interquartile Deviation which    */
                        /* gives the statistical picture without the extreme values */
                        if(4 <= lValidIntervalCount)
                        {
                            lSIQRem = lValidIntervalCount % 4;
                            switch(lSIQRem)
                            {
                            case 0:
                                lSIQCount = (lValidIntervalCount / 2);
                                break;
                            case 1:
                                lSIQCount = (lValidIntervalCount / 2) + 1;
                                break;
                            case 2:
                                lSIQCount = (lValidIntervalCount / 2) - 1;
                                break;
                            case 3:
                                lSIQCount = (lValidIntervalCount / 2);
                                break;
                            }
                            lSIQPlusMinus = lSIQCount / 2;
                            if(0 == lMedianRem)
                            {
                                lSIQLower = ((lMedianPos - 1) - lSIQPlusMinus) + 1;
                                lSIQUpper = (lMedianPos + lSIQPlusMinus) - 1;
                            }
                            else
                            {
                                lSIQLower = lMedianPos - lSIQPlusMinus;
                                lSIQUpper = lMedianPos + lSIQPlusMinus;
                            }
                            lSIQRange = (lSIQUpper - lSIQLower) + 1;
                            /* Get the minimum and maximum values */
                            dSIQMin = dpInterval[lSIQLower];
                            dSIQMax = dpInterval[lSIQUpper];
                            /* Median value does not change */
                            dSIQMedian = dMedian;
                            /* Calculate the SIQ total and average */
                            dSIQTotal = 0;
                            for(lIdx = lSIQLower; lIdx <= lSIQUpper; lIdx++)
                            {
                                dSIQTotal += dpInterval[lIdx];
                            }
                            dSIQAverage = dSIQTotal / ((double)lSIQRange);
                            /* Now we have the average calculate the standard deviation */
                            for(lIdx = lSIQLower; lIdx <= lSIQUpper; lIdx++)
                            {
                                dDiff = dSIQAverage - dpInterval[lIdx];
                                dSIQStdDevTotal += dDiff * dDiff;
                            }
                            dSIQStdDev = sqrt((dSIQStdDevTotal / ((double)lSIQRange)));
                        }
                    }

                    /* Finished calculating statistics. Free calculated  */
                    /* interval work area                                */
                    FreeIfNotNull((void**)&dpInterval);

                    /* Depending on the number of valid intervals report */
                    /* the various interval related statistics           */
                    if(0 == lValidIntervalCount)
                    {
                        ReportMsg(98);
                    }
                    else if(1 == lValidIntervalCount)
                    {
                        ReportMsg(99);
                        ReportMsg(101, dTotal, lValidIntervalCount);
                    }
                    else if((1 < lValidIntervalCount) && (4 > lValidIntervalCount))
                    {
                        ReportMsg(99);
                        ReportMsg(101, dTotal, lValidIntervalCount);
                        ReportMsg(102, dAverage, dMedian);
                        ReportMsg(103, dMin, dMax, dStdDev);
                    }
                    else
                    {
                        ReportMsg(99);
                        ReportMsg(101, dTotal, lValidIntervalCount);
                        ReportMsg(102, dAverage, dMedian);
                        ReportMsg(103, dMin, dMax, dStdDev);
                        ReportMsg(100);
                        ReportMsg(101, dSIQTotal, lSIQRange);
                        ReportMsg(102, dSIQAverage, dSIQMedian);
                        ReportMsg(103, dSIQMin, dSIQMax, dSIQStdDev);
                    }

                    /* Report number of invalid and out of sequence intervals */
                    if(0 < lStartOnlyCount)
                    {
                        ReportMsg(104, lStartOnlyCount);
                    }
                    if(0 < lEndOnlyCount)
                    {
                        ReportMsg(105, lEndOnlyCount);
                    }
                    if(0 < lEndBeforeStartCount)
                    {
                        ReportMsg(106, lEndBeforeStartCount);
                    }
                }
                else
                {
                    LogMsg(3, lpIIB->lIIRCount * sizeof(double), "IntervalStats", __LINE__, __FILE__);
                    kpMCB->lProcessing = FAILURE;
                    return;
                }
            }
            lpIIB = (IntervalInfoBlk*)lpIIB->lpIIBNext;
        }
    }
    ReportMsg(107);
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     CreateAndAddIntervalInfoRec                                          */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Create an IntervalInfoRec and add it to the IIB's IIR chain          */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpIIB Pointer to Interval Info block                                 */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     Pointer to the IntervalInfoRec                                       */
/*==========================================================================*/
IntervalInfoRec*
CreateAndAddIntervalInfoRec(IntervalInfoBlk* lpIIB)
{
    IntervalInfoRec* lpIIR;

    lpIIR = malloc(sizeof(IntervalInfoRec));
    if(lpIIR)
    {
        memset(lpIIR, 0x00, sizeof(IntervalInfoRec));
        lpIIR->lpIIRNext = NULL;
        lpIIR->lpIIRPrev = NULL;
        lpIIR->i64IntervalStart = 0;
        lpIIR->i64IntervalEnd = 0;
        lpIIR->bHaveStart = FALSE;
        lpIIR->bHaveEnd = FALSE;
        lpIIR->bEndBeforeStart = FALSE;
        AddToTailOfList((LinkedListBlk**)&lpIIB->lpIIRHead,
                        (LinkedListBlk**)&lpIIB->lpIIRTail,
                        (LinkedListBlk*)lpIIR);
        lpIIB->lIIRCount++;
    }
    else
    {
        ConsoleMsg(3, sizeof(IntervalInfoRec), "IntervalInfoRec", __LINE__, __FILE__);
        kpMCB->lProcessing = FAILURE;
        return NULL;
    }

    return lpIIR;
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     SetIntervalStart                                                     */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Set the interval start value in the interval info rec                */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpIIR Pointer to Interval Info rec                                   */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
SetIntervalStart(IntervalInfoRec* lpIIR)
{
    /* Set have start value flag and get the Performance Counter value */
    lpIIR->bHaveStart = TRUE;
    GetClockCount(&lpIIR->i64IntervalStart);
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     SetIntervalEnd                                                       */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Set the interval end value in the interval info rec                  */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpIIR Pointer to Interval Info rec                                   */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
SetIntervalEnd(IntervalInfoRec* lpIIR)
{
    /* Set have end value flag and get the Performance Counter value */
    lpIIR->bHaveEnd = TRUE;
    GetClockCount(&lpIIR->i64IntervalEnd);
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     SetInterval                                                          */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Write an interval value into an 8 byte area                          */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpIIR Pointer to Interval Info rec                                   */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
SetInterval(void* vp8Bytes)
{
    GetClockCount(vp8Bytes);
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     CalcInterval                                                         */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Calculate the interval value                                         */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpIIR Pointer to Interval Info rec                                   */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     Interval value as a double                                           */
/*==========================================================================*/
double
CalcInterval(IntervalInfoRec* lpIIR, double* lpdStart,
             double* lpdEnd, double* lpdInterval, double* lpdRunSeconds)
{
    IntervalInfoRec* lpIIRPrev;
    Int64            i64Interval;

    /* Calculate times relative to the previous interval */
    lpIIRPrev = (IntervalInfoRec*)lpIIR->lpIIRPrev;
    if(NULL == lpIIRPrev)
    {
        *lpdStart = 0;
        *lpdEnd = 0;
    }
    else
    {
        i64Interval = lpIIR->i64IntervalStart - lpIIRPrev->i64IntervalStart;
        *lpdStart = ((double)i64Interval) / ((double)kpMCB->i64CountFrequency);
        /* As out of sequence processing can occur it is possible to get  */
        /* a negative time difference value. Correct by multiplying by -1 */
        if(0 > *lpdStart)
        {
            *lpdStart *= -1;
        }
        i64Interval = lpIIR->i64IntervalEnd - lpIIRPrev->i64IntervalEnd;
        *lpdEnd = ((double)i64Interval) / ((double)kpMCB->i64CountFrequency);
    }

    /* Calculate the number of seconds from the RunStart timestamp to the interval start */
    i64Interval = lpIIR->i64IntervalStart - kpMCB->i64RunStart;
    *lpdRunSeconds = ((double)i64Interval) / ((double)kpMCB->i64CountFrequency);

    /* Calculate the actual interval */
    i64Interval = lpIIR->i64IntervalEnd - lpIIR->i64IntervalStart;
    *lpdInterval = ((double)i64Interval) / ((double)kpMCB->i64CountFrequency);

    return *lpdInterval;
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     CalcCmdExecTime                                                      */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Calculate the interval value                                         */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpIIR Pointer to Interval Info rec                                   */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     Interval value as a double                                           */
/*==========================================================================*/
double
CalcCmdExecTime(TestInfoBlk* lpTIB)
{
    Int64            i64Current;
    Int64            i64Result;
    double           dResult;

    GetClockCount(&i64Current);
    i64Result = i64Current - lpTIB->i64CommandInit;
    dResult = ((double)i64Result) / ((double)kpMCB->i64CountFrequency);

    return dResult;
}

/*==========================================================================*/
/* FUNCTION:                                                                */
/*     ControlMessageRate                                                   */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Control the rate at which messages get sent by sleeping the          */
/*     necessary amount of time                                             */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lpMRB Pointer to a message rate block                                */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
ControlMessageRate(TestInfoBlk* lpTIB, MessageRateBlk* lpMRB)
{
    double   dRandomValue;
    double   dRandomMultiplier;
    long     lNewNum;
    long     lCalcMillis;
    long     lWaitLoopCount;
    Int64    i64CalcMillis;

    if(TRUE == lpMRB->bInitForLoop)
    {
        /* First time through just wait until the second changes */
        /* in this way we should have nearly a full second for   */
        /* the first batch of messages. Note on a busy system    */
        /* or one with large time slices this method is not      */
        /* going to work properly                                */
        if(TRUE == lpMRB->bFirstTime)
        {
            time(&lpMRB->oProcessingSecond);
            time(&lpMRB->oNowSecond);
            while(lpMRB->oProcessingSecond == lpMRB->oNowSecond)
            {
                time(&lpMRB->oNowSecond);
                /* Keep handing control back to the operating sytem */
                /* rather than sitting in a full on loop            */
                SleepMillis(1);
            }
            lpMRB->bFirstTime = FALSE;
        }

        /* Resolve variables */
        ResolveValue(lpTIB, &lpMRB->lpszResFrom, &lpMRB->lpszFrom, TRUE);
        ResolveValue(lpTIB, &lpMRB->lpszResTo, &lpMRB->lpszTo, TRUE);
        ResolveValue(lpTIB, &lpMRB->lpszResIncr, &lpMRB->lpszIncr, TRUE);
        ResolveValue(lpTIB, &lpMRB->lpszResInterval, &lpMRB->lpszInterval, TRUE);
        if(FAILURE == lpTIB->bCmdResult)
        {
            return;
        }

        /* Convert to numeric values */
        lpMRB->lFrom = ConvNum(lpTIB, lpMRB->lpszResFrom, "From");
        lpMRB->lTo = ConvNum(lpTIB, lpMRB->lpszResTo, "To");
        lpMRB->lIncr = ConvNum(lpTIB, lpMRB->lpszResIncr, "Incr");
        lpMRB->lInterval = ConvNum(lpTIB, lpMRB->lpszResInterval, "Interval");
        if(FAILURE == lpTIB->bCmdResult)
        {
            return;
        }

        /* Copy the current time into the other timestamps      */
        memcpy(&lpMRB->oFirstSecond, &lpMRB->oNowSecond, sizeof(time_t));
        memcpy(&lpMRB->oProcessingSecond, &lpMRB->oNowSecond, sizeof(time_t));
        memcpy(&lpMRB->oIncrAtSecond, &lpMRB->oNowSecond, sizeof(time_t));
        lpMRB->oIncrAtSecond += lpMRB->lInterval;

        /* Determine the number of messages to be sent this second */
        if(lpMRB->lFrom == lpMRB->lTo)
        {
            lpMRB->lMessagesThisSecond = lpMRB->lFrom;
            lpMRB->lMessagesLeft = lpMRB->lMessagesThisSecond - 1;
            lpTIB->lTestCurrentMsgRate = lpMRB->lMessagesThisSecond;
        }
        else
        {
            switch(lpMRB->eMsgRateMode)
            {
            case M_BURST:
            case M_PACED:
            case M_RANDOM:
                /* Seed the random number generator */
                srand(lpMRB->lFrom);
                /* Some implementations seem to return the seed value the first time */
                rand();
                /* Generate a random number */
                dRandomValue = ((double) rand()) / ((double) RAND_MAX);
                dRandomMultiplier = (double)lpMRB->lTo - lpMRB->lFrom + 1;
                lNewNum = (long)(dRandomValue * dRandomMultiplier);
                lNewNum += lpMRB->lFrom;
                lpMRB->lMessagesThisSecond = lNewNum;
                lpMRB->lMessagesLeft = lpMRB->lMessagesThisSecond - 1;
                break;
            case M_INCREMENTAL:
                lpMRB->lMessagesThisSecond = lpMRB->lFrom;
                lpMRB->lMessagesLeft = lpMRB->lMessagesThisSecond - 1;
                break;
            }
            lpTIB->lTestCurrentMsgRate = lpMRB->lMessagesThisSecond;
        }

        /* Log the selected message rate */
        LogMsg(132, lpTIB->lTestNum, lpMRB->lMessagesThisSecond, lpMRB->lFrom, lpMRB->lTo);

        /* Set the high resolution timer stuff for PACED, RANDOM and INCREMENTAL */
        GetClockCount(&lpMRB->i64ProcessingStart);
        memcpy(&lpMRB->i64ProcessingNow, &lpMRB->i64ProcessingStart, sizeof(Int64));
        memcpy(&lpMRB->i64ProcessingNext, &lpMRB->i64ProcessingStart, sizeof(Int64));

        /* Calculate high resolution timer stuff for PACED, RANDOM and INCREMENTAL */
        lpMRB->i64ProcessingIncr = kpMCB->i64CountFrequency / lpMRB->lMessagesThisSecond;
        switch(lpMRB->eMsgRateMode)
        {
        case M_BURST:
            lpMRB->i64ProcessingIncr = 0;
            break;
        case M_PACED:
            lpMRB->i64ProcessingNext += lpMRB->i64ProcessingIncr;
            break;
        case M_RANDOM:
            /* Calculate the value of the random increment */
            dRandomValue = ((double) rand()) / ((double) RAND_MAX);
            /* Multiply to cater for the fact that random averages out at 50%  */
            /* but not too much so we dont miss too many message rate targets. */
            /* The more messages the more self correcting we are likely to be. */
            if(20 < lpMRB->lMessagesThisSecond)
            {
                dRandomMultiplier = ((double)lpMRB->i64ProcessingIncr) * 1.8;
            }
            else if(10 < lpMRB->lMessagesThisSecond)
            {
                dRandomMultiplier = ((double)lpMRB->i64ProcessingIncr) * 1.4;
            }
            else
            {
                dRandomMultiplier = (double)lpMRB->i64ProcessingIncr;
            }
            lpMRB->i64RandomIncr = (Int64)(dRandomValue * dRandomMultiplier);
            /* If its less than half a millisecond then make it a half millisecond */
            if(lpMRB->i64RandomIncr < (kpMCB->i64OneHalfMilliSecond))
            {
                lpMRB->i64RandomIncr = kpMCB->i64OneHalfMilliSecond;
            }
            lpMRB->i64ProcessingNext += lpMRB->i64RandomIncr;
            break;
        case M_INCREMENTAL:
            lpMRB->i64ProcessingNext += lpMRB->i64ProcessingIncr;
            break;
        }

        /* Set up for message rate is now done */
        lpMRB->bInitForLoop = FALSE;
    }
    else
    {
        /* Get the current time in seconds */
        time(&lpMRB->oNowSecond);
        GetClockCount(&lpMRB->i64ProcessingNow);
        /* If we are still in the same second work out what to do */
        /* depending on the mode of operation requested           */
        if(lpMRB->oProcessingSecond == lpMRB->oNowSecond)
        {
            if(0 < lpMRB->lMessagesLeft)
            {
                switch(lpMRB->eMsgRateMode)
                {
                /* In BURST mode we pump out messages as fast as we   */
                /* can until we reach the desired number of messages  */
                /* at which point in time we wait out the rest of the */
                /* current processing second                          */
                case M_BURST:
                    lpMRB->lMessagesLeft--;
                    break;
                case M_PACED:
                    if(lpMRB->i64ProcessingNow >= lpMRB->i64ProcessingNext)
                    {
                        while(lpMRB->i64ProcessingNow >= lpMRB->i64ProcessingNext)
                        {
                            lpMRB->i64ProcessingNext += lpMRB->i64ProcessingIncr;
                        }
                    }
                    else
                    {
                        i64CalcMillis = ((lpMRB->i64ProcessingNext - lpMRB->i64ProcessingNow) * 1000) / kpMCB->i64CountFrequency;
                        lCalcMillis = (long)i64CalcMillis;
                        if(5 < lCalcMillis)
                        {
                            lCalcMillis -= 5;
                            SleepMillis(lCalcMillis);
                        }
                        while(lpMRB->i64ProcessingNext > lpMRB->i64ProcessingNow)
                        {
                            SleepMillis(1);
                            GetClockCount(&lpMRB->i64ProcessingNow);
                        }
                        lpMRB->i64ProcessingNext += lpMRB->i64ProcessingIncr;
                    }
                    lpMRB->lMessagesLeft--;
                    break;
                case M_RANDOM:
                    if(lpMRB->i64ProcessingNow >= lpMRB->i64ProcessingNext)
                    {
                        while(lpMRB->i64ProcessingNow >= lpMRB->i64ProcessingNext)
                        {
                            lpMRB->i64ProcessingNext += lpMRB->i64RandomIncr;
                        }
                    }
                    else
                    {
                        i64CalcMillis = ((lpMRB->i64ProcessingNext - lpMRB->i64ProcessingNow) * 1000) / kpMCB->i64CountFrequency;
                        lCalcMillis = (long)i64CalcMillis;
                        if(5 < lCalcMillis)
                        {
                            lCalcMillis -= 5;
                            SleepMillis(lCalcMillis);
                        }
                        while(lpMRB->i64ProcessingNext > lpMRB->i64ProcessingNow)
                        {
                            SleepMillis(1);
                            GetClockCount(&lpMRB->i64ProcessingNow);
                        }
                    }
                    /* Calculate the value of the random increment */
                    dRandomValue = ((double) rand()) / ((double) RAND_MAX);
                    /* Multiply to cater for the fact that random averages out at 50%  */
                    /* but not too much so we dont miss too many message rate targets. */
                    /* The more messages the more self correcting we are likely to be. */
                    if(20 < lpMRB->lMessagesThisSecond)
                    {
                        dRandomMultiplier = ((double)lpMRB->i64ProcessingIncr) * 1.8;
                    }
                    else if(10 < lpMRB->lMessagesThisSecond)
                    {
                        dRandomMultiplier = ((double)lpMRB->i64ProcessingIncr) * 1.4;
                    }
                    else
                    {
                        dRandomMultiplier = (double)lpMRB->i64ProcessingIncr;
                    }
                    lpMRB->i64RandomIncr = (Int64)(dRandomValue * dRandomMultiplier);
                    /* If its less than half a millisecond then make it a half millisecond */
                    if(lpMRB->i64RandomIncr < (kpMCB->i64OneHalfMilliSecond))
                    {
                        lpMRB->i64RandomIncr = kpMCB->i64OneHalfMilliSecond;
                    }
                    lpMRB->i64ProcessingNext += lpMRB->i64RandomIncr;
                    lpMRB->lMessagesLeft--;
                    break;
                case M_INCREMENTAL:
                    if(lpMRB->i64ProcessingNow >= lpMRB->i64ProcessingNext)
                    {
                        while(lpMRB->i64ProcessingNow >= lpMRB->i64ProcessingNext)
                        {
                            lpMRB->i64ProcessingNext += lpMRB->i64ProcessingIncr;
                        }
                    }
                    else
                    {
                        i64CalcMillis = ((lpMRB->i64ProcessingNext - lpMRB->i64ProcessingNow) * 1000) / kpMCB->i64CountFrequency;
                        lCalcMillis = (long)i64CalcMillis;
                        if(5 < lCalcMillis)
                        {
                            lCalcMillis -= 5;
                            SleepMillis(lCalcMillis);
                        }
                        while(lpMRB->i64ProcessingNext > lpMRB->i64ProcessingNow)
                        {
                            SleepMillis(1);
                            GetClockCount(&lpMRB->i64ProcessingNow);
                        }
                        lpMRB->i64ProcessingNext += lpMRB->i64ProcessingIncr;
                    }
                    lpMRB->lMessagesLeft--;
                    break;
                }
            }
            else
            {
                /* Wait out the rest of the processing second */
                lWaitLoopCount = 0;
                while(lpMRB->oProcessingSecond == lpMRB->oNowSecond)
                {
                    time(&lpMRB->oNowSecond);
                    /* Keep handing control back to the operating sytem */
                    /* rather than sitting in a full on loop            */
                    SleepMillis(1);
                    lWaitLoopCount++;
                }
                LogMsg(130, lWaitLoopCount);
            }
        }

        /* Current design assumes a minimum rate of 1 message per    */
        /* second. For higher intervals a combination of <RandomNum> */
        /* and <WaitOnTime> can be used.                             */

        /* If we have changed seconds reset the current processing */
        /* second, determine if we did all the messages we were    */
        /* asked to and finally reset the message rate variables   */
        if(lpMRB->oProcessingSecond != lpMRB->oNowSecond)
        {
            /* Copy the current time into the processing timestamp */
            memcpy(&lpMRB->oProcessingSecond, &lpMRB->oNowSecond, sizeof(time_t));

            /* Report target and actual messages rates */
            LogMsg(129, lpMRB->lMessagesThisSecond, lpMRB->lMessagesThisSecond - lpMRB->lMessagesLeft);

            /* Changed second so check if target was met */
            if(0 < lpMRB->lMessagesLeft)
            {
                kpMCB->lMessageRateMissed++;
            }
            else
            {
                kpMCB->lMessageRateAchieved++;
            }

            /* Resolve variables */
            ResolveValue(lpTIB, &lpMRB->lpszResFrom, &lpMRB->lpszFrom, TRUE);
            ResolveValue(lpTIB, &lpMRB->lpszResTo, &lpMRB->lpszTo, TRUE);
            ResolveValue(lpTIB, &lpMRB->lpszResIncr, &lpMRB->lpszIncr, TRUE);
            ResolveValue(lpTIB, &lpMRB->lpszResInterval, &lpMRB->lpszInterval, TRUE);
            if(FAILURE == lpTIB->bCmdResult)
            {
                return;
            }

            /* Convert to numeric values */
            lpMRB->lFrom = ConvNum(lpTIB, lpMRB->lpszResFrom, "From");
            lpMRB->lTo = ConvNum(lpTIB, lpMRB->lpszResTo, "To");
            lpMRB->lIncr = ConvNum(lpTIB, lpMRB->lpszResIncr, "Incr");
            lpMRB->lInterval = ConvNum(lpTIB, lpMRB->lpszResInterval, "Interval");
            if(FAILURE == lpTIB->bCmdResult)
            {
                return;
            }

            /* Recalculate message rate  */
            if(lpMRB->lFrom == lpMRB->lTo)
            {
                switch(lpMRB->eMsgRateMode)
                {
                case M_BURST:
                case M_PACED:
                case M_RANDOM:
                    lpMRB->lMessagesThisSecond = lpMRB->lFrom;
                    lpMRB->lMessagesLeft = lpMRB->lMessagesThisSecond - 1;
                    break;
                case M_INCREMENTAL:
                    lpMRB->lMessagesThisSecond = lpMRB->lFrom;
                    lpMRB->lMessagesLeft = lpMRB->lMessagesThisSecond - 1;
                    if(lpMRB->oNowSecond >= lpMRB->oIncrAtSecond)
                    {
                        lpMRB->bIncrementalDone = TRUE;
                        return;
                    }
                    break;
                }
                lpTIB->lTestCurrentMsgRate = lpMRB->lMessagesThisSecond;
            }
            else
            {
                switch(lpMRB->eMsgRateMode)
                {
                case M_BURST:
                case M_PACED:
                case M_RANDOM:
                    /* Generate a random number */
                    dRandomValue = ((double) rand()) / ((double) RAND_MAX);
                    dRandomMultiplier = (double)lpMRB->lTo - lpMRB->lFrom + 1;
                    lNewNum = (long)(dRandomValue * dRandomMultiplier);
                    lNewNum += lpMRB->lFrom;
                    lpMRB->lMessagesThisSecond = lNewNum;
                    lpMRB->lMessagesLeft = lpMRB->lMessagesThisSecond - 1;
                    break;
                case M_INCREMENTAL:
                    if(lpMRB->oNowSecond >= lpMRB->oIncrAtSecond)
                    {
                        /* Increase message rate by the increment */
                        lpMRB->lMessagesThisSecond += lpMRB->lIncr;
                        lpMRB->lMessagesLeft = lpMRB->lMessagesThisSecond - 1;
                        memcpy(&lpMRB->oIncrAtSecond, &lpMRB->oNowSecond, sizeof(time_t));
                        lpMRB->oIncrAtSecond += lpMRB->lInterval;
                    }
                    else
                    {
                        lpMRB->lMessagesLeft = lpMRB->lMessagesThisSecond - 1;
                    }
                    /* Check to see if we are done */
                    if(lpMRB->lMessagesThisSecond > lpMRB->lTo)
                    {
                        lpMRB->bIncrementalDone = TRUE;
                        return;
                    }
                    break;
                }
                lpTIB->lTestCurrentMsgRate = lpMRB->lMessagesThisSecond;
            }
            /* Log the selected message rate */
            LogMsg(132, lpTIB->lTestNum, lpMRB->lMessagesThisSecond, lpMRB->lFrom, lpMRB->lTo);


            /* Set the high resolution timer stuff for PACED, RANDOM and INCREMENTAL */
            GetClockCount(&lpMRB->i64ProcessingStart);
            memcpy(&lpMRB->i64ProcessingNow, &lpMRB->i64ProcessingStart, sizeof(Int64));
            memcpy(&lpMRB->i64ProcessingNext, &lpMRB->i64ProcessingStart, sizeof(Int64));

            /* Calculate high resolution timer stuff for PACED, RANDOM and INCREMENTAL */
            lpMRB->i64ProcessingIncr = kpMCB->i64CountFrequency / lpMRB->lMessagesThisSecond;
            switch(lpMRB->eMsgRateMode)
            {
            case M_BURST:
                lpMRB->i64ProcessingIncr = 0;
                break;
            case M_PACED:
                lpMRB->i64ProcessingNext += lpMRB->i64ProcessingIncr;
                break;
            case M_RANDOM:
                /* Calculate the value of the random increment */
                dRandomValue = ((double) rand()) / ((double) RAND_MAX);
                /* Multiply to cater for the fact that random averages out at 50%  */
                /* but not too much so we dont miss too many message rate targets. */
                /* The more messages the more self correcting we are likely to be. */
                if(20 < lpMRB->lMessagesThisSecond)
                {
                    dRandomMultiplier = ((double)lpMRB->i64ProcessingIncr) * 1.8;
                }
                else if(10 < lpMRB->lMessagesThisSecond)
                {
                    dRandomMultiplier = ((double)lpMRB->i64ProcessingIncr) * 1.4;
                }
                else
                {
                    dRandomMultiplier = (double)lpMRB->i64ProcessingIncr;
                }
                lpMRB->i64RandomIncr = (Int64)(dRandomValue * dRandomMultiplier);
                /* If its less than half a millisecond then make it a half millisecond */
                if(lpMRB->i64RandomIncr < (kpMCB->i64OneHalfMilliSecond))
                {
                    lpMRB->i64RandomIncr = kpMCB->i64OneHalfMilliSecond;
                }
                lpMRB->i64ProcessingNext += lpMRB->i64RandomIncr;
                break;
            case M_INCREMENTAL:
                lpMRB->i64ProcessingNext += lpMRB->i64ProcessingIncr;
                break;
            }
        }
    }
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     SleepMillis                                                          */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Sleep the specified number of milliseconds                           */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lSleepMillis Number of milli seconds to sleep                        */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
void
SleepMillis(long lSleepMillis)
{
#ifdef WIN32
    Sleep(lSleepMillis);
#else
    usleep(lSleepMillis * 1000);
#endif /* WIN32 */
}


/*==========================================================================*/
/* FUNCTION:                                                                */
/*     Get a high precsion clock count                                      */
/*                                                                          */
/* PURPOSE:                                                                 */
/*     Sleep the specified number of milliseconds                           */
/*                                                                          */
/* INPUT PARAMETERS                                                         */
/*     lSleepMillis Number of milli seconds to sleep                        */
/*                                                                          */
/* RETURN VALUE                                                             */
/*     N/A                                                                  */
/*==========================================================================*/
#ifdef WIN32
void GetClockCount(Int64* lpi64ClockCount)
{
    QueryPerformanceCounter((LARGE_INTEGER*)lpi64ClockCount);
}
#else
void GetClockCount(Int64* lpi64ClockCount)
{
    struct timezone oTimeZone;
    struct timeval  oTimeVal;
    Int64           i64MicroSecondsTvSec;
    Int64           i64MicroSecondsTvUSec;

    gettimeofday(&oTimeVal, &oTimeZone);

    /* Convert longs in the timeval to 64 bits before using them */
    i64MicroSecondsTvSec = oTimeVal.tv_sec;
    i64MicroSecondsTvSec *= 1000000;
    i64MicroSecondsTvUSec = oTimeVal.tv_usec;

    /* Add up the pieces and return the ClockCount */
    *lpi64ClockCount = i64MicroSecondsTvSec + i64MicroSecondsTvUSec;
}
#endif /* WIN32 */
